@@ -4,7 +4,7 @@
 /*
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (C) 1993-2016  Otfried Cheong
+    Copyright (c) 1993-2019 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -49,6 +49,7 @@ AppUiBase::AppUiBase(lua_State *L0, int model)
 {
   L = L0;
   iModel = model;
+  isInkMode = false;
 
   iMouseIn = 0; // points
   iMouseFactor = 1.0;  // scale 1:1
@@ -56,10 +57,42 @@ AppUiBase::AppUiBase(lua_State *L0, int model)
   lua_getglobal(L, "prefs");
   lua_getfield(L, -1, "coordinates_format");
   if (lua_isstring(L, -1)) {
-    iCoordinatesFormat = lua_tostring(L, -1);
+    iCoordinatesFormat = lua_tolstring(L, -1, nullptr);
   }
   lua_pop(L, 1); // coordinates_format
 
+  lua_getfield(L, -1, "width_notes_bookmarks");
+  if (lua_isnumber(L, -1)) {
+    iWidthNotesBookmarks = lua_tointegerx(L, -1, nullptr);
+  }
+  lua_pop(L, 1); // width_notes_bookmarks
+
+  iUiScale = 100;
+  lua_getfield(L, -1, "ui_scale");
+  if (lua_isnumber(L, -1))
+    iUiScale = lua_tointegerx(L, -1, nullptr);
+  lua_pop(L, 1); // ui_scale
+
+  iToolbarScale = 100;
+  lua_getfield(L, -1, "toolbar_scale");
+  if (lua_isnumber(L, -1))
+    iToolbarScale = lua_tointegerx(L, -1, nullptr);
+  lua_pop(L, 1); // toolbar_scale
+
+  iUiGap = 0;
+  lua_getfield(L, -1, "ui_gap");
+  if (lua_isnumber(L, -1))
+    iUiGap = lua_tointegerx(L, -1, nullptr);
+  lua_pop(L, 1); // ui_gap
+
+  lua_getfield(L, -1, "mini_edit");
+  isMiniEdit = lua_toboolean(L, -1);
+  lua_pop(L, 1); // mini_edit
+
+  lua_getfield(L, -1, "left_dock_floats");
+  iLeftDockFloats = lua_toboolean(L, -1);
+  lua_pop(L, 1); // left_dock_floats
+
   iScalings.push_back(1);
   lua_getfield(L, -1, "scale_factors");
   if (lua_istable(L, -1)) {
@@ -67,7 +100,7 @@ AppUiBase::AppUiBase(lua_State *L0, int model)
     for (int i = 1; i <= n; ++i) {
       lua_rawgeti(L, -1, i);
       if (lua_isnumber(L, -1))
-	iScalings.push_back(lua_tointeger(L, -1));
+	iScalings.push_back(lua_tointegerx(L, -1, nullptr));
       lua_pop(L, 1); // number
     }
   }
@@ -134,10 +167,16 @@ void AppUiBase::buildMenus()
   addItem(EEditMenu, "Duplicate", "duplicate");
   addItem(EEditMenu, "Select all", "select_all");
   addItem(EEditMenu);
+  addItem(EEditMenu, "Pick properties", "pick_properties");
+  addItem(EEditMenu, "Apply properties", "apply_properties");
+  addItem(EEditMenu);
   addItem(EEditMenu, "Insert text box", "insert_text_box");
+  addItem(EEditMenu, "Change text width", "change_width");
   addItem(EEditMenu, "Edit object", "edit");
   addItem(EEditMenu, "Edit object as XML", "edit_as_xml");
-  addItem(EEditMenu, "Change text width", "change_width");
+  addItem(EEditMenu);
+  addItem(EEditMenu, "Edit group", "edit_group");
+  addItem(EEditMenu, "End group edit", "end_group_edit");
   addItem(EEditMenu);
   addItem(EEditMenu, "Document properties", "document_properties");
   addItem(EEditMenu, "Style sheets", "style_sheets");
@@ -159,8 +198,10 @@ void AppUiBase::buildMenus()
 
   addItem(EPropertiesMenu);
 
-  startSubMenu(EPropertiesMenu, "Text style", ESubmenuTextStyle);
+  startSubMenu(EPropertiesMenu, "Minipage style", ESubmenuTextStyle);
   iTextStyleMenu = endSubMenu();
+  startSubMenu(EPropertiesMenu, "Label style", ESubmenuLabelStyle);
+  iLabelStyleMenu = endSubMenu();
 
   startSubMenu(EPropertiesMenu, "Horizontal alignment");
   addSubItem("left", "horizontalalignment|left");
@@ -187,6 +228,8 @@ void AppUiBase::buildMenus()
   addItem(EModeMenu, "Rotate objects", "mode_rotate");
   addItem(EModeMenu, "Stretch objects (with Shift: scale objects)",
 	  "mode_stretch");
+  addItem(EModeMenu, "Shear objects", "mode_shear");
+  addItem(EModeMenu, "Move graph nodes", "mode_graph");
   addItem(EModeMenu, "Pan the canvas", "mode_pan");
   addItem(EModeMenu, "Shred objects", "mode_shredder");
   addItem(EModeMenu);
@@ -194,7 +237,13 @@ void AppUiBase::buildMenus()
   addItem(EModeMenu, "Mathematical symbols", "mode_math");
   addItem(EModeMenu, "Paragraphs", "mode_paragraph");
   addItem(EModeMenu, "Marks", "mode_marks");
-  addItem(EModeMenu, "Rectangles (with Shift: squares)", "mode_rectangles");
+  addItem(EModeMenu, "Axis-parallel rectangles (with Shift: squares)",
+	  "mode_rectangles1");
+  addItem(EModeMenu, "Axis-parallel rectangles, by center (with Shift: squares)",
+	  "mode_rectangles2");
+  addItem(EModeMenu, "Rectangles (with Shift: squares)", "mode_rectangles3");
+  addItem(EModeMenu, "Parallelograms (with Shift: axis-parallel)",
+	  "mode_parallelogram");
   addItem(EModeMenu, "Lines and polylines", "mode_lines");
   addItem(EModeMenu, "Polygons", "mode_polygons");
   addItem(EModeMenu, "Splines", "mode_splines");
@@ -228,9 +277,11 @@ void AppUiBase::buildMenus()
   startSubMenu(ESnapMenu, "Radial angle", ESubmenuAngleSize);
   iAngleSizeMenu = endSubMenu();
   addItem(ESnapMenu);
+  addItem(ESnapMenu, "Set origin", "@set_origin");
   addItem(ESnapMenu, "Set origin && snap", "@set_origin_snap");
-  addItem(ESnapMenu, "Hide axes", "@hide_axes");
+  addItem(ESnapMenu, "Show axes", "@*show_axes");
   addItem(ESnapMenu, "Set direction", "@set_direction");
+  addItem(ESnapMenu, "Set tangent direction", "@set_tangent_direction");
   addItem(ESnapMenu, "Reset direction", "@reset_direction");
   addItem(ESnapMenu, "Set line", "@set_line");
   addItem(ESnapMenu, "Set line && snap", "@set_line_snap");
@@ -247,10 +298,9 @@ void AppUiBase::buildMenus()
   endSubMenu();
 
   startSubMenu(EZoomMenu, "Coordinate scale");
-  for (uint i = 0; i < iScalings.size(); ++i) {
+  for (int s : iScalings) {
     char display[32];
     char action[32];
-    int s = iScalings[i];
     if (s < 0)
       sprintf(display, "%d:1", -s);
     else
@@ -320,6 +370,7 @@ void AppUiBase::buildMenus()
   addItem(EHelpMenu, "Ipe &manual", "manual");
   addItem(EHelpMenu, "Onscreen keyboard", "@keyboard");
   addItem(EHelpMenu, "Show &configuration", "show_configuration");
+  addItem(EHelpMenu, "Show &libraries", "show_libraries");
   addItem(EHelpMenu, "&Ipelet information", "about_ipelets");
 
   lua_getglobal(L, "prefs");
@@ -346,10 +397,10 @@ void AppUiBase::buildMenus()
       lua_pop(L, 2); // label, ipelet
       continue;
     }
-    String label(lua_tostring(L, -1));
+    String label(lua_tolstring(L, -1, nullptr));
     lua_pop(L, 1);
     lua_getfield(L, -1, "name");
-    String name(lua_tostring(L, -1));
+    String name(lua_tolstring(L, -1, nullptr));
     lua_pop(L, 1);
     lua_getfield(L, -1, "methods");
     char buf[20];
@@ -366,7 +417,7 @@ void AppUiBase::buildMenus()
 	sprintf(buf, "ipelet_%d_", j);
 	String action(buf);
 	action += name;
-	addSubItem(lua_tostring(L, -1), action.z());
+	addSubItem(lua_tolstring(L, -1, nullptr), action.z());
 	lua_pop(L, 2); // sublabel, method
       }
       endSubMenu();
@@ -378,15 +429,17 @@ void AppUiBase::buildMenus()
 
 // --------------------------------------------------------------------
 
-void AppUiBase::canvasObserverWheelMoved(double degrees, bool precise)
+void AppUiBase::canvasObserverWheelMoved(double xDegrees, double yDegrees,
+					 int kind)
 {
-  if (degrees != 0.0) {
+  if (xDegrees != 0.0 || yDegrees != 0.0) {
     lua_rawgeti(L, LUA_REGISTRYINDEX, iModel);
     lua_getfield(L, -1, "wheel_zoom");
     lua_insert(L, -2); // model
-    lua_pushnumber(L, degrees);
-    lua_pushboolean(L, precise);
-    lua_call(L, 3, 0);
+    lua_pushnumber(L, xDegrees);
+    lua_pushnumber(L, yDegrees);
+    lua_pushinteger(L, kind);
+    luacall(L, 4, 0);
   } else
     // result of a zoom gesture (Windows only, currently)
     setZoom(iCanvas->zoom());
@@ -394,7 +447,7 @@ void AppUiBase::canvasObserverWheelMoved(double degrees, bool precise)
 
 void AppUiBase::canvasObserverToolChanged(bool hasTool)
 {
-  setActionsEnabled(!hasTool);
+  setActionsEnabled(!hasTool || isInkMode);
 }
 
 static void adjust(double &x, int mode, double factor)
@@ -441,7 +494,7 @@ void AppUiBase::canvasObserverMouseAction(int button)
   lua_getfield(L, -1, "mouseButtonAction");
   lua_insert(L, -2); // model
   push_button(L, button);
-  lua_call(L, 3, 0);
+  luacall(L, 3, 0);
 }
 
 void AppUiBase::canvasObserverSizeChanged()
@@ -449,7 +502,7 @@ void AppUiBase::canvasObserverSizeChanged()
   lua_rawgeti(L, LUA_REGISTRYINDEX, iModel);
   lua_getfield(L, -1, "sizeChanged");
   lua_insert(L, -2); // model
-  lua_call(L, 1, 0);
+  luacall(L, 1, 0);
 }
 
 // --------------------------------------------------------------------
@@ -478,7 +531,7 @@ void AppUiBase::luaSelector(String name, String value)
     lua_pushboolean(L, false);
   else
     push_string(L, value);
-  lua_call(L, 3, 0);
+  luacall(L, 3, 0);
 }
 
 void AppUiBase::luaAbsoluteButton(const char *s)
@@ -488,13 +541,15 @@ void AppUiBase::luaAbsoluteButton(const char *s)
   lua_getfield(L, -1, "absoluteButton");
   lua_insert(L, -2); // method, model
   lua_pushstring(L, s);
-  lua_call(L, 2, 0);
+  luacall(L, 2, 0);
 }
 
 // --------------------------------------------------------------------
 
 void AppUiBase::luaAction(String name)
 {
+  if (isInkMode && iCanvas->tool() != nullptr)
+    return; // refuse any action while drawing ink
   if (name.left(12) == "coordinates|") {
     if (name.right(2) == "mm")
       iMouseIn = 1;
@@ -521,7 +576,7 @@ void AppUiBase::luaAction(String name)
     lua_getfield(L, -1, "action");
     lua_insert(L, -2); // before model
     push_string(L, name);
-    lua_call(L, 2, 0);
+    luacall(L, 2, 0);
   }
 }
 
@@ -531,7 +586,7 @@ void AppUiBase::luaShowPathStylePopup(Vector v)
   lua_getfield(L, -1, "showPathStylePopup");
   lua_insert(L, -2); // before model
   push_vector(L, v);
-  lua_call(L, 2, 0);
+  luacall(L, 2, 0);
 }
 
 void AppUiBase::luaShowLayerBoxPopup(Vector v, String layer)
@@ -541,7 +596,7 @@ void AppUiBase::luaShowLayerBoxPopup(Vector v, String layer)
   lua_insert(L, -2); // before model
   push_vector(L, v);
   push_string(L, layer);
-  lua_call(L, 3, 0);
+  luacall(L, 3, 0);
 }
 
 void AppUiBase::luaLayerAction(String name, String layer)
@@ -551,7 +606,7 @@ void AppUiBase::luaLayerAction(String name, String layer)
   lua_insert(L, -2); // before model
   push_string(L, name);
   push_string(L, layer);
-  lua_call(L, 3, 0);
+  luacall(L, 3, 0);
 }
 
 void AppUiBase::luaBookmarkSelected(int index)
@@ -560,7 +615,7 @@ void AppUiBase::luaBookmarkSelected(int index)
   lua_getfield(L, -1, "bookmark");
   lua_insert(L, -2); // method, model
   lua_pushnumber(L, index + 1);
-  lua_call(L, 2, 0);
+  luacall(L, 2, 0);
 }
 
 // --------------------------------------------------------------------
@@ -580,12 +635,12 @@ void AppUiBase::showInCombo(const Cascade *sheet, Kind kind,
 {
   AttributeSeq seq;
   sheet->allNames(kind, seq);
-  if (!seq.size() && deflt != 0) {
+  if (!seq.size() && deflt != nullptr) {
     addCombo(sel, deflt);
     iComboContents[sel].push_back(deflt);
   }
-  for (uint i = 0; i < seq.size(); ++i) {
-    String s = seq[i].string();
+  for (const auto & att : seq) {
+    String s = att.string();
     addCombo(sel, s);
     iComboContents[sel].push_back(s);
   }
@@ -595,8 +650,8 @@ void AppUiBase::showMarksInCombo(const Cascade *sheet)
 {
   AttributeSeq seq;
   sheet->allNames(ESymbol, seq);
-  for (uint i = 0; i < seq.size(); ++i) {
-    String s = stripMark(seq[i]);
+  for (const auto & att : seq) {
+    String s = stripMark(att);
     if (!s.empty()) {
       addCombo(EUiMarkShape, s);
       iComboContents[EUiMarkShape].push_back(s);
@@ -611,8 +666,8 @@ void AppUiBase::setupSymbolicNames(const Cascade *sheet)
     iComboContents[i].clear();
   AttributeSeq seq, absColor;
   sheet->allNames(EColor, seq);
-  for (uint i = 0; i < seq.size(); ++i)
-    absColor.push_back(sheet->find(EColor, seq[i]));
+  for (const auto & att : seq)
+    absColor.push_back(sheet->find(EColor, att));
   addComboColors(seq, absColor);
   showInCombo(sheet, EPen, EUiPen);
   showInCombo(sheet, ETextSize, EUiTextSize);
@@ -628,7 +683,7 @@ void AppUiBase::setGridAngleSize(Attribute abs_grid, Attribute abs_angle)
   iCascade->allNames(EGridSize, seq);
   if (!seq.size())
     setComboCurrent(EUiGridSize, 0);
-  for (uint i = 0; i < seq.size(); ++i) {
+  for (int i = 0; i < size(seq); ++i) {
     if (iCascade->find(EGridSize, seq[i]) == abs_grid) {
       setComboCurrent(EUiGridSize, i);
       break;
@@ -638,7 +693,7 @@ void AppUiBase::setGridAngleSize(Attribute abs_grid, Attribute abs_angle)
   iCascade->allNames(EAngleSize, seq);
   if (!seq.size())
     setComboCurrent(EUiAngleSize, 0);
-  for (uint i = 0; i < seq.size(); ++i) {
+  for (int i = 0; i < size(seq); ++i) {
     if (iCascade->find(EAngleSize, seq[i]) == abs_angle) {
       setComboCurrent(EUiAngleSize, i);
       break;
