@@ -3,7 +3,7 @@
 /*
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (C) 1993-2016  Otfried Cheong
+    Copyright (c) 1993-2019 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -33,10 +33,6 @@
 
 #include "ipelua.h"
 
-// for version info only
-#include "ipefonts.h"
-#include <zlib.h>
-
 #include <cstdio>
 #include <cstdlib>
 
@@ -45,7 +41,8 @@ using namespace ipelua;
 
 #include "ipeuilayout_cocoa.h"
 
-extern CGContextRef ipeGetCGContext();
+// in ipeui_cocoa:
+extern NSImage *colorIcon(double red, double green, double blue, int pixels);
 
 // --------------------------------------------------------------------
 
@@ -54,6 +51,21 @@ static const char * const snapbutton_action[] = {
   "snapgrid", "snapangle", "snapauto",
   "grid_visible" };
 
+static const char * const touchbar_action[] = {
+  "escape", "set_origin", "set_direction", "set_line",
+  "show_axes", "reset_direction", "set_tangent_direction",
+  "snapvtx", "snapctl", "snapbd", "snapint",
+  "snapgrid", "snapangle", "snapauto" };
+
+static const char * const touchbar_titles[] = {
+  "esc", "org", "dir", "line", "axes", "reset", "tangent",
+  nullptr, nullptr, nullptr, nullptr, nullptr, nullptr, nullptr };
+
+static const char * const touchbar_customization_label[] = {
+  "Escape", "Set origin", "Set direction", "Set to edge", "Show axes", "Reset direction", "Set to tangent direction",
+  "Snap to vertex", "Snap to control point", "Snap to boundary", "Snap to intersection", "Snap to grid",
+  "Angular snap", "Automatic snap" };
+
 @interface IpeAction : NSObject
 @property NSString *name;
 @property NSString *title;
@@ -87,14 +99,6 @@ static const char * const snapbutton_action[] = {
 
 // --------------------------------------------------------------------
 
-@interface IpeMenuItem : NSMenuItem
-@property NSString * ipeAction;
-
-- (instancetype) initWithTitle:(NSString *) aTitle
-		     ipeAction:(NSString *) anIpeAction
-		 keyEquivalent:(NSString *) aKey;
-@end
-
 @implementation IpeMenuItem
 
 - (instancetype) initWithTitle:(NSString *) aTitle
@@ -121,6 +125,11 @@ static const char * const snapbutton_action[] = {
       self.action = @selector(selectAll:);
     else if ([_ipeAction isEqualToString:@"fullscreen"])
       self.action = @selector(toggleFullScreen:);
+    // make sure these work when no window exists
+    else if ([@[@"show_configuration", @"new_window", @"open",
+		@"manual", @"about_ipelets"]
+		 containsObject:anIpeAction])
+      self.action = @selector(ipeAlwaysAction:);
   }
   return self;
 }
@@ -129,23 +138,12 @@ static const char * const snapbutton_action[] = {
 
 // --------------------------------------------------------------------
 
-#define COLORICONSIZE 24
-
-static NSImage *colorIcon(Color color)
+static NSImage *colorIcon(Color color, int pixels = 24)
 {
   double red = color.iRed.toDouble();
   double green = color.iGreen.toDouble();
   double blue = color.iBlue.toDouble();
-  NSSize size = { COLORICONSIZE, COLORICONSIZE };
-  NSImage *icon = [NSImage imageWithSize:size
-				 flipped:NO
-			  drawingHandler:^(NSRect rect) {
-      CGContextRef myContext = ipeGetCGContext();
-      CGContextSetRGBFillColor(myContext, red, green, blue, 1.0);
-      CGContextFillRect(myContext, CGRectMake(0, 0, COLORICONSIZE, COLORICONSIZE));
-      return YES;
-    }];
-  return icon;
+  return colorIcon(red, green, blue, pixels);
 }
 
 static NSImage *loadIcon(String action)
@@ -162,18 +160,9 @@ static NSImage *loadIcon(String action)
 
 // --------------------------------------------------------------------
 
-@interface IpeWindowDelegate :
-  NSObject <NSWindowDelegate, NSToolbarDelegate,
-	      NSComboBoxDelegate, IpeControlsDelegate>
-
-- (instancetype) initFor:(AppUi *) ui;
-- (void) ipeMenuAction:(id) sender;
-- (void) ipeToolbarAction:(id) sender;
-- (void) ipeAbsoluteButton:(id) sender;
-- (void) ipeSnapButton:(id) sender;
-- (void) toggleSnapbarShown:(id) sender;
-- (void) ipeSubmenu:(id) sender;
-
+// Window delegate is also an NSWindowController object to handle touch bar
+@interface IpeWindowDelegate : NSWindowController
+  <NSWindowDelegate, NSToolbarDelegate, IpeControlsDelegate, NSTouchBarDelegate>
 @end
 
 @implementation IpeWindowDelegate {
@@ -182,18 +171,13 @@ static NSImage *loadIcon(String action)
 
 - (instancetype) initFor:(AppUi *) ui
 {
-  self = [super init];
+  self = [super initWithWindow:nil];
   if (self) {
     appui = ui;
   }
   return self;
 }
 
-- (void) aboutIpe:(id) sender
-{
-  appui->aboutIpe();
-}
-
 - (void) ipeSubmenu:(id) sender
 {
   NSLog(@"ipeSubmenu: %@", sender);
@@ -212,6 +196,11 @@ static NSImage *loadIcon(String action)
   return YES;
 }
 
+- (BOOL) ipeIsModified:(id) sender
+{
+  return appui->isModified();
+}
+
 - (BOOL) windowShouldClose:(id) sender
 {
   return appui->closeRequested();
@@ -277,9 +266,9 @@ static NSImage *loadIcon(String action)
   appui->snapButton([sender tag]);
 }
 
-- (void) comboBoxSelectionDidChange:(NSNotification *) notification
+- (void) ipeSelectorChanged:(id) sender
 {
-  appui->selectorChanged([[notification object] tag]);
+  appui->selectorChanged([sender tag]);
 }
 
 - (void) toggleSnapbarShown:(id) sender
@@ -342,44 +331,103 @@ static NSImage *loadIcon(String action)
 - (NSArray<NSString *> *) toolbarAllowedItemIdentifiers:(NSToolbar *) toolbar
 {
   return @[ @"copy", @"cut", @"paste", @"delete", @"undo", @"redo",
-	    @"zoom_in", @"zoom_out", @"fit_objects", @"fit_page",
-	    @"fit_width", @"keyboard",
-	    @"mode_select", @"mode_translate", @"mode_rotate",
-	    @"mode_stretch", @"mode_pan", @"mode_shredder",
-	    @"mode_label", @"mode_math", @"mode_paragraph",
-	    @"mode_marks", @"mode_rectangles", @"mode_lines",
-	    @"mode_polygons", @"mode_splines", @"mode_splinegons",
-	    @"mode_arc1", @"mode_arc2", @"mode_arc3",
-	    @"mode_circle1", @"mode_circle2", @"mode_circle3",
-	    @"mode_ink",
-	    NSToolbarSpaceItemIdentifier,
-	    NSToolbarFlexibleSpaceItemIdentifier,
-	   ];
+	     @"zoom_in", @"zoom_out", @"fit_objects", @"fit_page",
+	     @"fit_width", @"keyboard",
+	     @"mode_select", @"mode_translate", @"mode_rotate",
+	     @"mode_stretch", @"mode_shear", @"mode_graph",
+	     @"mode_pan", @"mode_shredder",
+	     @"mode_label", @"mode_math", @"mode_paragraph",
+	     @"mode_marks", @"mode_rectangles1", @"mode_rectangles2",
+	     @"mode_rectangles3", @"mode_parallelogram", @"mode_lines",
+	     @"mode_polygons", @"mode_splines", @"mode_splinegons",
+	     @"mode_arc1", @"mode_arc2", @"mode_arc3",
+	     @"mode_circle1", @"mode_circle2", @"mode_circle3",
+	     @"mode_ink",
+	     NSToolbarSpaceItemIdentifier,
+	     NSToolbarFlexibleSpaceItemIdentifier,
+	    ];
 }
 
 - (NSArray<NSString *> *) toolbarDefaultItemIdentifiers:(NSToolbar *) toolbar
 {
   return @[ @"mode_select", @"mode_translate", @"mode_rotate",
-	    @"mode_stretch", @"mode_pan", @"mode_shredder",
-	    @"mode_label", @"mode_math", @"mode_paragraph",
-	    @"mode_marks", @"mode_rectangles", @"mode_lines",
-	    @"mode_polygons", @"mode_splines", @"mode_splinegons",
-	    @"mode_arc1", @"mode_arc2", @"mode_arc3",
-	    @"mode_circle1", @"mode_circle2", @"mode_circle3",
-	    @"mode_ink",
+	     @"mode_stretch", @"mode_shear", @"mode_graph",
+	     @"mode_pan", @"mode_shredder",
+	     @"mode_label", @"mode_math", @"mode_paragraph",
+	     @"mode_marks", @"mode_rectangles1", @"mode_rectangles2",
+	     @"mode_rectangles3", @"mode_parallelogram", @"mode_lines",
+	     @"mode_polygons", @"mode_splines", @"mode_splinegons",
+	     @"mode_arc1", @"mode_arc2", @"mode_arc3",
+	     @"mode_circle1", @"mode_circle2", @"mode_circle3",
+	     @"mode_ink",
 	   ];
 }
 
 - (NSArray *) toolbarSelectableItemIdentifiers: (NSToolbar *) toolbar
 {
   return @[ @"mode_select", @"mode_translate", @"mode_rotate",
-	    @"mode_stretch", @"mode_pan", @"mode_shredder",
-	    @"mode_label", @"mode_math", @"mode_paragraph",
-	    @"mode_marks", @"mode_rectangles", @"mode_lines",
-	    @"mode_polygons", @"mode_splines", @"mode_splinegons",
-	    @"mode_arc1", @"mode_arc2", @"mode_arc3",
-	    @"mode_circle1", @"mode_circle2", @"mode_circle3",
-	    @"mode_ink" ];
+	     @"mode_stretch", @"mode_shear", @"mode_graph",
+	     @"mode_pan", @"mode_shredder",
+	     @"mode_label", @"mode_math", @"mode_paragraph",
+	     @"mode_marks", @"mode_rectangles1", @"mode_rectangles2",
+	     @"mode_rectangles2", @"mode_rectangles3",
+	     @"mode_parallelogram", @"mode_lines",
+	     @"mode_polygons", @"mode_splines", @"mode_splinegons",
+	     @"mode_arc1", @"mode_arc2", @"mode_arc3",
+	     @"mode_circle1", @"mode_circle2", @"mode_circle3",
+	     @"mode_ink" ];
+}
+
+#if MAC_OS_X_VERSION_MAX_ALLOWED >= 101202
+#pragma clang diagnostic push
+#pragma clang diagnostic ignored "-Wdeprecated-declarations"
+- (NSTouchBar *) makeTouchBar
+{
+  NSTouchBar *bar = [[NSTouchBar alloc] init];
+  bar.delegate = self;
+  bar.defaultItemIdentifiers = @[ @"show_axes", @"set_origin", @"set_direction", @"snapvtx", @"snapctl", @"snapbd", @"snapint", @"snapgrid" ];
+  bar.customizationIdentifier = @"org.otfried.ipe.mainbar7_2_11";
+  bar.customizationAllowedItemIdentifiers =
+    @[ @"show_axes", @"set_origin", @"set_direction", @"reset_direction",
+	@"set_line", @"set_tangent_direction",
+	@"snapvtx", @"snapctl", @"snapbd", @"snapint", @"snapgrid", @"snapangle", @"snapauto" ];
+  bar.escapeKeyReplacementItemIdentifier = @"escape";
+  return bar;
+}
+
+- (NSTouchBarItem *) touchBar: (NSTouchBar *) bar
+	makeItemForIdentifier: (NSTouchBarItemIdentifier) identifier
+{
+  String idf = N2I(identifier);
+  for (size_t tag = 0; tag < (sizeof(touchbar_action)/sizeof(const char *)); ++tag) {
+    if (idf == touchbar_action[tag]) {
+      NSCustomTouchBarItem *item = [[NSCustomTouchBarItem alloc] initWithIdentifier:identifier];
+      auto title = touchbar_titles[tag];
+      NSButton *b;
+      if (title == nullptr) {
+	auto image = loadIcon(touchbar_action[tag]);
+	b = [NSButton buttonWithTitle:identifier image:image target:self action:@selector(ipeTouchBar:)];
+	b.buttonType = NSPushOnPushOffButton;
+	b.imagePosition = NSImageOnly;
+      } else {
+	b = [NSButton buttonWithTitle:C2N(title) target:self action:@selector(ipeTouchBar:)];
+	if (idf == "show_axes")
+	  b.buttonType = NSPushOnPushOffButton;
+      }
+      b.tag = tag;
+      item.view = b;
+      item.customizationLabel = C2N(touchbar_customization_label[tag]);
+      return item;
+    }
+  }
+  return nil;
+}
+#pragma clang diagnostic pop
+#endif
+
+- (void) ipeTouchBar:(id) sender
+{
+  appui->action(touchbar_action[[sender tag]]);
 }
 @end
 
@@ -394,7 +442,7 @@ void AppUi::addRootMenu(int id, const char *name)
 
 void AppUi::addItem(NSMenu *menu, const char *title, const char *name)
 {
-  if (title == 0) {
+  if (title == nullptr) {
     if (build_menus)
       [menu addItem:[NSMenuItem separatorItem]];
   } else {
@@ -502,7 +550,7 @@ void AppUi::addItem(int id, const char *title, const char *name)
   addItem(menu, title, name);
 }
 
-static NSMenu * __unsafe_unretained current_submenu = 0;
+static NSMenu * __unsafe_unretained current_submenu = nullptr;
 
 void AppUi::startSubMenu(int id, const char *name, int tag)
 {
@@ -513,7 +561,7 @@ void AppUi::startSubMenu(int id, const char *name, int tag)
 
   NSString *title = ipeui_set_mnemonic(C2N(name), nil);
   NSMenuItem *item = [[NSMenuItem alloc] initWithTitle:title
-						action:NULL
+						action:nullptr
 					 keyEquivalent:@""];
   NSMenu *submenu = [[NSMenu alloc] initWithTitle:title];
   item.submenu = submenu;
@@ -578,6 +626,8 @@ AppUi::AppUi(lua_State *L0, int model)
   [tb setSizeMode:NSToolbarSizeModeSmall];
   [iWindow setToolbar:tb];
 
+  [iWindow setWindowController:iDelegate];
+
   iContent = [[NSView alloc] initWithFrame:contentRect];
 
   makePropertiesTool();
@@ -585,7 +635,7 @@ AppUi::AppUi(lua_State *L0, int model)
   iLayerBox = [[NSBox alloc] initWithFrame:subRect];
   iLayerBox.title = @"Layers";
 
-  NSRect layerFrame = iLayerBox.contentView.frame;
+  NSRect layerFrame = [iLayerBox.contentView frame];
   iLayerView = [[IpeLayerView alloc] initWithFrame:layerFrame];
   iLayerView.delegate = iDelegate;
   iLayerView.toolTip = @"Layers of this page";
@@ -661,6 +711,9 @@ AppUi::AppUi(lua_State *L0, int model)
 	       forOrientation:NSLayoutConstraintOrientationHorizontal];
 
   [iWindow setContentView:iContent];
+
+  setCheckMark("coordinates|", "points");
+  setCheckMark("scaling|", "1");
 }
 
 AppUi::~AppUi()
@@ -680,15 +733,16 @@ void AppUi::makePropertiesTool()
     [iButton[i] setButtonType:NSMomentaryPushInButton];
     [iButton[i] setImagePosition:NSImageOnly];
     [iButton[i] setBezelStyle:NSRegularSquareBezelStyle];
-    [iButton[i] setAction:@selector(ipeAbsoluteButton:)];
-    [iButton[i] setTag:i];
+    iButton[i].action = @selector(ipeAbsoluteButton:);
+    iButton[i].tag = i;
     addToLayout(iPropertiesBox, iButton[i]);
 
-    iSelector[i] = [[NSComboBox alloc]
-		     initWithFrame:NSMakeRect(0., 0., 50., 20.)];
-    [iSelector[i] setEditable:NO];
-    [iSelector[i] setDelegate:iDelegate];
-    [iSelector[i] setTag:i];
+    iSelector[i] = [[NSPopUpButton alloc]
+		     initWithFrame:NSMakeRect(0., 0., 50., 20.)
+			 pullsDown:NO];
+    iSelector[i].target = iDelegate;
+    iSelector[i].action = @selector(ipeSelectorChanged:);
+    iSelector[i].tag = i;
     addToLayout(iPropertiesBox, iSelector[i]);
   }
   iButton[EUiMarkShape].enabled = NO;
@@ -826,18 +880,20 @@ void AppUi::makeSnapBar()
   }
 
   iSelector[EUiGridSize] =
-    [[NSComboBox alloc] initWithFrame:NSMakeRect(0, 0, 100, 40)];
-  iSelector[EUiGridSize].editable = NO;
+    [[NSPopUpButton alloc] initWithFrame:NSMakeRect(0, 0, 100, 40)
+			       pullsDown:NO];
   iSelector[EUiGridSize].toolTip = @"Grid size";
-  iSelector[EUiGridSize].delegate = iDelegate;
+  iSelector[EUiGridSize].target = iDelegate;
+  iSelector[EUiGridSize].action = @selector(ipeSelectorChanged:);
   iSelector[EUiGridSize].tag = EUiGridSize;
   addToLayout(iSnapBar, iSelector[EUiGridSize]);
 
   iSelector[EUiAngleSize] =
-    [[NSComboBox alloc] initWithFrame:NSMakeRect(0, 0, 100, 40)];
-  iSelector[EUiAngleSize].editable = NO;
+    [[NSPopUpButton alloc] initWithFrame:NSMakeRect(0, 0, 100, 40)
+			       pullsDown:NO];
   iSelector[EUiAngleSize].toolTip = @"Angle for angular snap";
-  iSelector[EUiAngleSize].delegate = iDelegate;
+  iSelector[EUiAngleSize].target = iDelegate;
+  iSelector[EUiAngleSize].action = @selector(ipeSelectorChanged:);
   iSelector[EUiAngleSize].tag = EUiAngleSize;
   addToLayout(iSnapBar, iSelector[EUiAngleSize]);
 
@@ -851,6 +907,8 @@ void AppUi::makeSnapBar()
   layout(iSelector[EUiAngleSize], iSnapBar, "t=t", MARGIN);
   layout(iSnapBar, iSelector[EUiGridSize], "b>b", MARGIN);
   layout(iSnapBar, iSelector[EUiAngleSize], "b>b", MARGIN);
+  layout(iSelector[EUiGridSize], nil, "w>0", 160);
+  layout(iSelector[EUiAngleSize], nil, "w>0", 100);
 
   layout(iSnapButton[0], iSnapBar, "l=l", MARGIN);
   layout(iSelector[EUiGridSize], iSnapButton[4], "l=r", PAD);
@@ -875,11 +933,13 @@ void AppUi::addComboColors(AttributeSeq &sym, AttributeSeq &abs)
 {
   iInUiUpdate = true;
   for (uint i = 0; i < sym.size(); ++i) {
-    // Color color = abs[i].color();
-    // NSImage *im = colorIcon(color);
+    Color color = abs[i].color();
+    NSImage *im = colorIcon(color, 12);
     String s = sym[i].string();
-    [iSelector[EUiStroke] addItemWithObjectValue:I2N(s)];
-    [iSelector[EUiFill] addItemWithObjectValue:I2N(s)];
+    [iSelector[EUiStroke] addItemWithTitle:I2N(s)];
+    [iSelector[EUiFill] addItemWithTitle:I2N(s)];
+    [iSelector[EUiStroke] lastItem].image = im;
+    [iSelector[EUiFill] lastItem].image = im;
     iComboContents[EUiStroke].push_back(s);
     iComboContents[EUiFill].push_back(s);
   }
@@ -889,7 +949,7 @@ void AppUi::addComboColors(AttributeSeq &sym, AttributeSeq &abs)
 void AppUi::addCombo(int sel, String s)
 {
   iInUiUpdate = true;
-  [iSelector[sel] addItemWithObjectValue:I2N(s)];
+  [iSelector[sel] addItemWithTitle:I2N(s)];
   iInUiUpdate = false;
 }
 
@@ -955,16 +1015,16 @@ void AppUi::setNumbers(String vno, bool vm, String pno, bool pm)
 BOOL AppUi::validateMenuItem(NSMenuItem *item, NSString *name)
 {
   if ([name isEqualToString:@"snapbar"]) {
-    item.title = iSnapBar.hidden ? @"Show Snap Toolbar" :
+    item.title = [iSnapBar isHidden] ? @"Show Snap Toolbar" :
       @"Hide Snap Toolbar";
     return YES;
   }
   IpeAction *s = findAction(name);
   if (s) {
     if ([name isEqualToString:@"toggle_notes"])
-      s.state = (iNotesPanel && iNotesPanel.visible);
+      s.state = (iNotesPanel && [iNotesPanel isVisible]);
     else if ([name isEqualToString:@"toggle_bookmarks"])
-      s.state = (iBookmarksPanel && iBookmarksPanel.visible);
+      s.state = (iBookmarksPanel && [iBookmarksPanel isVisible]);
     [item setState:(s.state ? NSOnState : NSOffState)];
     return actionsEnabled() || s.alwaysOn;
   }
@@ -1024,8 +1084,8 @@ void AppUi::fillDynamicSubmenu(NSMenuItem *item)
 
 void AppUi::toggleSnapbarShown()
 {
-  iSnapBar.hidden = !iSnapBar.hidden;
-  if (iSnapBar.hidden) {
+  iSnapBar.hidden = ![iSnapBar isHidden];
+  if ([iSnapBar isHidden]) {
     activateConstraint(iViewToSnapBar, NO);
     activateConstraint(iViewToTop, YES);
   } else {
@@ -1044,6 +1104,7 @@ void AppUi::setNotes(String notes)
     NSAttributedString *n =
       [[NSAttributedString alloc] initWithString:I2N(notes)];
     [s setAttributedString:n];
+    [iNotesField setTextColor: [NSColor textColor]];
   }
 }
 
@@ -1121,7 +1182,6 @@ void AppUi::absoluteButton(int sel)
 void AppUi::snapButton(int sel)
 {
   NSString *a = C2N(snapbutton_action[sel]);
-  NSLog(@"snapButton %d -> %@", sel, a);
   IpeAction *s = findAction(a);
   if (s) {
     s.state = (iSnapButton[sel].state == NSOnState);
@@ -1143,7 +1203,7 @@ IpeAction *AppUi::findAction(NSString *name) const
 }
 
 // Determine if an action is checked.
-// Used for viewmarked, pagemarked, snapXXX, grid_visible,
+// Used for viewmarked, pagemarked, snapXXX, grid_visible, show_axes,
 // pretty_display, toggle_notes, toggle_bookmarks.
 bool AppUi::actionState(const char *name)
 {
@@ -1171,11 +1231,26 @@ void AppUi::setActionState(const char *name, bool value)
       iModeIndicator.image = loadIcon(name);
       iWindow.toolbar.selectedItemIdentifier = a;
     }
-  } else if ([a hasPrefix:@"snap"] || [a isEqualToString:@"grid_visible"]) {
+  } else if ([a hasPrefix:@"snap"] || [a isEqualToString:@"grid_visible"] ||
+	     [a isEqualToString:@"show_axes"]) {
     for (int i = 0; i < 8; ++i) {
       if ([a isEqualToString:C2N(snapbutton_action[i])])
 	iSnapButton[i].state = value ? NSOnState : NSOffState;
     }
+    // update touch bar if it exists
+#if MAC_OS_X_VERSION_MAX_ALLOWED >= 101202
+#pragma clang diagnostic push
+#pragma clang diagnostic ignored "-Wdeprecated-declarations"
+    if (iDelegate.touchBar != nil) {
+      auto bar = (NSTouchBar *) iDelegate.touchBar;
+      auto item = [bar itemForIdentifier:a];
+      if (item != nil) {
+	auto b = (NSButton *) item.view;
+	b.state =  value ? NSOnState : NSOffState;
+      }
+    }
+#pragma clang diagnostic push
+#endif
   } else if ([a isEqualToString:@"shift_key"]) {
     int mod = s.state ? CanvasBase::EShift : 0;
     if (iCanvas)
@@ -1186,8 +1261,9 @@ void AppUi::setActionState(const char *name, bool value)
 void AppUi::action(String name)
 {
   ipeDebug("action %s", name.z());
-  if (name == "about") {
-    aboutIpe();
+  if (name == "escape") {
+    if (iCanvas && iCanvas->tool())
+      iCanvas->tool()->key("\033", 0);
     return;
   }
   // Implement radio buttons
@@ -1275,6 +1351,18 @@ BOOL AppUi::closeRequested()
   return result;
 }
 
+BOOL AppUi::isModified()
+{
+  // calls model
+  lua_rawgeti(L, LUA_REGISTRYINDEX, iModel);
+  lua_getfield(L, -1, "isModified");
+  lua_pushvalue(L, -2); // model
+  lua_remove(L, -3);
+  lua_call(L, 1, 1);
+  bool result = lua_toboolean(L, -1);
+  return result;
+}
+
 void AppUi::closeWindow()
 {
   [iWindow performClose:iDelegate];
@@ -1282,43 +1370,6 @@ void AppUi::closeWindow()
 
 // --------------------------------------------------------------------
 
-static const char * const about_text =
-  "Copyright (c) 1993-2016 Otfried Cheong\n\n"
-  "The extensible drawing editor Ipe creates figures "
-  "in Postscript and PDF format, "
-  "using LaTeX to format the text in the figures.\n"
-  "Ipe relies on the following fine pieces of software:\n\n"
-  " * Pdftex, Xetex, or Luatex\n"
-  " * Pdflatex\n"
-  " * %s (%d kB used)\n" // Lua
-  " * The font rendering library %s\n"
-  " * The rendering library Cairo %s\n"
-  " * The compression library zlib %s\n\n"
-  "Ipe is released under the GNU Public License.\n"
-  "See http://ipe.otfried.org for details.";
-
-void AppUi::aboutIpe()
-{
-  int luaSize = lua_gc(L, LUA_GCCOUNT, 0);
-  NSString *text = [NSString stringWithFormat:@"Ipe %d.%d.%d",
-			     IPELIB_VERSION / 10000,
-			     (IPELIB_VERSION / 100) % 100,
-			     IPELIB_VERSION % 100];
-  NSString *details = [NSString stringWithFormat:@(about_text),
-				LUA_RELEASE, luaSize,
-				Fonts::freetypeVersion().z(),
-				cairo_version_string(),
-				ZLIB_VERSION];
-
-  NSAlert *alert = [[NSAlert alloc] init];
-  [alert setMessageText:text];
-  [alert setInformativeText:details];
-  [alert setAlertStyle:NSInformationalAlertStyle];
-  [alert runModal];
-}
-
-// --------------------------------------------------------------------
-
 int AppUi::clipboard(lua_State *L)
 {
   // bool allow_bitmap = lua_toboolean(L, 2);
