@@ -4,7 +4,7 @@
 /*
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (C) 1993-2016  Otfried Cheong
+    Copyright (c) 1993-2019 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -34,10 +34,6 @@
 
 #include "ipelua.h"
 
-// for version info only
-#include "ipefonts.h"
-#include <zlib.h>
-
 #include <cstdio>
 #include <cstdlib>
 
@@ -60,23 +56,53 @@ using namespace ipelua;
 
 #define SIZE 16
 
+inline QSize adapt_size(const QSize &size, int factor)
+{
+  return QSize(factor * size.width() / 100, factor * size.height() / 100);
+}
+
+static bool add_bitmap(QIcon &icon, String base, const char *ext)
+{
+  if (Platform::fileExists(base + ext)) {
+    icon.addFile(QIpe(base + ext));
+    return true;
+  }
+  return false;
+}
+
 QIcon prefsIcon(String name)
 {
   String dir = ipeIconDirectory();
-  if (Platform::fileExists(dir + name + ".png"))
-    return QIcon(QIpe(dir + name + ".png"));
-  else
+  String base = dir + name;
+  if (Platform::fileExists(base + ".png")) {
+    QIcon icon(QIpe(base + ".png"));
+    if (!add_bitmap(icon, base, "@2x.png"))
+      icon.addPixmap(icon.pixmap(QSize(24, 24)).scaledToHeight(48));
+    add_bitmap(icon, base, "_32x32.png");
+    add_bitmap(icon, base, "_32x32@2x.png");
+    add_bitmap(icon, base, "128.png");  // for application icon
+    add_bitmap(icon, base, "512.png");  // for application icon
+    /*
+    auto l = icon.availableSizes();
+    for (int i = 0; i < l.size(); ++i)
+      ipeDebug("Icon %s has size %d x %d", name.z(), l[i].width(), l[i].height());
+    */
+    return icon;
+  } else
     return QIcon();
 }
 
-QPixmap prefsPixmap(String name)
+QPixmap AppUi::prefsPixmap(String name)
 {
-  return QPixmap(QIpe(ipeIconDirectory() + name + ".png"));
+  QIcon icon = prefsIcon(name);
+  int size = iUiScale * 24 / 100;
+  return icon.pixmap(QSize(size, size));
 }
 
-QIcon prefsColorIcon(Color color)
+QIcon AppUi::prefsColorIcon(Color color)
 {
-  QPixmap pixmap(SIZE, SIZE);
+  int size = iUiScale * SIZE / 100;
+  QPixmap pixmap(size, size);
   pixmap.fill(QIpe(color));
   return QIcon(pixmap);
 }
@@ -89,12 +115,11 @@ QAction *AppUi::findAction(const char *name) const
   if (it != iActions.end())
     return it->second;
   else
-    return 0;
+    return nullptr;
 }
 
 void AppUi::addItem(QMenu *m, const QString &title, const char *name)
 {
-  // bool canUseWhileDrawing = false;
   if (name[0] == '@') {
     // canUseWhileDrawing = true;
     name = name + 1;
@@ -106,6 +131,7 @@ void AppUi::addItem(QMenu *m, const QString &title, const char *name)
     name = name + 1;
   }
   QAction *a = new QAction(title, this);
+  a->setIconVisibleInMenu(false);
   if (checkable)
     a->setCheckable(true);
   lua_getglobal(L, "shortcuts");
@@ -151,13 +177,13 @@ void AppUi::addRootMenu(int id, const char *name)
 // if title and name == 0, add separator (default)
 void AppUi::addItem(int id, const char *title, const char *name)
 {
-  if (title == 0)
+  if (title == nullptr)
     iMenu[id]->addSeparator();
   else
     addItem(iMenu[id], QString::fromUtf8(title), name);
 }
 
-static QMenu *submenu = 0;
+static QMenu *submenu = nullptr;
 static int submenuId = 0;
 
 void AppUi::startSubMenu(int id, const char *name, int tag)
@@ -182,33 +208,44 @@ Qt::DockWidgetArea getDockSide(lua_State *L, const char *name,
 {
   Qt::DockWidgetArea r = deflt;
   lua_getglobal(L, "prefs");
-  lua_getfield(L, -1, "tools");
+  lua_getfield(L, -1, "tools_placement");
   if (lua_istable(L, -1)) {
     lua_getfield(L, -1, name);
-    const char *s = lua_tolstring(L, -1, 0);
+    const char *s = lua_tolstring(L, -1, nullptr);
     if (!strcmp(s, "left"))
       r = Qt::LeftDockWidgetArea;
     else if (!strcmp(s, "right"))
       r = Qt::RightDockWidgetArea;
     lua_pop(L, 1); // left or right
   }
-  lua_pop(L, 2); // prefs, tools
+  lua_pop(L, 2); // prefs, tools_placement
   return r;
 }
 
-// --------------------------------------------------------------------
+static void set_toolbar_size(QToolBar *bar, int factor)
+{
+  QSize size = bar->iconSize();
+  QSize size1 = adapt_size(size, factor);
+  bar->setIconSize(size1);
+  /*
+  ipeDebug("toolbar %s has size %d x %d, changed to %d x %d",
+	   bar->windowTitle().toUtf8().constData(), size.width(), size.height(),
+	   size1.width(), size1.height());
+  */
+}
 
-extern void qt_set_sequence_auto_mnemonic(bool b);
+// --------------------------------------------------------------------
 
 #if QT_VERSION < 0x050000
 AppUi::AppUi(lua_State *L0, int model, Qt::WFlags f)
 #else
 AppUi::AppUi(lua_State *L0, int model, Qt::WindowFlags f)
 #endif
-  : QMainWindow(0, f), AppUiBase(L0, model)
+    : QMainWindow(nullptr, f), AppUiBase(L0, model)
 {
   setWindowIcon(prefsIcon("ipe"));
   QMainWindow::setAttribute(Qt::WA_DeleteOnClose);
+  qApp->setAttribute(Qt::AA_UseHighDpiPixmaps);
   setDockOptions(AnimatedDocks); // do not allow stacking properties and layers
 
   Canvas *canvas = new Canvas(this);
@@ -220,6 +257,11 @@ AppUi::AppUi(lua_State *L0, int model, Qt::WindowFlags f)
   addToolBarBreak();
   iObjectTools = addToolBar("Objects");
 
+  ipeDebug("devicePixelRatio: %g", qApp->devicePixelRatio());
+  set_toolbar_size(iEditTools, iToolbarScale);
+  set_toolbar_size(iSnapTools, iToolbarScale);
+  set_toolbar_size(iObjectTools, iToolbarScale);
+
   iActionMap = new QSignalMapper(this);
   connect(iActionMap, SIGNAL(mapped(const QString &)),
 	  SLOT(qAction(const QString &)));
@@ -250,6 +292,8 @@ AppUi::AppUi(lua_State *L0, int model, Qt::WindowFlags f)
 	  SLOT(moveToLayerAction(QAction *)));
   connect(iTextStyleMenu, SIGNAL(triggered(QAction *)),
 	  SLOT(textStyleAction(QAction *)));
+  connect(iLabelStyleMenu, SIGNAL(triggered(QAction *)),
+	  SLOT(labelStyleAction(QAction *)));
   connect(iGridSizeMenu, SIGNAL(triggered(QAction *)),
 	  SLOT(gridSizeAction(QAction *)));
   connect(iAngleSizeMenu, SIGNAL(triggered(QAction *)),
@@ -337,20 +381,22 @@ AppUi::AppUi(lua_State *L0, int model, Qt::WindowFlags f)
   QWidget *wg = new QFrame();
   QGridLayout *lo = new QGridLayout();
   wg->setLayout(lo);
-  lo->setSpacing(0);
-  lo->setContentsMargins(0, 0, 0, 0);
+  int m = iUiScale * 2 / 100;
+  lo->setSpacing(1);
+  lo->setContentsMargins(m, m, m, m); // l t r b
   // wg->setMaximumWidth(150);
   lo->setSizeConstraint(QLayout::SetFixedSize);
   QButtonGroup *bg = new QButtonGroup(wg);
   bg->setExclusive(false);
   connect(bg, SIGNAL(buttonClicked(int)), this, SLOT(absoluteButton(int)));
-  iButton[EUiMarkShape] = 0; // no such button
+  iButton[EUiMarkShape] = nullptr;   // no such button
   for (int i = 0; i <= EUiSymbolSize; ++i) {
     if (i != EUiMarkShape) {
       iButton[i] = new QToolButton();
       iButton[i]->setFocusPolicy(Qt::NoFocus);
       bg->addButton(iButton[i], i);
       lo->addWidget(iButton[i], i > 2 ? i+1 : i, 0);
+      iButton[i]->setIconSize(adapt_size(iButton[i]->iconSize(), iUiScale));
     }
     iSelector[i] = new QComboBox();
     iSelector[i]->setFocusPolicy(Qt::NoFocus);
@@ -450,6 +496,8 @@ AppUi::AppUi(lua_State *L0, int model, Qt::WindowFlags f)
 	  this, SLOT(aboutToShowMoveToLayerMenu()));
   connect(iTextStyleMenu, SIGNAL(aboutToShow()),
 	  this, SLOT(aboutToShowTextStyleMenu()));
+  connect(iLabelStyleMenu, SIGNAL(aboutToShow()),
+	  this, SLOT(aboutToShowLabelStyleMenu()));
   connect(iGridSizeMenu, SIGNAL(aboutToShow()),
 	  this, SLOT(aboutToShowGridSizeMenu()));
   connect(iAngleSizeMenu, SIGNAL(aboutToShow()),
@@ -463,6 +511,7 @@ AppUi::AppUi(lua_State *L0, int model, Qt::WindowFlags f)
 
   iMouse = new QLabel(statusBar());
   findAction("coordinates|points")->setChecked(true);
+  findAction("scaling|1")->setChecked(true);
   statusBar()->addPermanentWidget(iMouse, 0);
   iMouse->setFont(font);
 
@@ -479,13 +528,22 @@ AppUi::~AppUi()
 
 // --------------------------------------------------------------------
 
+static void populateLayerMenu(LayerBox *layerList, QMenu *menu)
+{
+  menu->clear();
+  for (int i = 0; i < layerList->count(); ++i) {
+    LayerItem *item = dynamic_cast<LayerItem *>(layerList->item(i));
+    if (item) {
+      QAction *a = new QAction(QIpe(item->ipeLayerName), menu);
+      menu->addAction(a);
+    }
+  }
+
+}
+
 void AppUi::aboutToShowSelectLayerMenu()
 {
-  iSelectLayerMenu->clear();
-  for (int i = 0; i < iLayerList->count(); ++i) {
-    QAction *a = new QAction(iLayerList->item(i)->text(), iSelectLayerMenu);
-    iSelectLayerMenu->addAction(a);
-  }
+  populateLayerMenu(iLayerList, iSelectLayerMenu);
 }
 
 void AppUi::selectLayerAction(QAction *a)
@@ -495,11 +553,7 @@ void AppUi::selectLayerAction(QAction *a)
 
 void AppUi::aboutToShowMoveToLayerMenu()
 {
-  iMoveToLayerMenu->clear();
-  for (int i = 0; i < iLayerList->count(); ++i) {
-    QAction *a = new QAction(iLayerList->item(i)->text(), iMoveToLayerMenu);
-    iMoveToLayerMenu->addAction(a);
-  }
+  populateLayerMenu(iLayerList, iMoveToLayerMenu);
 }
 
 void AppUi::moveToLayerAction(QAction *a)
@@ -507,21 +561,31 @@ void AppUi::moveToLayerAction(QAction *a)
   action(String("movetolayer-") + IpeQ(a->text()));
 }
 
-void AppUi::aboutToShowTextStyleMenu()
+void AppUi::aboutToShowStyleMenu(Kind kind, MENUHANDLE menu, String current)
 {
   AttributeSeq seq;
-  iCascade->allNames(ETextStyle, seq);
-  iTextStyleMenu->clear();
+  iCascade->allNames(kind, seq);
+  menu->clear();
   for (uint i = 0; i < seq.size(); ++i) {
     String s = seq[i].string();
-    QAction *a = new QAction(QIpe(s), iTextStyleMenu);
+    QAction *a = new QAction(QIpe(s), menu);
     a->setCheckable(true);
-    if (s == iAll.iTextStyle.string())
+    if (s == current)
       a->setChecked(true);
-    iTextStyleMenu->addAction(a);
+    menu->addAction(a);
   }
 }
 
+void AppUi::aboutToShowTextStyleMenu()
+{
+  aboutToShowStyleMenu(ETextStyle, iTextStyleMenu, iAll.iTextStyle.string());
+}
+
+void AppUi::aboutToShowLabelStyleMenu()
+{
+  aboutToShowStyleMenu(ELabelStyle, iLabelStyleMenu, iAll.iLabelStyle.string());
+}
+
 void AppUi::aboutToShowGridSizeMenu()
 {
   iGridSizeMenu->clear();
@@ -538,6 +602,14 @@ void AppUi::aboutToShowGridSizeMenu()
 void AppUi::aboutToShowAngleSizeMenu()
 {
   iAngleSizeMenu->clear();
+  for (auto &name : iComboContents[EUiAngleSize]) {
+    QString s = QIpe(name);
+    QAction *a = new QAction(s, iAngleSizeMenu);
+    a->setCheckable(true);
+    if (s == iSelector[EUiAngleSize]->currentText())
+      a->setChecked(true);
+    iAngleSizeMenu->addAction(a);
+  }
 }
 
 void AppUi::textStyleAction(QAction *a)
@@ -545,6 +617,11 @@ void AppUi::textStyleAction(QAction *a)
   action(String("textstyle|") + IpeQ(a->text()));
 }
 
+void AppUi::labelStyleAction(QAction *a)
+{
+  action(String("labelstyle|") + IpeQ(a->text()));
+}
+
 void AppUi::gridSizeAction(QAction *a)
 {
   action(String("gridsize|") + IpeQ(a->text()));
@@ -674,7 +751,7 @@ void AppUi::bookmarkSelected(QListWidgetItem *item)
 
 void AppUi::setToolVisible(int m, bool vis)
 {
-  QWidget *tool = 0;
+  QWidget *tool = nullptr;
   switch (m) {
   case 0: tool = iPropertiesTools; break;
   case 1: tool = iBookmarkTools; break;
@@ -785,38 +862,32 @@ void AppUi::comboSelector(int id)
 // --------------------------------------------------------------------
 
 static const char * const aboutText =
-"<qt><h2>Ipe %d.%d.%d</h2>"
-"<p>Copyright (c) 1993-2016 Otfried Cheong</p>"
-"<p>The extensible drawing editor Ipe creates figures "
-"in Postscript and PDF format, "
-"using LaTeX to format the text in the figures.</p>"
-"<p>Ipe relies on the following fine pieces of software:"
-"<ul>"
-"<li> Pdftex, Xetex, or Luatex"
-"<li> %s (%d kB used)" // Lua
-"<li> The font rendering library %s"
-"<li> The rendering library Cairo %s / %s"
-"<li> The GUI toolkit Qt %s / %s"
-"<li> The compression library zlib %s"
-"</ul>"
-"<p>Ipe is released under the GNU Public License.</p>"
-"<p>See <a href=\"http://ipe.otfried.org\">ipe.otfried.org</a>"
-" for details.</p>"
-"</qt>";
+  "<qt><h1>Ipe %d.%d.%d</h1>"
+  "<p>Copyright (c) 1993-%d Otfried Cheong</p>"
+  "<p>The extensible drawing editor Ipe creates figures "
+  "in Postscript and PDF format, "
+  "using LaTeX to format the text in the figures.</p>"
+  "<p>Ipe is released under the GNU Public License.</p>"
+  "<p>See the <a href=\"http://ipe.otfried.org\">Ipe homepage</a>"
+  " for further information.</p>"
+  "<p>You can \"like\" Ipe and follow Ipe announcements on "
+  "<a href=\"http://www.facebook.com/drawing.editor.Ipe7\">Facebook</a>.</p>"
+  "<h3>Platinum sponsors</h3>"
+  "<ul><li>Martin Ziegler</li></ul>"
+  "<p>If you enjoy Ipe, you can become a member of the exclusive community of "
+  "<a href=\"http://patreon.com/otfried\">Ipe patrons</a>. "
+  "For the price of a cup of coffee per month you can make a meaningful contribution "
+  "to the continuing development of Ipe.</p>"
+  "</qt>";
 
 void AppUi::aboutIpe()
 {
-  int luaSize = lua_gc(L, LUA_GCCOUNT, 0);
   std::vector<char> buf(strlen(aboutText) + 100);
   sprintf(&buf[0], aboutText,
 	  IPELIB_VERSION / 10000,
 	  (IPELIB_VERSION / 100) % 100,
 	  IPELIB_VERSION % 100,
-	  LUA_RELEASE, luaSize,
-	  Fonts::freetypeVersion().z(),
-	  CAIRO_VERSION_STRING, cairo_version_string(),
-	  QT_VERSION_STR, qVersion(),
-	  ZLIB_VERSION);
+	  COPYRIGHT_YEAR);
 
   QMessageBox msgBox(this);
   msgBox.setWindowTitle("About Ipe");
@@ -950,26 +1021,16 @@ static int getImage(lua_State *L, QImage &img)
   QImage im1 = img.convertToFormat(QImage::Format_ARGB32);
   int w = im1.width();
   int h = im1.height();
-  bool isGray = im1.allGray();
-
-  Bitmap::TColorSpace cs = isGray ? Bitmap::EDeviceGray : Bitmap::EDeviceRGB;
-  int datalen = w * h * (isGray ? 1 : 3);
-  Buffer data(datalen);
-  char *d = data.data();
+  Buffer data(w * h * sizeof(uint32_t));
+  uint32_t *d = (uint32_t *) data.data();
   for (int y = 0; y < h; ++y) {
-    uint *p = (uint *) im1.scanLine(y);
+    uint32_t *p = (uint32_t *) im1.scanLine(y);
     for (int x = 0; x < w; ++x) {
-      if (isGray) {
-	*d++ = qRed(*p++);
-      } else {
-	*d++ = qRed(*p);
-	*d++ = qGreen(*p);
-	*d++ = qBlue(*p++);
-      }
+      *d++ = *p++;
     }
   }
 
-  Bitmap bitmap(w, h, cs, 8, data, Bitmap::EDirect, true);
+  Bitmap bitmap(w, h, Bitmap::ENative, data);
   Rect r(Vector::ZERO, Vector(w, h));
   Image *im = new Image(r, bitmap);
   push_object(L, im);
