@@ -4,7 +4,7 @@
 --[[
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (C) 1993-2016  Otfried Cheong
+    Copyright (c) 1993-2019 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -59,6 +59,10 @@ end
 -- (only useful if you develop ipelets or want to customize Ipe)
 prefs.developer = false
 
+-- Should Ipe terminate when the last window was closed?
+-- (Currently only on OS X)
+prefs.terminate_on_close = true
+
 -- Enable continuous spell-checking in text fields by default?
 -- (Currently only on OS X)
 prefs.spell_check = true
@@ -89,16 +93,24 @@ else
   prefs.delete_key = "\127"
 end
 
--- Where the tools should be put (left or right)
+-- Where the tools should be placed (left or right)
 -- and which ones should be displayed on start-up
--- (left/right currently only used with Qt toolkit)
-prefs.tools = {
+-- tools_placement is only used with Qt toolkit
+prefs.tools_placement = {
   properties = "left",
   bookmarks = "right",
   notes = "right",
-  layers = "left",
-  hidden = { "bookmarks", "notes" }
+  layers = "left"
 }
+prefs.tools_visible = {
+  properties = true,
+  bookmarks = false,
+  notes = false,
+  layers = true
+}
+
+-- initial ordering of the toolbars (windows only)
+prefs.toolbar_order = { "snap", "grid", "angle", "edit", "mode" }
 
 -- Should LaTeX be run automatically every time text has changed?
 prefs.auto_run_latex = true
@@ -109,9 +121,19 @@ prefs.auto_external_editor = nil
 -- Should closing the external editor immediately close the editing dialog?
 prefs.editor_closes_dialog = nil
 
+-- If the user interface is too small, scale it here (Windows and Qt only)
+-- 100 is the original size, 150 is 50% larger
+-- On Windows 10, the system dpi is used automatically if left at 100.
+prefs.ui_scale = 100
+-- The following influences the size of the toolbar buttons
+prefs.toolbar_scale = 100
+
 -- Minimum size of dialog window for creating/editing text objects
 prefs.editor_size = { 0, 0 }
 
+-- Width of Notes and Bookmarks tools (Windows only)
+prefs.width_notes_bookmarks = 200
+
 -- Size of main window at startup
 prefs.window_size = { 1080, 600 }
 
@@ -174,17 +196,40 @@ prefs.close_distance = 48
 
 -- Zoom factors, minimal and maximal possible zoom
 prefs.zoom_factor = 1.3
--- when zooming by scroll wheel or
--- by scroll gesture on OS X
-if config.toolkit == "cocoa" then
-  prefs.wheel_zoom_factor = 1.1
-  prefs.wheel_other_factor = 1.3 -- used on some external mice
-else
-  prefs.wheel_zoom_factor = 1.3
-end
 prefs.min_zoom = 0.1
 prefs.max_zoom = 100
 
+----------------------------------------------------------------------
+
+-- The following settings determine what the scroll wheel on a mouse or a
+-- two-finger pan gesture on a touchpad/trackpad  do.
+
+-- There are two sets of settings: "trackpad_scroll" applies to trackpads under OSX,
+-- "scroll" applies in all other cases (including external mice in OSX).
+
+-- When pan is true, the gesture moves (pans) the canvas around, and
+-- you can control the speed of panning with pixel_per_degree.
+-- The direction of the movement is determined by direction.x and direction.y (1 or -1).
+-- When pan is false, the vertical scroll movement is used to change the zoom level.
+-- You can control the zoom speed using zoom_per_degree.
+-- If you want to reverse the direction of zooming, change
+-- zoom_per_degree to something smaller than one, e.g. 0.9827
+
+prefs.scroll = { pan = true,
+	         pixel_per_degree = 2,
+		 zoom_per_degree = 1.0176,
+		 direction = { x=1, y=1 } }
+prefs.trackpad_scroll = { pan = true,
+	              	  pixel_per_degree = 2,
+		 	  zoom_per_degree = 1.0176,
+			  direction = { x=1, y=1 } }
+
+----------------------------------------------------------------------
+
+-- How close do two vertices need to be to be considered the same
+-- in the "Join paths" operation?
+prefs.join_threshold = 1e-6
+
 -- Should newly created text be transformable by default?
 prefs.text_transformable = false
 
@@ -195,20 +240,14 @@ prefs.automatic_use_title = false
 -- How to start browser to show Ipe manual:
 -- (not used with Windows or Cocoa toolkit)
 
--- On Linux, there are various options depending on your Desktop:
+prefs.browser = "xdg-open %s"
+-- On Linux, there are some other options you can try if xdg-open
+-- doesn't work on your desktop:
 -- Gnome 2: gnome-open
--- prefs.browser = "gnome-open %s"
 -- Gnome 3: gvfs-open
-prefs.browser = "gvfs-open %s"
 -- KDE: kde-open
--- prefs.browser = "kde-open %s"
 -- XFCE: exo-open
--- prefs.browser = "exo-open %s"
--- "sensible-browser" starts a new browser
-  local engines = { "default", "pdftex", "xetex" } -- "luatex" }
-  local engines = { "default", "pdftex", "xetex" } -- "luatex" }
--- (which is not nice, ideally we'd like a tab in an open browser window)
--- prefs.browser = "sensible-browser %s &"
+-- Ubuntu/Debian: sensible-browser
 
 -- How to start onscreen keyboard
 if config.platform == "win" then
@@ -220,10 +259,67 @@ else
   prefs.keyboard = nil
 end
 
--- Do not highlight current selection in "ink" drawing mode?
-prefs.no_ink_highlight = false
 -- Disable all snap modes in "ink" drawing mode?
-prefs.no_ink_snap = false
+prefs.no_ink_snap = true
+
+----------------------------------------------------------------------
+--
+-- The following settings can be used to tune the use of the pen in
+-- ink mode.  It seems digitizers differ a lot, here you can adapt Ipe
+-- to yours.
+--
+-- If your digitizer generates lots of pen move events, set
+-- ink_min_movement to avoid making thousands of tiny ink segments
+-- (start with a value like 0.2).
+--
+-- If there is a lot of noise in pen recognition, try setting
+-- ink_smoothing to a positive value k (smoothing is then performed
+-- over 2k+1 samples).  Try 3, for instance.
+--
+-- After a complete ink stroke has been drawn, Ipe simplifies it to
+-- remove excess vertices, by computing a path with the minimal number
+-- of vertices that differs by at most ink_tolerance.
+--
+-- Ink strokes that are very short (like a dot on an i) tend to appear
+-- too small. Use ink_dot_length and ink_dot_wider to enlarge them.
+--
+-- Finally, you can set ink_spline to have your ink strokes converted
+-- to a spline (rather than a polygonal curve).
+--
+----------------------------------------------------------------------
+
+-- Minimal distance for movement to update ink stroke
+prefs.ink_min_movement = 0
+
+-- To smooth ink strokes, perform sliding average over how many samples?
+prefs.ink_smoothing = 0
+
+-- When ink stroke is simplified, what is the tolerance?
+prefs.ink_tolerance = 0.3
+
+-- Threshold for considering an ink stroke a tap with little movement:
+prefs.ink_dot_length = 10.0
+
+-- When user taps the screen in ink mode with no or little movement,
+-- how much should the pen width be enlarged?
+-- nil means use current pen width
+prefs.ink_dot_wider = 2.0
+
+-- Turn ink strokes into splines?
+prefs.ink_spline = false
+
+----------------------------------------------------------------------
+
+-- Auto-exporting when document is being saved
+-- If this is set, then every time you save in xml format, say to
+-- mydoc.ipe, the same document is EXPORTED to mydoc.pdf.
+-- Careful: if mydoc.pdf exists, it is silently overwritten.
+-- mydoc.pdf cannot be opened by Ipe (it is an exported copy).
+
+prefs.auto_export_to_pdf = false
+
+----------------------------------------------------------------------
+
 -- Extended properties menu, perhaps useful for tablets:
 prefs.tablet_menu = false
 
