@@ -4,7 +4,7 @@
 /*
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (C) 1993-2016  Otfried Cheong
+    Copyright (c) 1993-2019 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -34,20 +34,10 @@
 
 using namespace ipe;
 
-extern bool dctDecode(Buffer dctData, Buffer pixelData, int components);
+extern bool dctDecode(Buffer dctData, Buffer pixelData);
 
 // --------------------------------------------------------------------
 
-/*! \class ipe::Bitmap::MRenderData
-  \ingroup base
-  \brief Abstract base class for pixmap data stored by a client.
-*/
-
-Bitmap::MRenderData::~MRenderData()
-{
-  // nothing
-}
-
 /*! \class ipe::Bitmap
   \ingroup base
   \brief A bitmap.
@@ -55,10 +45,7 @@ Bitmap::MRenderData::~MRenderData()
   Bitmaps are explicitely shared using reference-counting.  Copying is
   cheap, so Bitmap objects are meant to be passed by value.
 
-  The bitmap can cache data to speed up rendering. This data can be
-  set only once (as the bitmap is conceptually immutable).
-
-  The bitmap also provides a slot for short-term storage of an "object
+  The bitmap provides a slot for short-term storage of an "object
   number".  The PDF embedder, for instance, sets it to the PDF object
   number when embedding the bitmap, and can reuse it when "drawing"
   the bitmap.
@@ -67,123 +54,224 @@ Bitmap::MRenderData::~MRenderData()
 //! Default constructor constructs null bitmap.
 Bitmap::Bitmap()
 {
-  iImp = 0;
+  iImp = nullptr;
 }
 
 //! Create from XML stream.
-Bitmap::Bitmap(const XmlAttributes &attr, String data)
+Bitmap::Bitmap(const XmlAttributes &attr, String pcdata)
 {
-  int length = init(attr);
+  auto lengths = init(attr);
+  int length = lengths.first;
+  if (length == 0)
+    length = height() * width() * (isGray() ? 1 : 3);
+  int alphaLength = lengths.second;
   // decode data
   iImp->iData = Buffer(length);
   char *p = iImp->iData.data();
+  Buffer alpha;
+  char *q = nullptr;
+  if (alphaLength > 0) {
+    alpha = Buffer(alphaLength);
+    q = alpha.data();
+  }
   if (attr["encoding"] ==  "base64") {
-    Buffer dbuffer(data.data(), data.size());
+    Buffer dbuffer(pcdata.data(), pcdata.size());
     BufferSource source(dbuffer);
     Base64Source b64source(source);
     while (length-- > 0)
       *p++ = b64source.getChar();
+    while (alphaLength-- > 0)
+      *q++ = b64source.getChar();
   } else {
-    Lex datalex(data);
+    Lex datalex(pcdata);
     while (length-- > 0)
       *p++ = char(datalex.getHexByte());
+    while (alphaLength-- > 0)
+      *q++ = char(datalex.getHexByte());
   }
+  unpack(alpha);
   computeChecksum();
+  analyze();
 }
 
 //! Create from XML using external raw data
-Bitmap::Bitmap(const XmlAttributes &attr, Buffer data)
+Bitmap::Bitmap(const XmlAttributes &attr, Buffer data, Buffer alpha)
 {
-  int length = init(attr);
-  assert(length == data.size());
+  init(attr);
   iImp->iData = data;
+  unpack(alpha);
   computeChecksum();
+  analyze();
 }
 
-int Bitmap::init(const XmlAttributes &attr)
+std::pair<int, int> Bitmap::init(const XmlAttributes &attr)
 {
   iImp = new Imp;
   iImp->iRefCount = 1;
+  iImp->iFlags = 0;
+  iImp->iColorKey = -1;
+  iImp->iPixelsComputed = false;
   iImp->iObjNum = Lex(attr["id"]).getInt();
-  iImp->iRender = 0;
   iImp->iWidth = Lex(attr["width"]).getInt();
   iImp->iHeight = Lex(attr["height"]).getInt();
-  iImp->iColorKey = -1;
   int length = Lex(attr["length"]).getInt();
+  int alphaLength = Lex(attr["alphaLength"]).getInt();
   assert(iImp->iWidth > 0 && iImp->iHeight > 0);
   String cs = attr["ColorSpace"];
-  if (cs == "DeviceGray") {
-    iImp->iComponents = 1;
-    iImp->iColorSpace = EDeviceGray;
-  } else if (cs == "DeviceCMYK") {
-    iImp->iComponents = 4;
-    iImp->iColorSpace = EDeviceCMYK;
-  } else {
-    iImp->iComponents = 3;
-    iImp->iColorSpace = EDeviceRGB;
-  }
-  String cc;
-  if (iImp->iColorSpace == EDeviceRGB && attr.has("ColorKey", cc)) {
-    iImp->iColorKey = Lex(cc).getHexNumber();
+  if (cs.right(5) =="Alpha") {
+    iImp->iFlags |= EAlpha;
+    cs = cs.left(cs.size() - 5);
   }
+  if (cs == "DeviceRGB")
+    iImp->iFlags |= ERGB;
   String fi = attr["Filter"];
   if (fi == "DCTDecode")
-    iImp->iFilter = EDCTDecode;
+    iImp->iFlags |= EDCT;
   else if (fi == "FlateDecode")
-    iImp->iFilter = EFlateDecode;
-  else
-    iImp->iFilter = EDirect;
-  iImp->iBitsPerComponent = Lex(attr["BitsPerComponent"]).getInt();
-  if (length == 0) {
-    assert(iImp->iFilter == EDirect);
-    int bitsPerRow = width() * components() * bitsPerComponent();
-    int bytesPerRow = (bitsPerRow + 7) / 8;
-    length = height() * bytesPerRow;
-  }
-  return length;
+    iImp->iFlags |= EInflate;
+  String cc;
+  if (!isJpeg() && attr.has("ColorKey", cc))
+    iImp->iColorKey = Lex(cc).getHexNumber();
+  return std::make_pair(length, alphaLength);
 }
 
-//! Create a new image
-/*! \a filter specifies the format of the \a data.
-  If \a filter is EDirect, then setting \a deflate
-  compresses the bitmap and changes the filter to EFlateDecode.
- */
-Bitmap::Bitmap(int width, int height,
-	       TColorSpace colorSpace, int bitsPerComponent,
-	       Buffer data, TFilter filter, bool deflate)
+//! Create a new image from given image data.
+/*! If you already have data in native-endian ARGB32 without premultiplication,
+  pass it with flag ENative.
+  Otherwise pass a byte stream and set ERGB and EAlpha correctly:
+  EAlpha: each pixel starts with one byte of alpha channel,
+  ERGB: each pixel has three bytes of R, G, B, in this order,
+  otherwise each pixel has one byte of gray value. */
+Bitmap::Bitmap(int width, int height, uint32_t flags, Buffer data)
 {
   iImp = new Imp;
   iImp->iRefCount = 1;
+  iImp->iFlags = flags;
+  iImp->iColorKey = -1;
   iImp->iObjNum = -1;
-  iImp->iRender = 0;
   iImp->iWidth = width;
   iImp->iHeight = height;
-  iImp->iColorKey = -1;
+  iImp->iData = data;
+  iImp->iPixelsComputed = false;
   assert(iImp->iWidth > 0 && iImp->iHeight > 0);
-  iImp->iColorSpace = colorSpace;
-  switch (colorSpace) {
-  case EDeviceGray:
-    iImp->iComponents = 1;
-    break;
-  case EDeviceRGB:
-    iImp->iComponents = 3;
-    break;
-  case EDeviceCMYK:
-    iImp->iComponents = 4;
-    break;
-  }
-  iImp->iFilter = filter;
-  iImp->iBitsPerComponent = bitsPerComponent;
-  if (deflate && filter == EDirect) {
-    // optional deflation
-    int deflatedSize;
-    Buffer deflated = DeflateStream::deflate(data.data(), data.size(),
-					     deflatedSize, 9);
-    iImp->iData = Buffer(deflated.data(), deflatedSize);
-    iImp->iFilter = EFlateDecode;
-  } else
-    iImp->iData = data;
+  unpack(Buffer());
   computeChecksum();
+  analyze();
+}
+
+//! Take care of inflating, converting grayscale to rgb, and merging the alpha channel
+void Bitmap::unpack(Buffer alphaChannel)
+{
+  if (isJpeg() || iImp->iFlags & ENative)
+    return;
+  int npixels = width() * height();
+  if (iImp->iFlags & EInflate) {
+    // inflate data
+    int components = isGray() ? 1 : 3;
+    if (hasAlpha() && alphaChannel.size() == 0)
+      components += 1;
+    uLongf inflatedSize = npixels * components;
+    Buffer inflated(inflatedSize);
+    assert(uncompress((Bytef *) inflated.data(), &inflatedSize,
+		      (const Bytef *) iImp->iData.data(), iImp->iData.size()) == Z_OK);
+    iImp->iData = inflated;
+    if (alphaChannel.size() > 0) {
+      inflatedSize = npixels;
+      Buffer inflatedAlpha(inflatedSize);
+      assert(uncompress((Bytef *) inflatedAlpha.data(), &inflatedSize,
+			(const Bytef *) alphaChannel.data(), alphaChannel.size()) == Z_OK);
+      alphaChannel = inflatedAlpha;
+    }
+  }
+  // convert data to ARGB32 format
+  bool alphaInMain = hasAlpha() && alphaChannel.size() == 0;
+  Buffer pixels(npixels * sizeof(uint32_t));
+  const char *p = iImp->iData.data();
+  uint32_t *q = (uint32_t *) pixels.data();
+  uint32_t *fin = q + npixels;
+  if (!isGray()) {
+    while (q < fin) {
+      uint8_t alpha = (alphaInMain ? uint8_t(*p++) : 0xff);
+      uint8_t r = uint8_t(*p++);
+      uint8_t g = uint8_t(*p++);
+      uint8_t b = uint8_t(*p++);
+      uint32_t pixel = (alpha << 24) | (r << 16) | (g << 8) | b;
+      *q++ = pixel;
+    }
+  } else {
+    while (q < fin) {
+      uint8_t alpha = (alphaInMain ? uint8_t(*p++) : 0xff);
+      uint8_t r = uint8_t(*p++);
+      *q++ = (alpha << 24) | (r << 16) | (r << 8) | r;
+    }
+  }
+  // merge separate alpha channel
+  if (hasAlpha() && alphaChannel.size() > 0) {
+    q = (uint32_t *) pixels.data();
+    p = alphaChannel.data();
+    uint32_t pixel;
+    while (q < fin) {
+      pixel = *q;
+      pixel = (pixel & 0x00ffffff) | (*p++ << 24);
+      *q++ = pixel;
+    }
+  }
+  if (iImp->iColorKey >= 0) {
+    uint32_t colorKey = (iImp->iColorKey | 0xff000000);
+    q = (uint32_t *) pixels.data();
+    while (q < fin) {
+      if (*q == colorKey)
+	*q = iImp->iColorKey;
+      ++q;
+    }
+  }
+  iImp->iData = pixels;
+}
+
+//! Determine if bitmap has alpha channel, colorkey, rgb values (does nothing for JPG).
+void Bitmap::analyze()
+{
+  iImp->iColorKey = -1;
+  iImp->iFlags &= EDCT|ERGB;  // clear all other flags, we recompute them
+  if (isJpeg())
+    return;
+  iImp->iFlags &= EDCT;  // ERGB will also be recomputed
+  const uint32_t *q = (const uint32_t *) iImp->iData.data();
+  const uint32_t *fin = q + width() * height();
+  uint32_t pixel, gray;
+  while (q < fin) {
+    pixel = *q++ & 0x00ffffff;
+    gray = (pixel & 0xff);
+    gray |=  (gray << 8) | (gray << 16);
+    if (pixel != gray) {
+      iImp->iFlags |= ERGB;
+      break;
+    }
+  }
+  int candidate = -1, color;
+  uint32_t alpha;
+  q = (const uint32_t *) iImp->iData.data();
+  while (q < fin) {
+    pixel = *q++;
+    alpha = pixel & 0xff000000;
+    color = pixel & 0x00ffffff;
+    if (alpha != 0 && alpha != 0xff000000) {
+      iImp->iFlags |= EAlpha;
+      return;
+    }
+    if (alpha == 0) { // transparent color found
+      if (candidate < 0)
+	candidate = color;
+      else if (candidate != color) { // two different transparent colors
+	iImp->iFlags |= EAlpha;
+	return;
+      }
+    } else if (color == candidate) {   // opaque copy of candidate found
+      iImp->iFlags |= EAlpha;
+      return;
+    }
+  }
+  iImp->iColorKey = candidate;
 }
 
 //! Copy constructor.
@@ -199,7 +287,6 @@ Bitmap::Bitmap(const Bitmap &rhs)
 Bitmap::~Bitmap()
 {
   if (iImp && --iImp->iRefCount == 0) {
-    delete iImp->iRender;
     delete iImp;
   }
 }
@@ -218,81 +305,59 @@ Bitmap &Bitmap::operator=(const Bitmap &rhs)
   return *this;
 }
 
-//! Return rgb representation of the transparent color.
-/*! Returns -1 if the bitmap is not color keyed. */
-int Bitmap::colorKey() const
-{
-  return iImp->iColorKey;
-}
-
-//! Set transparent color.
-/*! Use \a key == -1 to disable color key. */
-void Bitmap::setColorKey(int key)
-{
-  iImp->iColorKey = key;
-}
-
 //! Save bitmap in XML stream.
 void Bitmap::saveAsXml(Stream &stream, int id, int pdfObjNum) const
 {
   assert(iImp);
-  stream << "<bitmap";
-  stream << " id=\"" << id << "\"";
-  stream << " width=\"" << width() << "\"";
-  stream << " height=\"" << height() << "\"";
-  stream << " length=\"" << size() << "\"";
-  switch (colorSpace()) {
-  case EDeviceGray:
-    stream << " ColorSpace=\"DeviceGray\"";
-    break;
-  case EDeviceRGB:
-    stream << " ColorSpace=\"DeviceRGB\"";
-    break;
-  case EDeviceCMYK:
-    stream << " ColorSpace=\"DeviceCMYK\"";
-    break;
-  }
-  switch (filter()) {
-  case EFlateDecode:
-    stream << " Filter=\"FlateDecode\"";
-    break;
-  case EDCTDecode:
-    stream << " Filter=\"DCTDecode\"";
-    break;
-  default:
-    // no filter
-    break;
-  }
-  stream << " BitsPerComponent=\"" << bitsPerComponent() << "\"";
+  stream << "<bitmap"
+	 << " id=\"" << id << "\""
+	 << " width=\"" << width() << "\""
+	 << " height=\"" << height() << "\""
+	 << " BitsPerComponent=\"8\""; // no longer used but required by earlier Ipe versions
 
-  if (iImp->iColorKey >= 0) {
+  stream << " ColorSpace=\"Device";
+  if (isGray())
+    stream << "Gray";
+  else
+    stream << "RGB";
+  if (hasAlpha())
+    stream << "Alpha";
+  stream << "\"";
+  if (isJpeg())
+    stream << " Filter=\"DCTDecode\"";
+  else
+    stream << " Filter=\"FlateDecode\"";
+
+  if (colorKey() >= 0) {
     char buf[10];
-    sprintf(buf, "%x", iImp->iColorKey);
+    sprintf(buf, "%x", colorKey());
     stream << " ColorKey=\"" << buf << "\"";
   }
 
   if (pdfObjNum >= 0) {
-    stream << " pdfObject=\"" << pdfObjNum << "\"/>\n";
+    stream << " pdfObject=\"" << pdfObjNum;
+    if (hasAlpha())
+      stream << " " << pdfObjNum-1;
+    stream << "\"/>\n";
   } else {
     // save data
+    auto data = embed();
+    stream << " length=\"" << data.first.size() << "\"";
+    if (hasAlpha())
+      stream << " alphaLength=\"" << data.second.size() << "\"";
     stream << " encoding=\"base64\">\n";
-    const char *p = data();
-    const char *fin = p + size();
     Base64Stream b64(stream);
-    while (p != fin)
-      b64.putChar(*p++);
+    for (const auto & buffer : { data.first, data.second } ) {
+      const char *p = buffer.data();
+      const char *fin = p + buffer.size();
+      while (p != fin)
+	b64.putChar(*p++);
+    }
     b64.close();
     stream << "</bitmap>\n";
   }
 }
 
-//! Set a cached bitmap for fast rendering.
-void Bitmap::setRenderData(MRenderData *data) const
-{
-  assert(iImp && iImp->iRender == 0);
-  iImp->iRender = data;
-}
-
 bool Bitmap::equal(Bitmap rhs) const
 {
   if (iImp == rhs.iImp)
@@ -300,13 +365,9 @@ bool Bitmap::equal(Bitmap rhs) const
   if (!iImp || !rhs.iImp)
     return false;
 
-  if (iImp->iColorSpace != rhs.iImp->iColorSpace ||
-      iImp->iBitsPerComponent != rhs.iImp->iBitsPerComponent ||
+  if (iImp->iFlags != rhs.iImp->iFlags ||
       iImp->iWidth != rhs.iImp->iWidth ||
       iImp->iHeight != rhs.iImp->iHeight ||
-      iImp->iComponents != rhs.iImp->iComponents ||
-      iImp->iColorKey != rhs.iImp->iColorKey ||
-      iImp->iFilter != rhs.iImp->iFilter ||
       iImp->iChecksum != rhs.iImp->iChecksum ||
       iImp->iData.size() != rhs.iImp->iData.size())
     return false;
@@ -333,63 +394,87 @@ void Bitmap::computeChecksum()
   iImp->iChecksum = s;
 }
 
+//! Create the data to be embedded in an XML or PDF file.
+/*! For Jpeg images, this is simply the bitmap data.  For other
+  images, rgb/grayscale data and alpha channel are split and deflated
+  separately. */
+std::pair<Buffer, Buffer> Bitmap::embed() const
+{
+  if (isJpeg())
+    return std::make_pair(iImp->iData, Buffer());
+  int npixels = width() * height();
+  uint32_t *src = (uint32_t *) iImp->iData.data();
+  uint32_t *fin = src + npixels;
+  uint32_t pixel;
+  Buffer rgb(npixels * (isGray() ? 1 : 3));
+  char *p = rgb.data();
+  while (src < fin) {
+    pixel = *src++;
+    if (isGray()) {
+      *p++ = pixel & 0xff;
+    } else {
+      *p++ = (pixel & 0xff0000) >> 16;
+      *p++ = (pixel & 0x00ff00) >> 8;
+      *p++ = (pixel & 0x0000ff);
+    }
+  }
+  int deflatedSize;
+  Buffer deflated = DeflateStream::deflate(rgb.data(), rgb.size(), deflatedSize, 9);
+  rgb = Buffer(deflated.data(), deflatedSize);
+  Buffer alpha;
+  if (hasAlpha()) {
+    alpha = Buffer(npixels);
+    src = (uint32_t *) iImp->iData.data();
+    p = alpha.data();
+    while (src < fin)
+      *p++ = (*src++ & 0xff000000) >> 24;
+    deflated = DeflateStream::deflate(alpha.data(), alpha.size(), deflatedSize, 9);
+    alpha = Buffer(deflated.data(), deflatedSize);
+  }
+  return std::make_pair(rgb, alpha);
+}
+
 // --------------------------------------------------------------------
 
-//! Convert bitmap data to a height x width pixel array in rgb format.
+//! Return pixels for rendering.
 /*! Returns empty buffer if it cannot decode the bitmap information.
-  Otherwise, returns a buffer of size width() * height() uint's. */
-Buffer Bitmap::pixelData() const
+  Otherwise, returns a buffer of size width() * height() uint32_t's.
+  The data is in cairo ARGB32 format, that is native-endian uint32_t's
+  with premultiplied alpha.
+*/
+Buffer Bitmap::pixelData()
 {
-  ipeDebug("pixelData %d x %d x %d, %d", width(), height(),
-	   components(), int(filter()));
-  if (bitsPerComponent() != 8)
-    return Buffer();
-  Buffer stream = iImp->iData;
-  Buffer pixels;
-  if (filter() == EDirect) {
-    pixels = stream;
-  } else if (filter() == EFlateDecode) {
-    // inflate data
-    uLong inflatedSize = width() * height() * components();
-    pixels = Buffer(inflatedSize);
-    if (uncompress((Bytef *) pixels.data(), &inflatedSize,
-		   (const Bytef *) stream.data(), stream.size()) != Z_OK
-	|| pixels.size() != int(inflatedSize))
-      return Buffer();
-  } else if (filter() == EDCTDecode) {
-    pixels = Buffer(width() * height() * components());
-    if (!dctDecode(stream, pixels, components()))
-      return Buffer();
-  }
-  Buffer data(height() * width() * sizeof(uint));
-  // convert pixels to data
-  const char *p = pixels.data();
-  uint *q = (uint *) data.data();
-  if (components() == 3) {
-    uint colorKey = (iImp->iColorKey | 0xff000000);
-    if (iImp->iColorKey < 0)
-      colorKey = 0;
-    for (int y = 0; y < iImp->iHeight; ++y) {
-      for (int x = 0; x < iImp->iWidth; ++x) {
-	uchar r = uchar(*p++);
-	uchar g = uchar(*p++);
-	uchar b = uchar(*p++);
-	uint pixel = 0xff000000 | (r << 16) | (g << 8) | b;
-	if (pixel == colorKey)
-	  *q++ = 0;
-	else
-	  *q++ = pixel;
-      }
-    }
-  } else if (components() == 1) {
-    for (int y = 0; y < iImp->iHeight; ++y) {
-      for (int x = 0; x < iImp->iWidth; ++x) {
-	uchar r = uchar(*p++);
-	*q++ = 0xff000000 | (r << 16) | (r << 8) | r;
-      }
+  if (!iImp->iPixelsComputed) {
+    iImp->iPixelsComputed = true;
+    if (isJpeg()) {
+      Buffer stream = iImp->iData;
+      Buffer pixels;
+      pixels = Buffer(4 * width() * height());
+      if (!dctDecode(stream, pixels))
+	return Buffer();
+      iImp->iPixelData = pixels;
+    } else {
+      if (hasAlpha() || colorKey() >= 0) {
+	// premultiply RGB data
+	iImp->iPixelData = Buffer(iImp->iData.size());
+	uint32_t *p = (uint32_t *) iImp->iData.data();
+	uint32_t *q = (uint32_t *) iImp->iPixelData.data();
+	uint32_t *fin = p + width() * height();
+	uint32_t pixel, alpha, alphaM, r, g, b;
+	while (p < fin) {
+	  pixel = *p++;
+	  alpha = (pixel & 0xff000000);
+	  alphaM = alpha >> 24;
+	  r = alphaM * (pixel & 0xff0000) / 255;
+	  g = alphaM * (pixel & 0x00ff00) / 255;
+	  b = alphaM * (pixel & 0x0000ff) / 255;
+	  *q++ = alpha | (r & 0xff0000) | (g & 0x00ff00) | (b & 0x0000ff);
+	}
+      } else
+	iImp->iPixelData = iImp->iData;
     }
   }
-  return data;
+  return iImp->iPixelData;
 }
 
 // --------------------------------------------------------------------
@@ -477,23 +562,24 @@ enum JPEG_MARKER {      /* JPEG marker codes                    */
 
 inline int read2bytes(FILE *f)
 {
-  uchar c1 = fgetc(f);
-  uchar c2 = fgetc(f);
+  uint8_t c1 = fgetc(f);
+  uint8_t c2 = fgetc(f);
   return (c1 << 8) + c2;
 }
 
 // --------------------------------------------------------------------
 
 //! Read information about JPEG image from file.
-/*! Returns NULL on success, an error message otherwise. */
+/*! Returns NULL on success, an error message otherwise.  Sets flags
+  to EDCT and possibly ERGB. */
 const char *Bitmap::readJpegInfo(FILE *file, int &width, int &height,
-				 Vector &dotsPerInch,
-				 TColorSpace &colorSpace,
-				 int &bitsPerComponent)
+				 Vector &dotsPerInch, uint32_t &flags)
 {
   static char jpg_id[] = "JFIF";
+  bool app0_seen = false;
 
   dotsPerInch = Vector(0, 0);
+  flags = EDCT;
 
   if (read2bytes(file) != 0xFFD8) {
     return "The file does not appear to be a JPEG image";
@@ -503,9 +589,13 @@ const char *Bitmap::readJpegInfo(FILE *file, int &width, int &height,
     int ch = fgetc(file);
     if (ch != 0xff)
       return "Reading JPEG image failed";
-    ch = fgetc(file);
+    do {
+      ch = fgetc(file);
+    } while (ch == 0xff);
+    ipeDebug("JPEG tag %x", ch & 0xff);
     int fpos = ftell(file);
     switch (ch & 0xff) {
+
     case M_SOF5:
     case M_SOF6:
     case M_SOF7:
@@ -516,34 +606,36 @@ const char *Bitmap::readJpegInfo(FILE *file, int &width, int &height,
     case M_SOF14:
     case M_SOF15:
       return "Unsupported type of JPEG compression";
-    case M_SOF2:
-      // cannot use progressive DCT with PDF-1.2
+
     case M_SOF0:
     case M_SOF1:
+    case M_SOF2: // progressive DCT allowed since PDF-1.3
     case M_SOF3:
       read2bytes(file);    /* read segment length  */
       ch = fgetc(file);
-      bitsPerComponent = ch;
+      if (ch != 8)
+	return "Unsupported bit width of pixels in JPEG image";
       height = read2bytes(file);
       width = read2bytes(file);
       ch = fgetc(file);
       switch (ch & 0xff) {
       case JPG_GRAY:
-	colorSpace = Bitmap::EDeviceGray;
         break;
       case JPG_RGB:
-	colorSpace = Bitmap::EDeviceRGB;
+	flags |= ERGB;
         break;
-      case JPG_CMYK:
-	colorSpace = Bitmap::EDeviceCMYK;
-	break;
       default:
 	return "Unsupported color space in JPEG image";
       }
       fseek(file, 0, SEEK_SET);
-      return 0;      //  success!
+      return nullptr;      //  success!
+
     case M_APP0: {
       int len = read2bytes(file);
+      if (app0_seen) {
+	fseek(file, fpos + len, SEEK_SET);
+	break;
+      }
       for (int i = 0; i < 5; i++) {
 	ch = fgetc(file);
 	if (ch != jpg_id[i]) {
@@ -566,8 +658,10 @@ const char *Bitmap::readJpegInfo(FILE *file, int &width, int &height,
 	  break;
 	}
       }
+      app0_seen = true;
       fseek(file, fpos + len, SEEK_SET);
       break; }
+
     case M_SOI:      // ignore markers without parameters
     case M_EOI:
     case M_TEM:
@@ -579,8 +673,8 @@ const char *Bitmap::readJpegInfo(FILE *file, int &width, int &height,
     case M_RST5:
     case M_RST6:
     case M_RST7:
-    case 0xff:       // What is this?
       break;
+
     default:         // skip variable length markers
 	fseek(file, fpos + read2bytes(file), SEEK_SET);
       break;
@@ -595,8 +689,7 @@ const char *Bitmap::readJpegInfo(FILE *file, int &width, int &height,
   If reading the file fails, returns a null Bitmap,
   and sets the error message \a errmsg.
 */
-Bitmap Bitmap::readJpeg(const char *fname, Vector &dotsPerInch,
-			const char * &errmsg)
+Bitmap Bitmap::readJpeg(const char *fname, Vector &dotsPerInch, const char * &errmsg)
 {
   FILE *file = Platform::fopen(fname, "rb");
   if (!file) {
@@ -604,19 +697,17 @@ Bitmap Bitmap::readJpeg(const char *fname, Vector &dotsPerInch,
     return Bitmap();
   }
 
-  int width, height, bitsPerComponent;
-  TColorSpace colorSpace;
+  int width, height;
+  uint32_t flags;
+
+  errmsg = Bitmap::readJpegInfo(file, width, height, dotsPerInch, flags);
 
-  errmsg = Bitmap::readJpegInfo(file, width, height, dotsPerInch,
-				colorSpace, bitsPerComponent);
   fclose(file);
   if (errmsg)
     return Bitmap();
 
   String a = Platform::readFile(fname);
-  return  Bitmap(width, height, colorSpace,
-		 bitsPerComponent, Buffer(a.data(), a.size()),
-		 Bitmap::EDCTDecode);
+  return  Bitmap(width, height, flags, Buffer(a.data(), a.size()));
 }
 
 // --------------------------------------------------------------------
