@@ -4,7 +4,7 @@
 /*
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (C) 1993-2016  Otfried Cheong
+    Copyright (c) 1993-2019 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -176,9 +176,13 @@ bool ImlParser::parseBitmap()
     return false;
   String objNumStr;
   if (att.slash() && att.has("pdfObject", objNumStr)) {
-    int objNum = Lex(objNumStr).getInt();
-    // ipeDebug("ParseBitmap: objNum = %d", objNum);
-    Bitmap bitmap(att, pdfStream(objNum));
+    Lex lex(objNumStr);
+    Buffer data = pdfStream(lex.getInt());
+    Buffer alpha;
+    lex.skipWhitespace();
+    if (!lex.eos())
+      alpha = pdfStream(lex.getInt());
+    Bitmap bitmap(att, data, alpha);
     iBitmaps.push_back(bitmap);
   } else {
     String bits;
@@ -220,9 +224,9 @@ bool ImlParser::parsePage(Page &page)
   if (tag == "notes") {
     XmlAttributes att;
     if (!parseAttributes(att))
-      return ESyntaxError;
+      return false;
     if (!parsePCDATA("notes", str))
-      return ESyntaxError;
+      return false;
     page.setNotes(str);
 
     tag = parseToTag();
@@ -292,6 +296,8 @@ bool ImlParser::parsePage(Page &page)
     if (tag == "/page") {
       return true;
     }
+    if (tag.empty())
+      return false;
     Object *obj = parseObject(tag, &page, &currentLayer);
     if (!obj)
       return false;
@@ -305,28 +311,28 @@ Page *ImlParser::parsePageSelection()
 {
   String tag = parseToTag();
   if (tag != "ipepage")
-    return 0;
+    return nullptr;
   XmlAttributes attr;
   if (!parseAttributes(attr))
-    return 0;
+    return nullptr;
   tag = parseToTag();
 
   while (tag == "bitmap") {
     if (!parseBitmap())
-      return 0;
+      return nullptr;
     tag = parseToTag();
   }
 
   if (tag != "page")
-    return 0;
+    return nullptr;
 
   std::unique_ptr<Page> page(new Page);
   if (!parsePage(*page))
-    return 0;
+    return nullptr;
 
   tag = parseToTag();
   if (tag != "/ipepage")
-    return 0;
+    return nullptr;
   return page.release();
 }
 
@@ -341,7 +347,7 @@ bool ImlParser::parseSelection(PageObjectSeq &seq)
     return false;
   XmlAttributes attr;
   if (!parseAttributes(attr))
-    return 0;
+    return nullptr;
   tag = parseToTag();
 
   while (tag == "bitmap") {
@@ -368,11 +374,11 @@ Object *ImlParser::parseObject(String tag, Page *page,
 			       int *currentLayer)
 {
   if (tag[0] == '/')
-    return 0;
+    return nullptr;
 
   XmlAttributes attr;
   if (!parseAttributes(attr))
-    return 0;
+    return nullptr;
 
   String layer;
   if (page && currentLayer && attr.has("layer", layer)) {
@@ -393,14 +399,14 @@ Object *ImlParser::parseObject(String tag, Page *page,
       }
       Object *obj = parseObject(tag);
       if (!obj)
-	return 0;
+	return nullptr;
       group.push_back(obj);
     }
   }
 
   String pcdata;
   if (!attr.slash() && !parsePCDATA(tag, pcdata))
-    return 0;
+    return nullptr;
   String bitmapId;
   if (tag == "image" && attr.has("bitmap", bitmapId)) {
     int objNum = Lex(bitmapId).getInt();
@@ -435,6 +441,9 @@ bool ImlParser::parseStyle(StyleSheet &sheet)
   if (att.has("name", name))
     sheet.setName(name);
 
+  if (att.slash())
+    return true;
+
   String tag = parseToTag();
   while (tag != "/ipestyle") {
     if (tag == "bitmap") {
@@ -456,7 +465,7 @@ bool ImlParser::parseStyle(StyleSheet &sheet)
       else if (att["transformations"] == "translations")
 	symbol.iTransformations = ETransformationsTranslations;
       if (att["xform"] == "yes") {
-	uint flags = Reference::flagsFromName(name);
+	uint32_t flags = Reference::flagsFromName(name);
 	if ((flags & (Reference::EHasStroke|
 		      Reference::EHasFill|
 		      Reference::EHasPen|
@@ -503,14 +512,18 @@ bool ImlParser::parseStyle(StyleSheet &sheet)
       ts.iVerticalAlignment = Text::makeVAlign(att["valign"], EAlignBaseline);
       sheet.setTitleStyle(ts);
     } else if (tag == "pagenumberstyle") {
-      if (!parseAttributes(att) || !att.slash())
+      if (!parseAttributes(att))
 	return false;
       StyleSheet::PageNumberStyle pns;
       pns.iDefined = true;
       Lex lex1(att["pos"]);
       lex1 >> pns.iPos.x >> pns.iPos.y;
-      pns.iFontSize = Lex(att["size"]).getDouble();
-      pns.iColor = Color(att["color"]);
+      pns.iSize = Attribute::makeTextSize(att["size"]);
+      pns.iColor = Attribute::makeColor(att["color"], Attribute::BLACK());
+      pns.iVerticalAlignment = Text::makeVAlign(att["valign"], EAlignBaseline);
+      pns.iHorizontalAlignment = Text::makeHAlign(att["halign"], EAlignLeft);
+      if (!att.slash() && !parsePCDATA(tag, pns.iText))
+	return false;
       sheet.setPageNumberStyle(pns);
     } else if (tag == "preamble") {
       if (!parseAttributes(att))
@@ -609,7 +622,7 @@ bool ImlParser::parseStyle(StyleSheet &sheet)
       }
       if (s.iStops.front().offset < 0.0 || s.iStops.front().offset > 1.0)
 	return false;
-      for (uint i = 1; i < s.iStops.size(); ++i) {
+      for (int i = 1; i < size(s.iStops); ++i) {
 	if (s.iStops[i].offset < s.iStops[i-1].offset)
 	  return false;
       }
@@ -651,8 +664,8 @@ bool ImlParser::parseStyle(StyleSheet &sheet)
       String value = att["begin"];
       value += '\0';
       value += att["end"];
-      sheet.add(ETextStyle, Attribute(true, name),
-		Attribute(false, value));
+      Kind k = (att["type"] == "label") ? ELabelStyle : ETextStyle;
+      sheet.add(k, Attribute(true, name), Attribute(false, value));
     } else {
       Kind kind;
       if (tag == "pen")
@@ -676,6 +689,9 @@ bool ImlParser::parseStyle(StyleSheet &sheet)
 					      Attribute::NORMAL());
       if (name.empty() || value.isSymbolic())
 	return false;
+      if (kind == EGridSize &&
+	  (!value.isNumber() || !value.number().isInteger()))
+	return false; // refuse non-integer gridsize
       sheet.add(kind, Attribute(true, name), value);
     }
     tag = parseToTag();
@@ -693,16 +709,16 @@ StyleSheet *ImlParser::parseStyleSheet()
   if (tag == "?xml") {
     XmlAttributes attr;
     if (!parseAttributes(attr, true))
-      return 0;
+      return nullptr;
     tag = parseToTag();
   }
   if (tag != "ipestyle")
-    return 0;
+    return nullptr;
   StyleSheet *sheet = new StyleSheet();
   if (parseStyle(*sheet))
     return sheet;
   delete sheet;
-  return 0;
+  return nullptr;
 }
 
 // --------------------------------------------------------------------
