@@ -4,7 +4,7 @@
 /*
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (C) 1993-2016  Otfried Cheong
+    Copyright (c) 1993-2019 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -207,8 +207,8 @@ int Page::findLayer(String name) const
 const char * const layerNames[] = {
   "alpha", "beta", "gamma", "delta", "epsilon", "zeta", "eta",
   "theta", "iota", "kappa", "lambda", "mu", "nu", "xi",
-  "omicron", "pi", "rho", "sigma", "tau", "phi", "chi", "xi",
-  "omega" };
+  "omicron", "pi", "rho", "sigma", "tau", "upsilon", "phi",
+  "chi", "psi", "omega" };
 
 //! Add a new layer with unique name.
 void Page::addLayer()
@@ -275,6 +275,16 @@ void Page::removeLayer(String name)
   iLayers.erase(iLayers.begin() + index);
 }
 
+//! Return number of objects in each layer
+void Page::objectsPerLayer(std::vector<int> &objcounts) const
+{
+  objcounts.clear();
+  for (int i = 0; i < countLayers(); ++i)
+    objcounts.push_back(0);
+  for (const auto & obj : iObjects)
+    objcounts[obj.iLayer] += 1;
+}
+
 //! Rename a layer.
 void Page::renameLayer(String oldName, String newName)
 {
@@ -285,10 +295,10 @@ void Page::renameLayer(String oldName, String newName)
 }
 
 //! Returns a precise bounding box for the artwork on the page.
-/*! This is meant to be used as the bounding box in PDF and EPS
-  output.  It is computed by rendering all objects on the page that
-  are visible in at least one view, plus all objects in a layer named
-  "BBOX" (even if that is not visible), using a BBoxPainter. */
+/*! This is meant to be used as the bounding box in PDF output.  It is
+  computed by rendering all objects on the page that are visible in at
+  least one view, plus all objects in a layer named "BBOX" (even if
+  that is not visible), using a BBoxPainter. */
 Rect Page::pageBBox(const Cascade *sheet) const
 {
   // make table with all layers to be rendered
@@ -325,6 +335,13 @@ Rect Page::viewBBox(const Cascade *sheet, int view) const
   return bboxPainter.bbox();
 }
 
+//! Snapping occurs if the layer is visible and has snapping enabled.
+bool Page::objSnapsInView(int objNo, int view) const noexcept
+{
+  int layer = layerOf(objNo);
+  return hasSnapping(layer) && visible(view, layer);
+}
+
 // --------------------------------------------------------------------
 
 //! Set effect of view.
@@ -392,11 +409,35 @@ int Page::countMarkedViews() const
   return (count == 0) ? 1 : count;
 }
 
+//! Return index of view with given number or name.
+/*! Input numbers are one-based.
+  Returns -1 if no such view exists.
+  View names are not yet implemented!!!
+*/
+int Page::findView(String s) const
+{
+  if (s.empty())
+    return -1;
+  if ('0' <= s[0] && s[0] <= '9') {
+    int no = Lex(s).getInt();
+    if (no <= 0 || no > countViews())
+      return -1;
+    return no - 1;
+  }
+  /*
+    for (size_t i = 0; i < countViews(); ++i) {
+    if (s == <name of view i>)
+    return i;
+    }
+  */
+  return -1;
+}
+
 // --------------------------------------------------------------------
 
 Page::SObject::SObject()
 {
-  iObject = 0;
+  iObject = nullptr;
   iLayer = 0;
   iSelect = ENotSelected;
 }
@@ -407,7 +448,7 @@ Page::SObject::SObject(const SObject &rhs)
   if (rhs.iObject)
     iObject = rhs.iObject->clone();
   else
-    iObject = 0;
+    iObject = nullptr;
 }
 
 Page::SObject &Page::SObject::operator=(const SObject &rhs)
@@ -419,7 +460,7 @@ Page::SObject &Page::SObject::operator=(const SObject &rhs)
     if (rhs.iObject)
       iObject = rhs.iObject->clone();
     else
-      iObject = 0;
+      iObject = nullptr;
     iBBox.clear(); // invalidate
   }
   return *this;
@@ -497,13 +538,13 @@ void Page::invalidateBBox(int i) const
 
 //! Return a bounding box for the object at index \a i.
 /*! This is a bounding box including the control points of the object.
-    If you need a tight bounding box, you'll need to use the Object
-    directly.
+  If you need a tight bounding box, you'll need to use the Object
+  directly.
 
-    The Page caches the box the first time it is computed.
+  The Page caches the box the first time it is computed.
 
-    Make sure you call Page::transform instead of Object::setMatrix,
-    as the latter would not invalidate the bounding box.
+  Make sure you call Page::transform instead of Object::setMatrix,
+  as the latter would not invalidate the bounding box.
 */
 Rect Page::bbox(int i) const
 {
@@ -547,7 +588,7 @@ void Page::snapBnd(int i, const Vector &mouse,
 
 //! Set attribute \a prop of object at index \a i to \a value.
 /*! This method automatically invalidates the bounding box if a
-    ETextSize property is actually changed. */
+  ETextSize property is actually changed. */
 bool Page::setAttribute(int i, Property prop, Attribute value)
 {
   bool changed = object(i)->setAttribute(prop, value);
@@ -613,7 +654,7 @@ void Page::setMarked(bool marked)
 const Text *Page::titleText() const
 {
   if (title().empty())
-    return 0;
+    return nullptr;
   return &iTitleObject;
 }
 
