@@ -4,7 +4,7 @@
 /*
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (C) 1993-2016  Otfried Cheong
+    Copyright (c) 1993-2019 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -62,7 +62,7 @@ Path *Path::create(const XmlAttributes &attr, String data)
 {
   std::unique_ptr<Path> self(new Path(attr));
   if (!self->iShape.load(data))
-    return 0;
+    return nullptr;
   self->makeArrowData();
   return self.release();
 }
@@ -172,16 +172,12 @@ void Path::init(const AllAttributes &attr, bool withArrows)
   iFArrowShape = iRArrowShape = Attribute::ARROW_NORMAL();
   iFArrowSize = iRArrowSize = Attribute::NORMAL();
   if (withArrows) {
-    if (attr.iFArrow) {
-      iHasFArrow = true;
-      iFArrowShape = attr.iFArrowShape;
-      iFArrowSize = attr.iFArrowSize;
-    }
-    if (attr.iRArrow) {
-      iHasRArrow = true;
-      iRArrowShape = attr.iRArrowShape;
-      iRArrowSize = attr.iRArrowSize;
-    }
+    iFArrowShape = attr.iFArrowShape;
+    iRArrowShape = attr.iRArrowShape;
+    iFArrowSize = attr.iFArrowSize;
+    iRArrowSize = attr.iRArrowSize;
+    iHasFArrow = attr.iFArrow;
+    iHasRArrow = attr.iRArrow;
   }
 }
 
@@ -312,6 +308,7 @@ void Path::drawArrow(Painter &painter, Vector pos, Angle angle,
   if (symbol) {
     double s = painter.cascade()->find(EArrowSize, size).number().toDouble();
     Color color = painter.stroke();
+    // Fixed opaq = painter.opacity();
 
     painter.push();
     painter.pushMatrix();
@@ -322,6 +319,8 @@ void Path::drawArrow(Painter &painter, Vector pos, Angle angle,
     bool cw = (radius < 0);
     if (cw)
       radius = -radius;
+    bool pointy = (shape == Attribute::ARROW_PTARC() ||
+		   shape == Attribute::ARROW_FPTARC());
 
     if (shape.isArcArrow() && (radius > s)) {
       Angle delta = s / radius;
@@ -352,15 +351,20 @@ void Path::drawArrow(Painter &painter, Vector pos, Angle angle,
 	painter.setFill(Attribute(Color(1000, 1000, 1000)));
       else
 	painter.setFill(Attribute(color));
+      // painter.setOpacity(Attribute(opaq));
       painter.newPath();
       painter.moveTo(arc1.beginp());
       painter.drawArc(arc1);
-      if (shape == Attribute::ARROW_PTARC() ||
-	  shape == Attribute::ARROW_FPTARC()) {
-	painter.lineTo(arc3.endp());
+      if (cw) {
+	if (pointy)
+	  painter.lineTo(arc3.endp());
+	painter.lineTo(arc2.beginp());
+	painter.drawArc(arc2);
+      } else {
+	painter.drawArc(arc2);
+	if (pointy)
+	  painter.lineTo(arc3.beginp());
       }
-      painter.lineTo(arc2.beginp());
-      painter.drawArc(arc2);
       painter.closePath();
       painter.drawPath(EStrokedAndFilled);
     } else {
@@ -434,8 +438,7 @@ void Path::draw(Painter &painter) const
     }
     if (iHasRArrow && iRArrowOk) {
       double r = 0.0;
-      if (iRArrowArc && (iRArrowShape == Attribute::ARROW_ARC() ||
-			 iRArrowShape == Attribute::ARROW_FARC())) {
+      if (iRArrowArc && iRArrowShape.isArcArrow()) {
 	CurveSegment seg = iShape.subPath(0)->asCurve()->segment(0);
 	Vector center = painter.matrix() * seg.matrix().translation();
 	r = (center - painter.matrix() * iRArrowPos).len();
@@ -531,7 +534,6 @@ void Path::setOpacity(Attribute opaq)
 //! Set pen.
 void Path::setPen(Attribute pen)
 {
-  ipeDebug("set pen");
   iPen = pen;
 }
 
@@ -706,7 +708,7 @@ bool Path::setAttribute(Property prop, Attribute value)
   return false;
 }
 
-Attribute Path::getAttribute(Property prop)
+Attribute Path::getAttribute(Property prop) const noexcept
 {
   switch (prop) {
   case EPropPathMode:
