@@ -4,7 +4,7 @@
 /*
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (C) 1993-2016  Otfried Cheong
+    Copyright (c) 1993-2019 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -175,7 +175,7 @@ void PdfPainter::doDrawPath(TPathMode mode)
   // if (!mode)
   // iStream << "n\n"; // no op path
 
-  const Gradient *g = 0;
+  const Gradient *g = nullptr;
   Attribute grad = iState.back().iGradient;
   if (!grad.isNormal())
     g = iCascade->findGradient(grad);
@@ -209,76 +209,13 @@ void PdfPainter::doDrawText(const Text *text)
   const Text::XForm *xf = text->getXForm();
   if (!xf)
     return;
-
-  Color s = stroke();
   drawOpacity();
-
+  pushMatrix();
+  transform(Matrix(xf->iStretch, 0, 0, xf->iStretch, 0, 0));
+  translate(xf->iTranslation);
   iStream << matrix() << " cm " ;
-  if (xf->iStretch != 1.0 || xf->iBBox.bottomLeft() != Vector::ZERO)
-    iStream << Matrix(xf->iStretch, 0, 0, xf->iStretch,
-		      -xf->iBBox.bottomLeft().x,
-		      -xf->iBBox.bottomLeft().y) << " cm ";
-
-  BufferSource source(xf->iStream);
-  PdfParser parser(source);
-  std::vector<const PdfObj *> args;
-  while (!parser.eos()) {
-    PdfToken tok = parser.token();
-    if (tok.iType != PdfToken::EOp) {
-      const PdfObj *obj = parser.getObject();
-      if (!obj) {
-	ipeDebug("Error parsing PDF text");
-	break; // no further parsing attempted
-      }
-      args.push_back(obj);
-    } else {
-      // its an operator, execute it
-      String op = tok.iString;
-      parser.getToken();
-      if (op == "k") {
-	drawColor(iStream, s, "g", "rg");
-      } else if (op == "K") {
-	drawColor(iStream, s, "G", "RG");
-      } else if (op == "sym") {
-	if (args.size() == 5) {
-	  String s[5];
-	  for (int i = 0; i < 5; ++i) {
-	    if (args[i]->string())
-	      s[i] = args[i]->string()->value();
-	  }
-	  // ipeDebug("render Symbol %s, %s, %s, %s, %s",
-	  // s[0].z(), s[1].z(), s[2].z(), s[3].z(), s[4].z());
-	  const Symbol *symbol =
-	    cascade()->findSymbol(Attribute(true, s[0]));
-	  if (symbol) {
-	    PdfPainter sympt(cascade(), iStream);
-	    sympt.pushMatrix();
-	    double sa = Lex(s[1]).getDouble();
-	    Matrix m(sa, 0, 0, sa, 0, 0);
-	    sympt.transform(m);
-	    sympt.setSymStroke(Attribute::makeColor(s[2], Attribute::BLACK()));
-	    sympt.setSymFill(Attribute::makeColor(s[3], Attribute::WHITE()));
-	    sympt.setSymPen(Attribute::makeScalar(s[4], Attribute::NORMAL()));
-	    symbol->iObject->draw(sympt);
-	    sympt.popMatrix();
-	  }
-	}
-      } else {
-	for (uint i = 0; i < args.size(); ++i)
-	  iStream << args[i]->repr() << " ";
-	iStream << op << "\n";
-      }
-      while (!args.empty()) {
-	delete args.back();
-	args.pop_back();
-      }
-    }
-  }
-  // delete remaining args
-  while (!args.empty()) {
-    delete args.back();
-    args.pop_back();
-  }
+  iStream << "/" << xf->iName << " Do\n";
+  popMatrix();
 }
 
 void PdfPainter::doDrawSymbol(Attribute symbol)
@@ -310,25 +247,26 @@ void PdfPainter::doDrawSymbol(Attribute symbol)
     - 0: Must be left empty (a PDF restriction).
     - 1:  XML stream.
     - 2: Parent of all pages objects.
-
+    - 3: ExtGState resource from pdflatex
+    - 4: Shading resource from pdflatex
+    - 5: Pattern resource from pdflatex
+    - 6: ColorSpace resource from pdflatex
 */
 
 //! Create a PDF writer operating on this (open and empty) file.
-PdfWriter::PdfWriter(TellStream &stream, const Document *doc,
-		     const PdfResources *resources,
-		     bool markedView, int fromPage, int toPage,
+PdfWriter::PdfWriter(TellStream &stream, const Document *doc, const PdfResources *resources,
+		     uint32_t flags, int fromPage, int toPage,
 		     int compression)
-  : iStream(stream), iDoc(doc), iMarkedView(markedView),
+  : iStream(stream), iDoc(doc), iResources(resources), iSaveFlags(flags),
     iFromPage(fromPage), iToPage(toPage)
 {
   iCompressLevel = compression;
-  iObjNum = 3;  // 0 - 2 are reserved
+  iObjNum = 7;  // 0 - 6 are reserved
   iXmlStreamNum = -1; // no XML stream yet
-  iResourceNum = -1;
-  iPageNumberFont = -1;
   iExtGState = -1;
   iPatternNum = -1;
   iBookmarks = -1;
+  iDests = -1;
 
   if (iFromPage < 0 || iFromPage >= iDoc->countPages())
     iFromPage = 0;
@@ -347,23 +285,23 @@ PdfWriter::PdfWriter(TellStream &stream, const Document *doc,
 
   iStream << "%PDF-1.4\n";
 
-  if (iDoc->properties().iNumberPages) {
-    iPageNumberFont = startObject();
-    iStream << "<<\n"
-	    << "/Type /Font\n"
-	    << "/Subtype /Type1\n"
-	    << "/BaseFont /Helvetica\n"
-	    << ">> endobj\n";
-  }
+  // embed all fonts and other resources from Pdflatex
+  embedResources();
 
-  // embed all opacities
+  // embed resources from pdflatex
+  embedLatexResource(3, "ExtGState");
+  embedLatexResource(4, "Shading");
+  embedLatexResource(5, "Pattern");
+  embedLatexResource(6, "ColorSpace");
+
+  // embed all extgstate objects
   AttributeSeq os;
   iDoc->cascade()->allNames(EOpacity, os);
-  if (os.size()) {
+  if (os.size() > 0) {
     iExtGState = startObject();
     iStream << "<<\n";
-    for (uint i = 0; i < os.size(); ++i) {
-      Attribute alpha = iDoc->cascade()->find(EOpacity, os[i]);
+    for (const auto & obj : os) {
+      Attribute alpha = iDoc->cascade()->find(EOpacity, obj);
       assert(alpha.isNumber());
       iStream << opacityName(alpha.number()) << " << /CA " << alpha.number()
 	      << " /ca " << alpha.number() << " >>\n";
@@ -374,8 +312,8 @@ PdfWriter::PdfWriter(TellStream &stream, const Document *doc,
   // embed all gradients
   AttributeSeq gs;
   iDoc->cascade()->allNames(EGradient, gs);
-  for (uint i = 0; i < gs.size(); ++i) {
-    const Gradient *g = iDoc->cascade()->findGradient(gs[i]);
+  for (const auto & grad : gs) {
+    const Gradient *g = iDoc->cascade()->findGradient(grad);
     int num = startObject();
     iStream << "<<\n"
 	    << " /ShadingType " << int(g->iType) << "\n"
@@ -401,7 +339,7 @@ PdfWriter::PdfWriter(TellStream &stream, const Document *doc,
 	      << "  /FunctionType 3 /Domain [ 0 1 ]\n"
 	      << "  /Bounds [";
       int count = 0;
-      for (uint i = 1; i < g->iStops.size() - 1; ++i) {
+      for (int i = 1; i < size(g->iStops) - 1; ++i) {
 	if (g->iStops[i].offset > g->iStops[i-1].offset) {
 	  iStream << g->iStops[i].offset << " ";
 	  ++count;
@@ -411,7 +349,7 @@ PdfWriter::PdfWriter(TellStream &stream, const Document *doc,
       for (int i = 0; i <= count; ++i)
 	iStream << "0.0 1.0 ";
       iStream << "]\n  /Functions [\n";
-      for (uint i = 1; i < g->iStops.size(); ++i) {
+      for (int i = 1; i < size(g->iStops); ++i) {
 	if (g->iStops[i].offset > g->iStops[i-1].offset) {
 	  iStream << "   << /FunctionType 2 /Domain [ 0 1 ] /N 1 /C0 [";
 	  g->iStops[i-1].color.saveRGB(iStream);
@@ -423,7 +361,7 @@ PdfWriter::PdfWriter(TellStream &stream, const Document *doc,
       iStream << "] >>\n";
     }
     iStream << ">> endobj\n";
-    iGradients[gs[i].index()] = num;
+    iGradients[grad.index()] = num;
   }
 
   // embed all tilings
@@ -431,8 +369,8 @@ PdfWriter::PdfWriter(TellStream &stream, const Document *doc,
   std::map<int, int> patterns;
   iDoc->cascade()->allNames(ETiling, ts);
   if (ts.size() > 0) {
-    for (uint i = 0; i < ts.size(); ++i) {
-      const Tiling *t = iDoc->cascade()->findTiling(ts[i]);
+    for (const auto & tiling : ts) {
+      const Tiling *t = iDoc->cascade()->findTiling(tiling);
       Linear m(t->iAngle);
       int num = startObject();
       iStream << "<<\n"
@@ -441,7 +379,7 @@ PdfWriter::PdfWriter(TellStream &stream, const Document *doc,
 	      << "/PaintType 2\n"    // uncolored pattern
 	      << "/TilingType 2\n"   // faster
 	      << "/BBox [ 0 0 100 " << t->iStep << " ]\n"
-	      << "/XStep 99\n"
+	      << "/XStep 100\n"
 	      << "/YStep " << t->iStep << "\n"
 	      << "/Resources << >>\n"
 	      << "/Matrix [" << m << " 0 0]\n";
@@ -449,28 +387,25 @@ PdfWriter::PdfWriter(TellStream &stream, const Document *doc,
       StringStream ss(s);
       ss << "0 0 100 " << t->iWidth << " re f\n";
       createStream(s.data(), s.size(), false);
-      patterns[ts[i].index()] = num;
+      patterns[tiling.index()] = num;
     }
 
     // create pattern dictionary
     iPatternNum = startObject();
     iStream << "<<\n";
-    for (uint i = 0; i < ts.size(); ++i) {
-      iStream << "/Pat" << ts[i].index() << " "
-	      << patterns[ts[i].index()] << " 0 R\n";
+    for (const auto & pattern : ts) {
+      iStream << "/Pat" << pattern.index() << " "
+	      << patterns[pattern.index()] << " 0 R\n";
     }
     iStream << ">> endobj\n";
   }
 
-  // embed all fonts and other resources from Pdflatex
-  embedResources(resources);
-
   // embed all symbols with xform attribute
   AttributeSeq sys;
   iDoc->cascade()->allNames(ESymbol, sys);
   if (sys.size()) {
-    for (uint i = 0; i < sys.size(); ++i) {
-      const Symbol *sym = iDoc->cascade()->findSymbol(sys[i]);
+    for (const auto & sysi : sys) {
+      const Symbol *sym = iDoc->cascade()->findSymbol(sysi);
       if (sym->iXForm) {
 	// compute bbox for object
 	BBoxPainter bboxPainter(iDoc->cascade());
@@ -491,7 +426,7 @@ PdfWriter::PdfWriter(TellStream &stream, const Document *doc,
 	PdfPainter painter(iDoc->cascade(), ss);
 	sym->iObject->draw(painter);
 	createStream(s.data(), s.size(), false);
-	iSymbols[sys[i].index()] = num;
+	iSymbols[sysi.index()] = num;
       }
     }
   }
@@ -515,26 +450,125 @@ int PdfWriter::startObject(int objnum)
   return objnum;
 }
 
+bool PdfWriter::hasResource(String kind) const noexcept
+{
+  return iResources && iResources->resourcesOfKind(kind);
+}
+
 /*! Write all PDF resources to the PDF file, and record their object
   numbers.  Embeds nothing if \c resources is nullptr, but must be
   called nevertheless. */
-void PdfWriter::embedResources(const PdfResources *resources)
+void PdfWriter::embedResources()
 {
-  std::unordered_map<int, int> resourceNumber;
-
-  if (resources) {
-    const std::vector<int> &seq = resources->embedSequence();
+  bool inflate = (iCompressLevel == 0);
+  if (iResources) {
+    const std::vector<int> &seq = iResources->embedSequence();
     for (auto &num : seq) {
-      const PdfObj *obj = resources->object(num);
+      const PdfObj *obj = iResources->object(num);
       int embedNum = startObject();
-      obj->write(iStream, &resourceNumber);
+      if (obj->dict() && obj->dict()->get("IpeId", nullptr))
+	embedIpeXForm(obj->dict());
+      else
+	obj->write(iStream, &iResourceNumber, inflate);
       iStream << " endobj\n";
-      resourceNumber[num] = embedNum;
+      iResourceNumber[num] = embedNum;
     }
-    iResourceNum = startObject();
-    const PdfDict *font = resources->get("Font");
-    font->write(iStream,  &resourceNumber);
-    iStream << " endobj\n";
+  }
+}
+
+void PdfWriter::embedIpeXForm(const PdfDict *d)
+{
+  bool inflate = (iCompressLevel == 0) && d->deflated();
+  iStream << "<<";
+  for (int i = 0; i < d->count(); ++i) {
+    String key = d->key(i);
+    // skip /IpeId etc keys
+    if (key.left(3) == "Ipe")
+      continue;
+    if ((inflate && key == "Filter") || key == "Length")
+      continue;   // remove /FlateDecode filter and /Length
+    iStream << "/" << key << " ";
+    if (key == "Resources") {
+      const PdfObj *res = d->value(i);
+      if (res->ref())
+	res = iResources->object(res->ref()->value());
+      if (res->dict())
+	embedXFormResource(res->dict());
+      else	// should not happen!
+	d->value(i)->write(iStream, &iResourceNumber);
+    } else if (key == "BBox") {
+      const TextPadding *pad = iDoc->cascade()->findTextPadding();
+      std::vector<double> bbox;
+      d->getNumberArray("BBox", nullptr, bbox);
+      if (pad && bbox.size() == 4) {
+	bbox[0] -= pad->iLeft;
+	bbox[1] -= pad->iBottom;
+	bbox[2] += pad->iRight;
+	bbox[3] += pad->iTop;
+      }
+      iStream << "[";
+      for (auto it = bbox.begin(); it != bbox.end(); ++it)
+	iStream << *it << " ";
+      iStream << "]";
+    } else
+      d->value(i)->write(iStream, &iResourceNumber);
+    iStream << " ";
+  }
+  Buffer stream = inflate ? d->inflate() : d->stream();
+  if (stream.size() > 0) {
+    iStream << "/Length " << stream.size()
+	    << ">>\nstream\n";
+    for (int i = 0; i < stream.size(); ++i)
+      iStream.putChar(stream[i]);
+    iStream << "\nendstream";
+  } else
+    iStream << ">>";
+}
+
+void PdfWriter::embedXFormResource(const PdfDict *d)
+{
+  iStream << "<<";
+  for (int i = 0; i < d->count(); ++i) {
+    String key = d->key(i);
+    iStream << "/" << key << " ";
+    if (key == "ColorSpace" || key == "Shading" || key == "Pattern" ||
+	key == "ExtGState") {
+      ipeDebug("PDF Writer: Conflicting resource in XForm: %s", key.z());
+    } else
+      d->value(i)->write(iStream, &iResourceNumber);
+  }
+  if (hasResource("ExtGState"))
+    iStream << "/ExtGState 3 0 R\n";
+  if (hasResource("Shading"))
+    iStream << "/ColorSpace 4 0 R\n";
+  if (hasResource("Pattern"))
+    iStream << "/Pattern 5 0 R\n";
+  if (hasResource("ColorSpace"))
+    iStream << "/ColorSpace 6 0 R\n";
+  iStream << ">>";
+}
+
+void PdfWriter::embedLatexResource(int num, String kind)
+{
+  if (hasResource(kind)) {
+    startObject(num);
+    iStream << "<<\n";
+    embedResource(kind);
+    iStream << ">> endobj\n";
+  }
+}
+
+void PdfWriter::embedResource(String kind)
+{
+  if (!iResources)
+    return;
+  const PdfDict *d = iResources->resourcesOfKind(kind);
+  if (!d)
+    return;
+  for (int i = 0; i < d->count(); ++i) {
+    iStream << "/" << d->key(i) << " ";
+    d->value(i)->write(iStream, &iResourceNumber);
+    iStream << " ";
   }
 }
 
@@ -556,7 +590,7 @@ void PdfWriter::createStream(const char *data, int size, bool preCompressed)
   if (iCompressLevel > 0) {
     int deflatedSize;
     Buffer deflated = DeflateStream::deflate(data, size, deflatedSize,
-						   iCompressLevel);
+					     iCompressLevel);
     iStream << "/Length " << deflatedSize
 	    << " /Filter /FlateDecode >>\nstream\n";
     iStream.putRaw(deflated.data(), deflatedSize);
@@ -572,46 +606,48 @@ void PdfWriter::createStream(const char *data, int size, bool preCompressed)
 
 void PdfWriter::embedBitmap(Bitmap bitmap)
 {
+  int smaskNum = -1;
+  auto embed = bitmap.embed();
+  if (bitmap.hasAlpha() && embed.second.size() > 0) {
+    smaskNum = startObject();
+    iStream << "<<\n";
+    iStream << "/Type /XObject\n";
+    iStream << "/Subtype /Image\n";
+    iStream << "/Width " << bitmap.width() << "\n";
+    iStream << "/Height " << bitmap.height() << "\n";
+    iStream << "/ColorSpace /DeviceGray\n";
+    iStream << "/Filter /FlateDecode\n";
+    iStream << "/BitsPerComponent 8\n";
+    iStream << "/Length " << embed.second.size() << "\n>> stream\n";
+    iStream.putRaw(embed.second.data(), embed.second.size());
+    iStream << "\nendstream endobj\n";
+  }
   int objnum = startObject();
   iStream << "<<\n";
   iStream << "/Type /XObject\n";
   iStream << "/Subtype /Image\n";
   iStream << "/Width " << bitmap.width() << "\n";
   iStream << "/Height " << bitmap.height() << "\n";
-  switch (bitmap.colorSpace()) {
-  case Bitmap::EDeviceGray:
+  if (bitmap.isGray())
     iStream << "/ColorSpace /DeviceGray\n";
-    break;
-  case Bitmap::EDeviceRGB:
+  else
     iStream << "/ColorSpace /DeviceRGB\n";
-    break;
-  case Bitmap::EDeviceCMYK:
-    iStream << "/ColorSpace /DeviceCMYK\n";
-    // apparently JPEG CMYK images need this decode array??
-    // iStream << "/Decode [1 0 1 0 1 0 1 0]\n";
-    break;
-  }
-  switch (bitmap.filter()) {
-  case Bitmap::EFlateDecode:
-    iStream << "/Filter /FlateDecode\n";
-    break;
-  case Bitmap::EDCTDecode:
+  if (bitmap.isJpeg())
     iStream << "/Filter /DCTDecode\n";
-    break;
-  default:
-    // no filter
-    break;
-  }
-  iStream << "/BitsPerComponent " << bitmap.bitsPerComponent() << "\n";
-  if (bitmap.colorKey() >= 0) {
+  else
+    iStream << "/Filter /FlateDecode\n";
+  iStream << "/BitsPerComponent 8\n";
+  if (smaskNum >= 0) {
+    iStream << "/SMask " << smaskNum << " 0 R\n";
+  } else if (bitmap.colorKey() >= 0) {
     int r = (bitmap.colorKey() >> 16) & 0xff;
     int g = (bitmap.colorKey() >> 8) & 0xff;
     int b = bitmap.colorKey() & 0xff;
     iStream << "/Mask [" << r << " " << r << " " << g << " " << g
 	    << " " << b << " " << b << "]\n";
   }
-  iStream << "/Length " << bitmap.size() << "\n>> stream\n";
-  iStream.putRaw(bitmap.data(), bitmap.size());
+  iStream << "/Length " << embed.first.size() << "\n>> stream\n";
+  iStream.putRaw(embed.first.data(), embed.first.size());
   iStream << "\nendstream endobj\n";
   bitmap.setObjNum(objnum);
 }
@@ -636,15 +672,21 @@ void PdfWriter::embedBitmaps(const BitmapFinder &bm)
 
 void PdfWriter::createResources(const BitmapFinder &bm)
 {
-  iStream << "/Resources <<\n  /ProcSet [ /PDF ";
-  if (iResourceNum >= 0)
+  // These are only the resources needed by Ipe drawing directly.
+  // Resources used from inside text objects (e.g. by tikz)
+  // are stored inside the XObject resources.
+  // Font is only used inside XObject, no font resource needed on page
+  // Resources:
+  // ProcSet, ExtGState, ColorSpace, Pattern, Shading, XObject, Font
+  // not used now: Properties
+  // ProcSet
+  iStream << "/Resources <<\n  /ProcSet [/PDF";
+  if (iResources)
     iStream << "/Text";
   if (!bm.iBitmaps.empty())
-    iStream << "/ImageB /ImageC";
-  iStream << " ]\n";
-  if (iResourceNum >= 0)
-    iStream << "  /Font " << iResourceNum << " 0 R\n";
-  // embed resources for gradients
+    iStream << "/ImageB/ImageC";
+  iStream << "]\n";
+  // Shading
   if (iGradients.size()) {
     iStream << "  /Shading <<";
     for (std::map<int,int>::const_iterator it = iGradients.begin();
@@ -652,14 +694,21 @@ void PdfWriter::createResources(const BitmapFinder &bm)
       iStream << " /Grad" << it->first << " " << it->second << " 0 R";
     iStream << " >>\n";
   }
+  // ExtGState
   if (iExtGState >= 0)
     iStream << "  /ExtGState " << iExtGState << " 0 R\n";
-  // patterns
+  // ColorSpace
   if (iPatternNum >= 0) {
-    iStream << "  /ColorSpace << /PCS [/Pattern /DeviceRGB] >>\n";
-    iStream << "  /Pattern " << iPatternNum << " 0 R\n";
+    iStream << "  /ColorSpace << /PCS [/Pattern /DeviceRGB] ";
+    iStream << ">>\n";
   }
-  if (!bm.iBitmaps.empty() || !iSymbols.empty()) {
+  // Pattern
+  if (iPatternNum >= 0)
+    iStream << "  /Pattern " << iPatternNum << " 0 R\n";
+  // XObject
+  // TODO: Is "hasResource" here used correctly?
+  // From Tikz xobject resources seem to go into the Ipe xform.
+  if (!bm.iBitmaps.empty() || !iSymbols.empty() || hasResource("XObject")) {
     iStream << "  /XObject << ";
     for (BmIter it = bm.iBitmaps.begin(); it != bm.iBitmaps.end(); ++it) {
       // mention each PDF object only once
@@ -673,6 +722,7 @@ void PdfWriter::createResources(const BitmapFinder &bm)
     for (std::map<int,int>::const_iterator it = iSymbols.begin();
 	 it != iSymbols.end(); ++it)
       iStream << "/Symbol" << it->first << " " << it->second << " 0 R ";
+    embedResource("XObject");
     iStream << ">>\n";
   }
   iStream << "  >>\n";
@@ -690,18 +740,10 @@ void PdfWriter::paintView(Stream &stream, int pno, int view)
   if (background && page->findLayer("BACKGROUND") < 0)
     painter.drawSymbol(Attribute::BACKGROUND());
 
-  // page numbers
-  if (iPageNumberFont >= 0) {
-    const StyleSheet::PageNumberStyle *pns =
-      iDoc->cascade()->findPageNumberStyle();
-    stream << "q ";
-    PdfPainter::drawColor(stream, pns->iColor, "g", "rg");
-    stream << "BT /F" << iPageNumberFont << " " << pns->iFontSize
-	   << " Tf " << pns->iPos << " Td\n"
-	   << "[(" << (pno + 1);
-    if (page->countMarkedViews() > 1)
-      stream << "-" << (view + 1);
-    stream << ")] TJ ET Q\n";
+  if (iDoc->properties().iNumberPages && iResources) {
+    const Text *pn = iResources->pageNumber(pno, view);
+    if (pn)
+      pn->draw(painter);
   }
 
   const Text *title = page->titleText();
@@ -737,12 +779,66 @@ void PdfWriter::createPageView(int pno, int view)
   } else
     paintView(sstream, pno, view);
 
+  int firstLink = -1;
+  int lastLink = -1;
+  for (int i = 0; i < page->count(); ++i) {
+    const Group *g = page->object(i)->asGroup();
+    if (g && page->objectVisible(view, i) && !g->url().empty()) {
+      lastLink = startObject();
+      if (firstLink < 0) firstLink = lastLink;
+      iStream << "<<\n"
+	      << "/Type /Annot\n"
+	      << "/Subtype /Link\n"
+	      << "/H /N\n"
+	      << "/Rect [" << page->bbox(i) << "]\n"
+	      << "/A <</Type/Action/S";
+      String url = g->url();
+      if (url.left(6) == "named:") {
+	iStream << "/Named/N/" << url.substr(6);
+      } else {
+	if (url.left(7) == "launch:") {
+	  url = url.substr(7);
+	  iStream << "/Launch/F";
+	} else if (url.left(5) == "goto:") {
+	  url = url.substr(5);
+	  iStream << "/GoTo/D";
+	} else
+	  iStream << "/URI/URI";
+	writeString(url);
+      }
+      iStream << ">>\n>> endobj\n";
+    }
+  }
+  int notesObj = -1;
+  if (!page->notes().empty() &&
+      (!(iSaveFlags & SaveFlag::Export) || (iSaveFlags & SaveFlag::KeepNotes))) {
+    notesObj = startObject();
+    iStream << "<<\n"
+	    << "/Type /Annot\n"
+	    << "/Subtype /Text\n"
+	    << "/Rect [20 40 30 40]\n"
+	    << "/F 4\n"
+	    << "/Contents ";
+    writeString(page->notes());
+    iStream << "\n>> endobj\n";
+  }
+
   int contentsobj = startObject();
   iStream << "<<\n";
   createStream(pagedata.data(), pagedata.size(), (iCompressLevel > 0));
   int pageobj = startObject();
   iStream << "<<\n";
   iStream << "/Type /Page\n";
+  if (firstLink >= 0 || notesObj >= 0) {
+    iStream << "/Annots [ ";
+    if (firstLink >= 0) {
+      while (firstLink <= lastLink)
+	iStream << firstLink++ << " 0 R ";
+    }
+    if (notesObj >= 0)
+      iStream << notesObj << " 0 R";
+    iStream << "]\n";
+  }
   iStream << "/Contents " << contentsobj << " 0 R\n";
   // iStream << "/Rotate 0\n";
   createResources(bm);
@@ -766,17 +862,17 @@ void PdfWriter::createPageView(int pno, int view)
     iStream << "/ArtBox [" << bbox << "]\n";
   iStream << "/Parent 2 0 R\n";
   iStream << ">> endobj\n";
-  iPageObjectNumbers.push_back(pageobj);
+  iPageObjectNumbers.push_back({ pno, view, pageobj });
 }
 
 //! Create all PDF pages.
 void PdfWriter::createPages()
 {
   for (int page = iFromPage; page <= iToPage; ++page) {
-    if (iMarkedView && !iDoc->page(page)->marked())
+    if ((iSaveFlags & SaveFlag::MarkedView) && !iDoc->page(page)->marked())
       continue;
     int nViews = iDoc->page(page)->countViews();
-    if (iMarkedView) {
+    if (iSaveFlags & SaveFlag::MarkedView) {
       bool shown = false;
       for (int view = 0; view < nViews; ++view) {
 	if (iDoc->page(page)->markedView(view)) {
@@ -839,12 +935,20 @@ void PdfWriter::writeString(String text)
 
 // --------------------------------------------------------------------
 
+int PdfWriter::pageObjectNumber(int page, int view)
+{
+  auto it = std::find_if(iPageObjectNumbers.begin(), iPageObjectNumbers.end(),
+			 [=](const PON &pon) {return pon.page == page && pon.view == view;});
+  if (it != iPageObjectNumbers.end())
+    return it->objNum;
+  ipeDebug("pageObjectNumber not found, this is a bug!");
+  return 0;
+}
+
 struct Section {
   int iPage;
-  int iSeqPage;
   int iObjNum;
   std::vector<int> iSubPages;
-  std::vector<int> iSubSeqPages;
 };
 
 //! Create the bookmarks (PDF outline).
@@ -852,31 +956,25 @@ void PdfWriter::createBookmarks()
 {
   // first collect all information
   std::vector<Section> sections;
-  int seqPg = 0;
   for (int pg = iFromPage; pg <= iToPage; ++pg) {
+    if ((iSaveFlags & SaveFlag::MarkedView) && !iDoc->page(pg)->marked())
+      continue;
     String s = iDoc->page(pg)->section(0);
     String ss = iDoc->page(pg)->section(1);
     if (!s.empty()) {
       Section sec;
       sec.iPage = pg;
-      sec.iSeqPage = seqPg;
       sections.push_back(sec);
-      // ipeDebug("section on page %d, seq %d", pg, seqPg);
     }
-    if (!sections.empty() && !ss.empty()) {
+    if (!sections.empty() && !ss.empty())
       sections.back().iSubPages.push_back(pg);
-      sections.back().iSubSeqPages.push_back(seqPg);
-      // ipeDebug("subsection on page %d, seq %d", pg, seqPg);
-    }
-    seqPg += iMarkedView ? iDoc->page(pg)->countMarkedViews() :
-      iDoc->page(pg)->countViews();
   }
   if (sections.empty())
     return;
   // reserve outline object
   iBookmarks = iObjNum++;
   // assign object numbers
-  for (uint s = 0; s < sections.size(); ++s) {
+  for (int s = 0; s < size(sections); ++s) {
     sections[s].iObjNum = iObjNum++;
     iObjNum += sections[s].iSubPages.size(); // leave space for subsections
   }
@@ -885,7 +983,7 @@ void PdfWriter::createBookmarks()
   iStream << "<<\n/First " << sections[0].iObjNum << " 0 R\n"
 	  << "/Count " << int(sections.size()) << "\n"
 	  << "/Last " << sections.back().iObjNum << " 0 R\n>> endobj\n";
-  for (uint s = 0; s < sections.size(); ++s) {
+  for (int s = 0; s < size(sections); ++s) {
     int count = sections[s].iSubPages.size();
     int obj = sections[s].iObjNum;
     // embed section
@@ -893,11 +991,11 @@ void PdfWriter::createBookmarks()
     iStream << "<<\n/Title ";
     writeString(iDoc->page(sections[s].iPage)->section(0));
     iStream << "\n/Parent " << iBookmarks << " 0 R\n"
-	    << "/Dest [ " << iPageObjectNumbers[sections[s].iSeqPage]
+	    << "/Dest [ " << pageObjectNumber(sections[s].iPage, 0)
 	    << " 0 R /XYZ null null null ]\n";
     if (s > 0)
       iStream << "/Prev " << sections[s-1].iObjNum << " 0 R\n";
-    if (s < sections.size() - 1)
+    if (s < size(sections) - 1)
       iStream << "/Next " << sections[s+1].iObjNum << " 0 R\n";
     if (count > 0)
       iStream << "/Count " << -count << "\n"
@@ -907,12 +1005,11 @@ void PdfWriter::createBookmarks()
     // using ids obj + 1 .. obj + count for the subsections
     for (int ss = 0; ss < count; ++ss) {
       int pageNo = sections[s].iSubPages[ss];
-      int seqPageNo = sections[s].iSubSeqPages[ss];
       startObject(obj + ss + 1);
       iStream << "<<\n/Title ";
       writeString(iDoc->page(pageNo)->section(1));
       iStream << "\n/Parent " << obj << " 0 R\n"
-	      << "/Dest [ " << iPageObjectNumbers[seqPageNo]
+	      << "/Dest [ " << pageObjectNumber(pageNo, 0)
 	      << " 0 R /XYZ null null null ]\n";
       if (ss > 0)
 	iStream << "/Prev " << (obj + ss) << " 0 R\n";
@@ -923,6 +1020,33 @@ void PdfWriter::createBookmarks()
   }
 }
 
+//! Create the named destinations.
+void PdfWriter::createNamedDests()
+{
+  std::vector<std::pair<String, int>> dests;
+  for (int pg = iFromPage; pg <= iToPage; ++pg) {
+    if ((iSaveFlags & SaveFlag::MarkedView) && !iDoc->page(pg)->marked())
+      continue;
+    String s = iDoc->page(pg)->section(0);
+    if (!s.empty())
+      dests.push_back(std::make_pair(s, pageObjectNumber(pg, 0)));
+  }
+  if (dests.empty())
+    return;
+  std::sort(dests.begin(), dests.end());
+  iDests = startObject();
+  iStream << "<<\n/Limits [";
+  writeString(dests.front().first);
+  iStream << " ";
+  writeString(dests.back().first);
+  iStream << "]\n/Names [\n";
+  for (const auto & dest : dests) {
+    writeString(dest.first);
+    iStream << " [" << dest.second << " 0 R /XYZ null null null]\n";
+  }
+  iStream << "]>> endobj\n";
+}
+
 // --------------------------------------------------------------------
 
 //! Create the root objects and trailer of the PDF file.
@@ -934,13 +1058,23 @@ void PdfWriter::createTrailer()
   iStream << "<<\n" << "/Type /Pages\n";
   iStream << "/Count " << int(iPageObjectNumbers.size()) << "\n";
   iStream << "/Kids [ ";
-  for (std::vector<int>::const_iterator it = iPageObjectNumbers.begin();
-       it != iPageObjectNumbers.end(); ++it)
-    iStream << (*it) << " 0 R ";
+  for (auto pon : iPageObjectNumbers)
+    iStream << pon.objNum << " 0 R ";
   iStream << "]\n>> endobj\n";
+  // create Name dictionary
+  int nameDict = -1;
+  if (iDests >= 0) {
+    nameDict = startObject();
+    iStream << "<</Dests " << iDests << " 0 R>> endobj\n";
+  }
+  // create PieceInfo
+  int pieceInfo = startObject();
+  iStream << "<</Ipe<</Private 1 0 R/LastModified("
+	  << props.iModified << ")>> >> endobj\n";
   // create /Catalog
   int catalogobj = startObject();
-  iStream << "<<\n/Type /Catalog\n/Pages 2 0 R\n";
+  iStream << "<<\n/Type /Catalog\n/Pages 2 0 R\n"
+	  << "/PieceInfo " << pieceInfo << " 0 R\n";
   if (props.iFullScreen)
     iStream << "/PageMode /FullScreen\n";
   if (iBookmarks >= 0) {
@@ -948,12 +1082,14 @@ void PdfWriter::createTrailer()
       iStream << "/PageMode /UseOutlines\n";
     iStream << "/Outlines " << iBookmarks << " 0 R\n";
   }
+  if (nameDict >= 0)
+    iStream << "/Names " << nameDict << " 0 R\n";
   if (iDoc->countTotalViews() > 1) {
     iStream << "/PageLabels << /Nums [ ";
     int count = 0;
     for (int page = 0; page < iDoc->countPages(); ++page) {
-      if (!iMarkedView || iDoc->page(page)->marked()) {
-	int nviews = iMarkedView ? iDoc->page(page)->countMarkedViews() :
+      if (!(iSaveFlags & SaveFlag::MarkedView) || iDoc->page(page)->marked()) {
+	int nviews = (iSaveFlags & SaveFlag::MarkedView) ? iDoc->page(page)->countMarkedViews() :
 	  iDoc->page(page)->countViews();
 	if (nviews > 1) {
 	  iStream << count << " <</S /D /P (" << (page + 1) << "-)>>";
