@@ -4,7 +4,7 @@
 /*
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (C) 1993-2016  Otfried Cheong
+    Copyright (c) 1993-2019 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -38,11 +38,15 @@
 #include <gdiplus.h>
 #else
 #include <sys/wait.h>
-#include <xlocale.h>
 #include <dirent.h>
 #endif
 #ifdef __APPLE__
 #include <mach-o/dyld.h>
+#include <xlocale.h>
+// sys/param.h triggers legacy mode for realpath, so that the
+// last component of the path does not need to exist.
+#include <sys/param.h>
+#include <sys/errno.h>
 #endif
 
 #include <cstdlib>
@@ -51,6 +55,7 @@
 #include <cstdarg>
 #include <unistd.h>
 #include <clocale>
+#include <cstring>
 
 using namespace ipe;
 
@@ -77,14 +82,36 @@ static bool initialized = false;
 static bool showDebug = false;
 static Platform::DebugHandler debugHandler = nullptr;
 
+#ifdef IPEAPPIMAGE
+static String appImageBase;
+#endif
+
 #ifdef WIN32
 static ULONG_PTR gdiplusToken = 0;
-// not yet available in MINGW RT library
-// _locale_t ipeLocale;
+_locale_t ipeLocale;
+typedef _locale_t (*LPCreateLocale)(int category, const char *locale);
+static LPCreateLocale p_create_locale = nullptr;
+typedef void (*LPFreeLocale)(_locale_t locale);
+static LPFreeLocale p_free_locale = nullptr;
+typedef double (*LPStrtodL)(const char *strSource, char **endptr, _locale_t locale);
+static LPStrtodL p_strtod_l = nullptr;
 #else
 locale_t ipeLocale;
 #endif
 
+#ifndef WIN32
+static String dotIpe()
+{
+  const char *home = getenv("HOME");
+  if (!home)
+    return String();
+  String res = String(home) + "/.ipe";
+  if (!Platform::fileExists(res) && mkdir(res.z(), 0700) != 0)
+    return String();
+  return res + "/";
+}
+#endif
+
 #ifdef WIN32
 static void readIpeConf()
 {
@@ -97,9 +124,21 @@ static void readIpeConf()
   int i = 0;
   while (i < conf.size()) {
     String line = conf.getLine(i);
-    std::vector<wchar_t> wline;
-    Platform::utf8ToWide(line.z(), wline);
-    _wputenv(&wline[0]);
+    _wputenv(line.w().data());
+  }
+}
+#else
+static void readIpeConf()
+{
+  String fname = dotIpe() + "/ipe.conf";
+  String conf = Platform::readFile(fname);
+  if (conf.empty())
+    return;
+  ipeDebug("ipe.conf = %s", conf.z());
+  int i = 0;
+  while (i < conf.size()) {
+    String line = conf.getLine(i);
+    putenv(strdup(line.z()));  // a leak, but necessary
   }
 }
 #endif
@@ -119,7 +158,8 @@ static void shutdownIpelib()
 {
 #ifdef WIN32
   Gdiplus::GdiplusShutdown(gdiplusToken);
-  // _free_locale(ipeLocale);
+  if (p_create_locale != nullptr)
+    p_free_locale(ipeLocale);
 #else
   freelocale(ipeLocale);
 #endif
@@ -141,19 +181,33 @@ void Platform::initLib(int version)
   if (initialized)
     return;
   initialized = true;
+  readIpeConf();
   showDebug = false;
-  if (getenv("IPEDEBUG") != 0) {
+  if (getenv("IPEDEBUG")) {
     showDebug = true;
     fprintf(stderr, "Debug messages enabled\n");
   }
   debugHandler = debugHandlerImpl;
+#ifdef IPEAPPIMAGE
+  const char *cwd = getcwd(nullptr, 0);
+  if (cwd)
+    appImageBase = cwd;
+  else
+    appImageBase = ".";
+#endif
 #ifdef WIN32
-  readIpeConf();
-  // ipeLocale = ::_create_locale(LC_NUMERIC, "C");
+  HMODULE hDll = LoadLibraryA("msvcrt.dll");
+  if (hDll) {
+    p_create_locale = (LPCreateLocale) GetProcAddress(hDll, "_create_locale");
+    p_free_locale = (LPFreeLocale) GetProcAddress(hDll, "_free_locale");
+    p_strtod_l = (LPStrtodL) GetProcAddress(hDll, "_strtod_l");
+  }
+  if (p_create_locale != nullptr)
+    ipeLocale = p_create_locale(LC_NUMERIC, "C");
   Gdiplus::GdiplusStartupInput gdiplusStartupInput;
-  Gdiplus::GdiplusStartup(&gdiplusToken, &gdiplusStartupInput, 0);
+  Gdiplus::GdiplusStartup(&gdiplusToken, &gdiplusStartupInput, nullptr);
 #else
-  ipeLocale = newlocale(LC_NUMERIC_MASK, "C", NULL);
+  ipeLocale = newlocale(LC_NUMERIC_MASK, "C", nullptr);
 #endif
   atexit(shutdownIpelib);
 #ifndef WIN32
@@ -196,10 +250,8 @@ void ipeDebug(const char *msg, ...) noexcept
 String Platform::currentDirectory()
 {
 #ifdef WIN32
-  wchar_t *buffer;
-  if ((buffer = _wgetcwd(0, 0)) == 0)
-    return String();
-  return Platform::wideToUtf8(buffer);
+  wchar_t *buffer = _wgetcwd(nullptr, 0);
+  return String(buffer);
 #else
   char buffer[1024];
   if (getcwd(buffer, 1024) != buffer)
@@ -208,22 +260,40 @@ String Platform::currentDirectory()
 #endif
 }
 
+//! Returns drive on which Ipe executable exists.
+/*! On Linux and OSX, returns empty string. */
+String Platform::ipeDrive()
+{
+#ifdef WIN32
+  String fname = ipeDir("", nullptr);
+  if (fname.size() > 2 && fname[1] == ':')
+    return fname.left(2);
+#endif
+  return String();
+}
+
 #ifdef IPEBUNDLE
 String Platform::ipeDir(const char *suffix, const char *fname)
 {
+#ifdef IPESNAPCRAFT
+  const char *p = getenv("SNAP");
+  String exe = p ? p : "/snap/ipe/current";
+  exe += "/ipe/";
+#else
+#ifdef IPEAPPIMAGE
+  String exe = appImageBase + "/ipe/";
+#else
 #ifdef WIN32
   wchar_t exename[OFS_MAXPATHNAME];
-  GetModuleFileNameW(0, exename, OFS_MAXPATHNAME);
-  String exe = Platform::wideToUtf8(exename);
+  GetModuleFileNameW(nullptr, exename, OFS_MAXPATHNAME);
+  String exe(exename);
 #else
-  char path[PATH_MAX];
+  char path[PATH_MAX], rpath[PATH_MAX];
   uint32_t size = sizeof(path);
   String exe;
-  if (_NSGetExecutablePath(path, &size) == 0) {
-    char *rp = realpath(path, NULL);
-    exe = String(rp);
-    free(rp);
-  } else
+  if (_NSGetExecutablePath(path, &size) == 0 && realpath(path, rpath) != nullptr)
+    exe = String(rpath);
+  else
     ipeDebug("ipeDir: buffer too small; need size %u", size);
 #endif
   int i = exe.rfind(IPESEP);
@@ -241,6 +311,8 @@ String Platform::ipeDir(const char *suffix, const char *fname)
     exe += "/Resources/";
 #else
   exe += IPESEP;
+#endif
+#endif
 #endif
   exe += suffix;
   if (fname) {
@@ -251,19 +323,6 @@ String Platform::ipeDir(const char *suffix, const char *fname)
 }
 #endif
 
-#ifndef WIN32
-static String dotIpe()
-{
-  const char *home = getenv("HOME");
-  if (!home)
-    return String();
-  String res = String(home) + "/.ipe";
-  if (!Platform::fileExists(res) && mkdir(res.z(), 0700) != 0)
-    return String();
-  return res + "/";
-}
-#endif
-
 //! Return path for the directory containing pdflatex and xelatex.
 /*! If empty means look on PATH. */
 String Platform::latexPath()
@@ -272,7 +331,9 @@ String Platform::latexPath()
 #ifdef WIN32
   const wchar_t *p = _wgetenv(L"IPELATEXPATH");
   if (p)
-    result = Platform::wideToUtf8(p);
+    result = String(p);
+  if (result.left(4) == "ipe:")
+    result = ipeDrive() + result.substr(4);
 #else
   char *p = getenv("IPELATEXPATH");
   if (p)
@@ -292,16 +353,18 @@ String Platform::latexDirectory()
   String latexDir;
   const wchar_t *p = _wgetenv(L"IPELATEXDIR");
   if (p)
-    latexDir = Platform::wideToUtf8(p);
+    latexDir = String(p);
+  if (latexDir.left(4) == "ipe:")
+    latexDir = ipeDrive() + latexDir.substr(4);
   else {
     wchar_t szPath[MAX_PATH];
-    if (SUCCEEDED(SHGetFolderPathW(NULL, CSIDL_LOCAL_APPDATA,
-				   NULL, 0, szPath))) {
-      latexDir = Platform::wideToUtf8(szPath) + "\\ipe";
+    if (SUCCEEDED(SHGetFolderPathW(nullptr, CSIDL_LOCAL_APPDATA,
+				   nullptr, 0, szPath))) {
+      latexDir = String(szPath) + "\\ipe";
     } else {
       p = _wgetenv(L"LOCALAPPDATA");
       if (p)
-	latexDir = Platform::wideToUtf8(p) + "\\ipe";
+	latexDir = String(p) + "\\ipe";
       else
 	latexDir = ipeDir("latexrun");
     }
@@ -335,37 +398,54 @@ String Platform::latexDirectory()
 bool Platform::fileExists(String fname)
 {
 #ifdef WIN32
-  std::vector<wchar_t> wname;
-  utf8ToWide(fname.z(), wname);
-  return (_waccess(&wname[0], F_OK) == 0);
+  return (_waccess(fname.w().data(), F_OK) == 0);
 #else
   return (access(fname.z(), F_OK) == 0);
 #endif
 }
 
+//! Convert relative filename to absolute.
+/*! This also works when the filename does not exist, or at least it tries. */
+String Platform::realPath(String fname)
+{
+#ifdef WIN32
+  wchar_t wresult[MAX_PATH];
+  // this function works also when fname does not exist
+  GetFullPathNameW(fname.w().data(), MAX_PATH, wresult, nullptr);
+  return String(wresult);
+#else
+  char rpath[PATH_MAX];
+  if (realpath(fname.z(), rpath))
+    return String(rpath);
+  if (errno != ENOENT || fname.left(1) == "/")
+    return fname;  // not much we can do
+  if (realpath(".", rpath) == nullptr)
+    return fname;  // nothing we can do
+  return String(rpath) + "/" + fname;
+#endif
+}
+
 //! List all files in directory
 /*! Return true if successful, false on error. */
 bool Platform::listDirectory(String path, std::vector<String> &files)
 {
 #ifdef WIN32
   String pattern = path + "\\*";
-  std::vector<wchar_t> wname;
-  utf8ToWide(pattern.z(), wname);
   struct _wfinddata_t info;
-  intptr_t h = _wfindfirst(&wname[0], &info);
+  intptr_t h = _wfindfirst(pattern.w().data(), &info);
   if (h == -1L)
     return false;
-  files.push_back(wideToUtf8(info.name));
+  files.push_back(String(info.name));
   while (_wfindnext(h, &info) == 0)
-    files.push_back(wideToUtf8(info.name));
+    files.push_back(String(info.name));
   _findclose(h);
   return true;
 #else
   DIR *dir = opendir(path.z());
-  if (dir == 0)
+  if (dir == nullptr)
     return false;
   struct dirent *entry = readdir(dir);
-  while (entry != 0) {
+  while (entry != nullptr) {
     String s(entry->d_name);
     if (s != "." && s != "..")
       files.push_back(s);
@@ -399,8 +479,13 @@ int Platform::runLatex(String dir, LatexType engine) noexcept
     "xelatex" : (engine == LatexType::Luatex) ?
     "lualatex" : "pdflatex";
 #ifdef WIN32
-  String bat;
+  if (getenv("IPETEXFORMAT")) {
+    latex = (engine == LatexType::Xetex) ?
+      "xetex ^&latex" : (engine == LatexType::Luatex) ?
+      "luatex ^&latex" : "pdftex ^&pdflatex";
+  }
 
+  String bat;
   if (dir.size() > 2 && dir[1] == ':') {
     bat += dir.substr(0, 2);
     bat += "\r\n";
@@ -420,10 +505,9 @@ int Platform::runLatex(String dir, LatexType engine) noexcept
 
   // CMD.EXE input needs to be encoded in "OEM codepage",
   // which can be different from "Windows codepage"
-  std::vector<wchar_t> wbat;
-  utf8ToWide(bat.z(), wbat);
+  std::wstring wbat = bat.w();
   Buffer oemBat(2 * wbat.size() + 1);
-  CharToOemW(&wbat[0], oemBat.data());
+  CharToOemW(wbat.data(), oemBat.data());
 
   String s = dir + "runlatex.bat";
   std::FILE *f = Platform::fopen(s.z(), "wb");
@@ -441,12 +525,12 @@ int Platform::runLatex(String dir, LatexType engine) noexcept
   startupInfo.cb = sizeof(startupInfo);
 
   // Call the executable program
-  s = String("cmd /c call \"") + dir + String("runlatex.bat\"");
-  utf8ToWide(s.z(), wbat);
+  String cmd = String("cmd /c call \"") + dir + String("runlatex.bat\"");
+  std::wstring wcmd = cmd.w();
 
-  int result = CreateProcessW(NULL, &wbat[0], NULL, NULL, FALSE,
+  int result = CreateProcessW(nullptr, wcmd.data(), nullptr, nullptr, FALSE,
 			      NORMAL_PRIORITY_CLASS|CREATE_NO_WINDOW,
-			      NULL, NULL, &startupInfo, &processInformation);
+			      nullptr, nullptr, &startupInfo, &processInformation);
   if (result == 0)
     return -1;  // failure to create process
 
@@ -463,6 +547,11 @@ int Platform::runLatex(String dir, LatexType engine) noexcept
     Sleep(Lex(wine).getInt());
   return 0;
 #else
+  if (getenv("IPETEXFORMAT")) {
+    latex = (engine == LatexType::Xetex) ?
+      "xetex \\&latex" : (engine == LatexType::Luatex) ?
+      "luatex \\&latex" : "pdftex \\&pdflatex";
+  }
   String s("cd \"");
   s += dir;
   s += "\"; rm -f ipetemp.log; ";
@@ -482,33 +571,38 @@ int Platform::runLatex(String dir, LatexType engine) noexcept
 #ifdef WIN32
 FILE *Platform::fopen(const char *fname, const char *mode)
 {
-  std::vector<wchar_t> wname;
-  utf8ToWide(fname, wname);
-  std::vector<wchar_t> wmode;
-  utf8ToWide(mode, wmode);
-  return _wfopen(&wname[0], &wmode[0]);
+  return _wfopen(String(fname).w().data(), String(mode).w().data());
 }
 
 int Platform::mkdir(const char *dname)
 {
-  std::vector<wchar_t> wname;
-  utf8ToWide(dname, wname);
-  return _wmkdir(&wname[0]);
+  return _wmkdir(String(dname).w().data());
 }
 
-String Platform::wideToUtf8(const wchar_t *wbuf)
+//! Return a wide string including a terminating zero character.
+std::wstring String::w() const noexcept
 {
-  int rm = WideCharToMultiByte(CP_UTF8, 0, wbuf, -1, NULL, 0, NULL, NULL);
-  char multi[rm];
-  WideCharToMultiByte(CP_UTF8, 0, wbuf, -1, multi, rm, NULL, NULL);
-  return String(multi);
+  if (empty())
+    return L"\0";
+  int rw = MultiByteToWideChar(CP_UTF8, 0, data(), size(), nullptr, 0);
+  std::wstring result(rw + 1, wchar_t(0));
+  MultiByteToWideChar(CP_UTF8, 0, data(), size(), &result[0], rw);
+  return result;
 }
 
-void Platform::utf8ToWide(const char *utf8, std::vector<wchar_t> &wbuf)
+String::String(const wchar_t *wbuf)
 {
-  int rw = MultiByteToWideChar(CP_UTF8, 0, utf8, -1, NULL, 0);
-  wbuf.resize(rw);
-  MultiByteToWideChar(CP_UTF8, 0, utf8, -1, &wbuf[0], rw);
+  if (!wbuf) {
+    iImp = emptyString();
+  } else {
+    int rm = WideCharToMultiByte(CP_UTF8, 0, wbuf, -1, nullptr, 0, nullptr, nullptr);
+    iImp = new Imp;
+    iImp->iRefCount = 1;
+    iImp->iSize = rm - 1; // rm includes the trailing zero
+    iImp->iCapacity = (rm + 32) & ~15;
+    iImp->iData = new char[iImp->iCapacity];
+    WideCharToMultiByte(CP_UTF8, 0, wbuf, -1, iImp->iData, rm, nullptr, nullptr);
+  }
 }
 #endif
 
@@ -517,10 +611,12 @@ void Platform::utf8ToWide(const char *utf8, std::vector<wchar_t> &wbuf)
 double Platform::toDouble(String s)
 {
 #ifdef WIN32
-  return strtod(s.z(), 0);
-  // return _strtod_l(s.z(), 0, ipeLocale);
+  if (p_create_locale != nullptr)
+    return _strtod_l(s.z(), nullptr, ipeLocale);
+  else
+    return strtod(s.z(), nullptr);
 #else
-  return strtod_l(s.z(), 0, ipeLocale);
+  return strtod_l(s.z(), nullptr, ipeLocale);
 #endif
 }
 
@@ -529,7 +625,6 @@ void ipeAssertionFailed(const char *file, int line, const char *assertion)
   fprintf(stderr, "Assertion failed on line #%d (%s): '%s'\n",
 	  line, file, assertion);
   abort();
-  // exit(-1);
 }
 
 // --------------------------------------------------------------------
