@@ -4,7 +4,7 @@
 /*
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (C) 1993-2016  Otfried Cheong
+    Copyright (c) 1993-2019 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -48,7 +48,7 @@ using namespace ipe;
 //! Default constructor
 Symbol::Symbol()
 {
-  iObject = 0;
+  iObject = nullptr;
   iXForm = false;
   iTransformations = ETransformationsAffine;
 }
@@ -64,7 +64,7 @@ Symbol::Symbol(Object *object)
 //! Copy constructor.
 Symbol::Symbol(const Symbol &rhs)
 {
-  iObject = rhs.iObject ? rhs.iObject->clone() : 0;
+  iObject = rhs.iObject ? rhs.iObject->clone() : nullptr;
   iXForm = rhs.iXForm;
   iTransformations = rhs.iTransformations;
 }
@@ -74,7 +74,7 @@ Symbol &Symbol::operator=(const Symbol &rhs)
 {
   if (this != &rhs) {
     delete iObject;
-    iObject = rhs.iObject ? rhs.iObject->clone() : 0;
+    iObject = rhs.iObject ? rhs.iObject->clone() : nullptr;
     iXForm = rhs.iXForm;
     iTransformations = rhs.iTransformations;
   }
@@ -134,7 +134,7 @@ void StyleSheet::setLayout(const Layout &layout)
 const Layout *StyleSheet::layout() const
 {
   if (iLayout.isNull())
-    return 0;
+    return nullptr;
   else
     return &iLayout;
 }
@@ -143,7 +143,7 @@ const Layout *StyleSheet::layout() const
 const TextPadding *StyleSheet::textPadding() const
 {
   if (iTextPadding.iLeft < 0)
-    return 0;
+    return nullptr;
   else
     return &iTextPadding;
 }
@@ -166,7 +166,7 @@ const StyleSheet::TitleStyle *StyleSheet::titleStyle() const
   if (iTitleStyle.iDefined)
     return &iTitleStyle;
   else
-    return 0;
+    return nullptr;
 }
 
 //! Set style of page numbering.
@@ -176,12 +176,12 @@ void StyleSheet::setPageNumberStyle(const PageNumberStyle &pns)
 }
 
 //! Return page number style.
-const StyleSheet::PageNumberStyle *StyleSheet::pageNumberStyle()
+const StyleSheet::PageNumberStyle *StyleSheet::pageNumberStyle() const
 {
   if (iPageNumberStyle.iDefined)
     return &iPageNumberStyle;
   else
-    return 0;
+    return nullptr;
 }
 
 //! Add gradient to this style sheet.
@@ -195,12 +195,12 @@ void StyleSheet::addGradient(Attribute name, const Gradient &s)
 const Gradient *StyleSheet::findGradient(Attribute sym) const
 {
   if (!sym.isSymbolic())
-    return 0;
+    return nullptr;
   GradientMap::const_iterator it = iGradients.find(sym.index());
   if (it != iGradients.end())
     return &it->second;
   else
-    return 0;
+    return nullptr;
 }
 
 //! Add tiling to this style sheet.
@@ -214,12 +214,12 @@ void StyleSheet::addTiling(Attribute name, const Tiling &s)
 const Tiling *StyleSheet::findTiling(Attribute sym) const
 {
   if (!sym.isSymbolic())
-    return 0;
+    return nullptr;
   TilingMap::const_iterator it = iTilings.find(sym.index());
   if (it != iTilings.end())
     return &it->second;
   else
-    return 0;
+    return nullptr;
 }
 
 void StyleSheet::addEffect(Attribute name, const Effect &e)
@@ -231,12 +231,12 @@ void StyleSheet::addEffect(Attribute name, const Effect &e)
 const Effect *StyleSheet::findEffect(Attribute sym) const
 {
   if (!sym.isSymbolic())
-    return 0;
+    return nullptr;
   EffectMap::const_iterator it = iEffects.find(sym.index());
   if (it != iEffects.end())
     return &it->second;
   else
-    return 0;
+    return nullptr;
 }
 
 // --------------------------------------------------------------------
@@ -273,12 +273,12 @@ void StyleSheet::addSymbol(Attribute name, const Symbol &symbol)
 const Symbol *StyleSheet::findSymbol(Attribute attr) const
 {
   if (!attr.isSymbolic())
-    return 0;
+    return nullptr;
   SymbolMap::const_iterator it = iSymbols.find(attr.index());
   if (it != iSymbols.end())
     return &it->second;
   else
-    return 0;
+    return nullptr;
 }
 
 // --------------------------------------------------------------------
@@ -397,9 +397,9 @@ void StyleSheet::allNames(Kind kind, AttributeSeq &seq) const
 	seq.push_back(attr);
     }
   } else {
-    uint kindMatch = (kind << SHIFT);
+    uint32_t kindMatch = (kind << SHIFT);
     for (Map::const_iterator it = iMap.begin(); it != iMap.end(); ++it) {
-      if (uint(it->first & KINDMASK) == kindMatch) {
+      if (uint32_t(it->first & KINDMASK) == kindMatch) {
 	Attribute attr(true, (it->first & MASK));
 	if (std::find(seq.begin(), seq.end(), attr) == seq.end())
 	  seq.push_back(attr);
@@ -457,11 +457,14 @@ void StyleSheet::saveAsXml(Stream &stream, bool saveBitmaps) const
 
   for (Map::const_iterator it = iMap.begin(); it != iMap.end(); ++it) {
     int kind = (it->first >> SHIFT);
-    stream << "<" << kind_names[kind]
+    int kind1 = (kind == ELabelStyle) ? ETextStyle : kind;
+    stream << "<" << kind_names[kind1]
 	   << " name=\"" << rep->toString(it->first & MASK) << "\"";
-    if (kind == ETextStyle) {
+    if (kind1 == ETextStyle) {
       String s = it->second.string();
       int i = s.find('\0');
+      if (kind == ELabelStyle)
+	stream << " type=\"label\"";
       stream << " begin=\"" << s.substr(0, i) << "\" end=\""
 	     << s.substr(i+1) << "\"/>\n";
     } else
@@ -491,10 +494,13 @@ void StyleSheet::saveAsXml(Stream &stream, bool saveBitmaps) const
 
   }
   if (iPageNumberStyle.iDefined) {
-    stream << "<pagenumberstyle pos=\"" << iPageNumberStyle.iPos
-	   << "\" size=\"" << iPageNumberStyle.iFontSize
-	   << "\" color=\"" << iPageNumberStyle.iColor
-	   << "\"/>\n";
+    stream << "<pagenumberstyle pos=\"" << iPageNumberStyle.iPos << "\""
+	   << " color=\"" << iPageNumberStyle.iColor.string() << "\""
+	   << " size=\"" << iPageNumberStyle.iSize.string() << "\"";
+    Text::saveAlignment(stream,
+			iPageNumberStyle.iHorizontalAlignment,
+			iPageNumberStyle.iVerticalAlignment);
+    stream << ">" << iPageNumberStyle.iText << "</pagenumberstyle>\n";
   }
   if (iTitleStyle.iDefined) {
     stream << "<titlestyle pos=\"" << iTitleStyle.iPos
@@ -535,7 +541,7 @@ void StyleSheet::saveAsXml(Stream &stream, bool saveBitmaps) const
     if (!g.iMatrix.isIdentity())
       stream << " matrix=\"" << g.iMatrix << "\"";
     stream << ">\n";
-    for (uint i = 0; i < g.iStops.size(); ++i)
+    for (int i = 0; i < size(g.iStops); ++i)
       stream << " <stop offset=\"" << g.iStops[i].offset
 	     << "\" color=\"" << g.iStops[i].color << "\"/>\n";
     stream << "</gradient>\n";
@@ -586,9 +592,9 @@ Cascade::Cascade()
 
 static void destruct_sheets(std::vector<StyleSheet *> &sheets)
 {
-  for (uint i = 0; i < sheets.size(); ++i) {
+  for (int i = 0; i < size(sheets); ++i) {
     delete sheets[i];
-    sheets[i] = 0;
+    sheets[i] = nullptr;
   }
   sheets.clear();
 }
@@ -672,7 +678,7 @@ const Symbol *Cascade::findSymbol(Attribute sym) const
     const Symbol *s = iSheets[i]->findSymbol(sym);
     if (s) return s;
   }
-  return 0;
+  return nullptr;
 }
 
 const Gradient *Cascade::findGradient(Attribute sym) const
@@ -681,7 +687,7 @@ const Gradient *Cascade::findGradient(Attribute sym) const
     const Gradient *s = iSheets[i]->findGradient(sym);
     if (s) return s;
   }
-  return 0;
+  return nullptr;
 }
 
 const Tiling *Cascade::findTiling(Attribute sym) const
@@ -690,7 +696,7 @@ const Tiling *Cascade::findTiling(Attribute sym) const
     const Tiling *s = iSheets[i]->findTiling(sym);
     if (s) return s;
   }
-  return 0;
+  return nullptr;
 }
 
 const Effect *Cascade::findEffect(Attribute sym) const
@@ -699,7 +705,7 @@ const Effect *Cascade::findEffect(Attribute sym) const
     const Effect *s = iSheets[i]->findEffect(sym);
     if (s) return s;
   }
-  return 0;
+  return nullptr;
 }
 
 //! Find page layout (such as text margins).
@@ -711,7 +717,7 @@ const Layout *Cascade::findLayout() const
   }
   // must never happen
   assert(false);
-  return 0;
+  return nullptr;
 }
 
 //! Find text padding (for text bbox computation).
@@ -723,7 +729,7 @@ const TextPadding *Cascade::findTextPadding() const
   }
   // must never happen
   assert(false);
-  return 0;
+  return nullptr;
 }
 
 //! Get style of page titles (or 0 if none defined).
@@ -733,7 +739,7 @@ const StyleSheet::TitleStyle *Cascade::findTitleStyle() const
     const StyleSheet::TitleStyle *ts = iSheets[i]->titleStyle();
     if (ts) return ts;
   }
-  return 0;
+  return nullptr;
 }
 
 //! Return style of page numbering (or 0 if none defined).
@@ -743,7 +749,7 @@ const StyleSheet::PageNumberStyle *Cascade::findPageNumberStyle() const
     const StyleSheet::PageNumberStyle *pns = iSheets[i]->pageNumberStyle();
     if (pns) return pns;
   }
-  return 0;
+  return nullptr;
 }
 
 //! Return total LaTeX preamble (of the whole cascade).
