@@ -4,7 +4,7 @@
 /*
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (C) 1993-2015  Otfried Cheong
+    Copyright (c) 1993-2019 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -33,6 +33,7 @@
 #include "ipebitmap.h"
 
 #include <cerrno>
+#include <cstring>
 
 using namespace ipe;
 using namespace ipelua;
@@ -67,7 +68,7 @@ bool ipelua::is_type(lua_State *L, int ud, const char *tname)
 
 String ipelua::check_filename(lua_State *L, int index)
 {
-  return String(luaL_checkstring(L, index));
+  return String(luaL_checklstring(L, index, nullptr));
 }
 
 // --------------------------------------------------------------------
@@ -78,7 +79,7 @@ static int document_constructor(lua_State *L)
 {
   bool has_fname = (lua_gettop(L) > 0);
   Document **d = (Document **) lua_newuserdata(L, sizeof(Document *));
-  *d = 0;
+  *d = nullptr;
   luaL_getmetatable(L, "Ipe.document");
   lua_setmetatable(L, -2);
 
@@ -125,7 +126,7 @@ static int document_destruct(lua_State *L)
 {
   Document **d = check_document(L, 1);
   delete (*d);
-  *d = 0;
+  *d = nullptr;
   return 0;
 }
 
@@ -153,7 +154,7 @@ static int document_index(lua_State *L)
     int n = check_pageno(L, 2, *d);
     push_page(L, (*d)->page(n), false);
   } else {
-    const char *key = luaL_checkstring(L, 2);
+    const char *key = luaL_checklstring(L, 2, nullptr);
     if (!luaL_getmetafield(L, 1, key))
       lua_pushnil(L);
   }
@@ -193,28 +194,28 @@ static int document_pages(lua_State *L)
   return 3;
 }
 
-// "export", "nozip", "markedview", "nocolor"
-static uint check_flags(lua_State *L, int index)
+// "export", "nozip", "markedview"
+static uint32_t check_flags(lua_State *L, int index)
 {
   if (lua_isnoneornil(L, index))
     return 0;
   luaL_argcheck(L, lua_istable(L, index), index, "argument is not a table");
-  uint flags = 0;
+  uint32_t flags = 0;
   lua_getfield(L, index, "export");
   if (lua_toboolean(L, -1))
-    flags |= Document::EExport;
+    flags |= SaveFlag::Export;
   lua_pop(L, 1);
   lua_getfield(L, index, "nozip");
   if (lua_toboolean(L, -1))
-    flags |= Document::ENoZip;
+    flags |= SaveFlag::NoZip;
+  lua_pop(L, 1);
+  lua_getfield(L, index, "keepnotes");
+  if (lua_toboolean(L, -1))
+    flags |= SaveFlag::KeepNotes;
   lua_pop(L, 1);
   lua_getfield(L, index, "markedview");
   if (lua_toboolean(L, -1))
-    flags |= Document::EMarkedView;
-  lua_pop(L, 1);
-  lua_getfield(L, index, "nocolor");
-  if (lua_toboolean(L, -1))
-    flags |= Document::ENoColor;
+    flags |= SaveFlag::MarkedView;
   lua_pop(L, 1);
   return flags;
 }
@@ -223,13 +224,13 @@ static int document_save(lua_State *L)
 {
   Document **d = check_document(L, 1);
   String fname = check_filename(L, 2);
-  int format;
+  FileFormat format;
   if (lua_isnoneornil(L, 3))
     format = Document::formatFromFilename(fname);
   else
-    format = luaL_checkoption(L, 3, 0, format_name);
-  uint flags = check_flags(L, 4);
-  bool result = (*d)->save(fname.z(), Document::TFormat(format), flags);
+    format = FileFormat(luaL_checkoption(L, 3, nullptr, format_name));
+  uint32_t flags = check_flags(L, 4);
+  bool result = (*d)->save(fname.z(), format, flags);
   lua_pushboolean(L, result);
   return 1;
 }
@@ -238,7 +239,7 @@ static int document_exportPages(lua_State *L)
 {
   Document **d = check_document(L, 1);
   String fname = check_filename(L, 2);
-  uint flags = check_flags(L, 3);
+  uint32_t flags = check_flags(L, 3);
   int fromPage = check_pageno(L, 4, *d);
   int toPage = check_pageno(L, 5, *d);
   bool result = (*d)->exportPages(fname.z(), flags, fromPage, toPage);
@@ -250,16 +251,15 @@ static int document_exportView(lua_State *L)
 {
   Document **d = check_document(L, 1);
   String fname = check_filename(L, 2);
-  int format;
+  FileFormat format;
   if (lua_isnoneornil(L, 3))
     format = Document::formatFromFilename(fname);
   else
-    format = luaL_checkoption(L, 3, 0, format_name);
-  uint flags = check_flags(L, 4);
+    format = FileFormat(luaL_checkoption(L, 3, nullptr, format_name));
+  uint32_t flags = check_flags(L, 4);
   int pno = check_pageno(L, 5, *d);
   int vno = check_viewno(L, 6, (*d)->page(pno));
-  bool result = (*d)->exportView(fname.z(), Document::TFormat(format),
-				 flags, pno, vno);
+  bool result = (*d)->exportView(fname.z(), format, flags, pno, vno);
   lua_pushboolean(L, result);
   return 1;
 }
@@ -272,35 +272,6 @@ static int document_countTotalViews(lua_State *L)
   return 1;
 }
 
-// Document --> boolean
-static int document_has(lua_State *L)
-{
-  static const char * const what_names[] =
-    { "truetype", "gradients", "tilings", "transparency", 0 };
-
-  Document **d = check_document(L, 1);
-  int what = luaL_checkoption(L, 2, 0, what_names);
-  bool result = false;
-  switch (what) {
-  case 0: //
-    result = (*d)->hasTrueTypeFonts();
-    break;
-  case 1:
-    result = (*d)->hasGradients();
-    break;
-  case 2:
-    result = (*d)->hasTilings();
-    break;
-  case 3:
-    result = (*d)->hasTransparency();
-    break;
-  default:
-    break;
-  }
-  lua_pushboolean(L, result);
-  return 1;
-}
-
 static int document_sheets(lua_State *L)
 {
   Document **d = check_document(L, 1);
@@ -362,11 +333,6 @@ static int document_runLatex(lua_State *L)
       lua_pushliteral(L, "There was an error reading the Pdflatex output");
       lua_pushliteral(L, "latexoutput");
       break;
-    case Document::ErrNoIconv:
-      lua_pushliteral(L, "This document needs charset conversion to run "
-		      "Pdflatex, but Ipe is compiled without this feature");
-      lua_pushliteral(L, "noiconv");
-      break;
     }
   }
   push_string(L, log);
@@ -379,7 +345,7 @@ static int document_checkStyle(lua_State *L)
   AttributeSeq seq;
   (*d)->checkStyle(seq);
   lua_createtable(L, 0, seq.size());
-  for (uint i = 0; i < seq.size(); ++i) {
+  for (int i = 0; i < size(seq); ++i) {
     push_attribute(L, seq[i]);
     lua_rawseti(L, -2, i+1);
   }
@@ -422,6 +388,9 @@ static int document_remove(lua_State *L)
   return 1;
 }
 
+static const char * const tex_engine_names[] = {
+  "default", "pdftex", "xetex", "luatex" };
+
 static int document_properties(lua_State *L)
 {
   Document **d = check_document(L, 1);
@@ -447,6 +416,8 @@ static int document_properties(lua_State *L)
   lua_setfield(L, -2, "fullscreen");
   lua_pushboolean(L, prop.iNumberPages);
   lua_setfield(L, -2, "numberpages");
+  lua_pushstring(L, tex_engine_names[int(prop.iTexEngine)]);
+  lua_setfield(L, -2, "tex");
   return 1;
 }
 
@@ -462,7 +433,7 @@ static void propString(lua_State *L, const char *name, String &str)
 {
   lua_getfield(L, 2, name);
   if (lua_isstring(L, -1))
-    str = lua_tostring(L, -1);
+    str = lua_tolstring(L, -1, nullptr);
   lua_pop(L, 1);
 }
 
@@ -482,6 +453,12 @@ static int document_setProperties(lua_State *L)
   propString(L, "created", prop.iCreated);
   propString(L, "modified", prop.iModified);
   propString(L, "creator", prop.iCreator);
+  String tex;
+  propString(L, "tex", tex);
+  for (int i = 0; i < 4; ++i) {
+    if (!strcmp(tex.z(), tex_engine_names[i]))
+      prop.iTexEngine = LatexType(i);
+  }
   (*d)->setProperties(prop);
   return 0;
 }
@@ -504,12 +481,11 @@ static const struct luaL_Reg document_methods[] = {
   { "countTotalViews", document_countTotalViews },
   { "sheets", document_sheets },
   { "replaceSheets", document_replaceSheets },
-  { "has", document_has },
   { "runLatex", document_runLatex },
   { "checkStyle", document_checkStyle },
   { "properties", document_properties },
   { "setProperties", document_setProperties },
-  { NULL, NULL }
+  { nullptr, nullptr },
 };
 
 // --------------------------------------------------------------------
@@ -521,9 +497,9 @@ static int file_format(lua_State *L)
   if (!fd)
     luaL_error(L, "fopen error: %s", strerror(errno));
   FileSource source(fd);
-  int format = Document::fileFormat(source);
+  FileFormat format = Document::fileFormat(source);
   fclose(fd);
-  lua_pushstring(L, format_name[format]);
+  lua_pushstring(L, format_name[int(format)]);
   return 1;
 }
 
@@ -558,7 +534,7 @@ static int ipe_splinetobeziers(lua_State *L)
   else
     Bezier::spline(v.size(), &v[0], result);
   lua_createtable(L, result.size(), 0);
-  for (uint i = 0; i < result.size(); ++i) {
+  for (int i = 0; i < size(result); ++i) {
     lua_createtable(L, 4, 1);
     lua_pushliteral(L, "spline");
     lua_setfield(L, -2, "type");
@@ -581,19 +557,71 @@ static int ipe_fileExists(lua_State *L)
   return 1;
 }
 
+static int ipe_realpath(lua_State *L)
+{
+  String s = check_filename(L, 1);
+  push_string(L, Platform::realPath(s));
+  return 1;
+}
+
+static int ipe_directory(lua_State *L)
+{
+  const char *path = luaL_checklstring(L, 1, nullptr);
+  std::vector<String> files;
+  if (!Platform::listDirectory(path, files))
+    luaL_error(L, "cannot list directory '%s'", path);
+  lua_createtable(L, 0, files.size());
+  for (int i = 0; i < size(files); ++i) {
+    push_string(L, files[i]);
+    lua_rawseti(L, -2, i+1);
+  }
+  return 1;
+}
+
+#define l_checkmode(mode) \
+	(*mode != '\0' && strchr("rwa", *(mode++)) != nullptr &&	\
+	(*mode != '+' || ++mode) &&  /* skip if char is '+' */	\
+	(*mode != 'b' || ++mode) &&  /* skip if char is 'b' */	\
+	(*mode == '\0'))
+
+static int ipe_fclose (lua_State *L) {
+  luaL_Stream *p = (luaL_Stream *) luaL_checkudata(L, 1, LUA_FILEHANDLE);
+  int res = fclose(p->f);
+  return luaL_fileresult(L, (res == 0), nullptr);
+}
+
+// open a file for reading or writing from Lua,
+// correctly handling UTF-8 filenames on Windows
+static int ipe_openFile(lua_State *L)
+{
+  const char *filename = luaL_checklstring(L, 1, nullptr);
+  const char *mode = luaL_optlstring(L, 2, "r", nullptr);
+
+  luaL_Stream *p = (luaL_Stream *)lua_newuserdata(L, sizeof(luaL_Stream));
+  p->closef = nullptr;  /* mark file handle as 'closed' */
+  luaL_setmetatable(L, LUA_FILEHANDLE);
+  p->f = nullptr;
+  p->closef = &ipe_fclose;
+
+  const char *md = mode;  /* to traverse/check mode */
+  luaL_argcheck(L, l_checkmode(md), 2, "invalid mode");
+  p->f = Platform::fopen(filename, mode);
+  return (p->f == nullptr) ? luaL_fileresult(L, 0, filename) : 1;
+}
+
 static const char * const image_format_names[] = { "png", "jpeg" };
 
 static int ipe_readImage(lua_State *L)
 {
   String s = check_filename(L, 1);
-  int fmt = luaL_checkoption(L, 2, NULL, image_format_names);
+  int fmt = luaL_checkoption(L, 2, nullptr, image_format_names);
 
   Vector dotsPerInch;
-  const char *errmsg = 0;
+  const char *errmsg = nullptr;
 
   Bitmap bmp = fmt ?
     Bitmap::readJpeg(s.z(), dotsPerInch, errmsg) :
-    Bitmap::readPNG(s.z(), true, dotsPerInch, errmsg);
+    Bitmap::readPNG(s.z(), dotsPerInch, errmsg);
 
   if (bmp.isNull()) {
     lua_pushnil(L);
@@ -650,9 +678,12 @@ static const struct luaL_Reg ipelib_functions[] = {
   { "normalizeAngle", ipe_normalizeangle },
   { "splineToBeziers", ipe_splinetobeziers },
   { "fileExists", ipe_fileExists },
+  { "realPath", ipe_realpath },
+  { "directory", ipe_directory },
+  { "openFile", ipe_openFile },
   { "readImage", ipe_readImage },
   { "Image", image_constructor },
-  { NULL, NULL }
+  { nullptr, nullptr }
 };
 
 extern "C" int luaopen_ipe(lua_State *L)
