@@ -5,7 +5,7 @@
 /*
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (C) 1993-2016  Otfried Cheong
+    Copyright (c) 1993-2019 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -43,27 +43,15 @@
 
 #define COLORICONSIZE 12
 
-inline String N2I(NSString *aStr) { return String(aStr.UTF8String); }
 inline const char *N2C(NSString *aStr) { return aStr.UTF8String; }
-inline NSString *I2N(String s) {return [NSString stringWithUTF8String:s.z()];}
 inline NSString *C2N(const char *s) {return [NSString stringWithUTF8String:s];}
-
-// --------------------------------------------------------------------
-
-CGContextRef ipeGetCGContext()
-{
-  if ([NSGraphicsContext instancesRespondToSelector:@selector(CGContext)])
-    return [[NSGraphicsContext currentContext] CGContext];
-  else
-    return (CGContextRef) [[NSGraphicsContext currentContext] graphicsPort];
-}
+inline NSString *S2N(const std::string &s) { return C2N(s.c_str()); }
 
 // --------------------------------------------------------------------
 
 class PDialog;
 
 @interface IpeDialogDelegate : NSObject <NSWindowDelegate,
-					   NSComboBoxDelegate,
 					   NSTableViewDataSource,
 					   NSTableViewDelegate>
 
@@ -83,6 +71,7 @@ void addToLayout(NSView *view, NSView *subview)
 
 id layoutGuide(NSView *owner)
 {
+#if MAC_OS_X_VERSION_MAX_ALLOWED >= 101100
   if ([owner respondsToSelector:@selector(addLayoutGuide:)]) {
 #pragma clang diagnostic push
 #pragma clang diagnostic ignored "-Wdeprecated-declarations"
@@ -90,11 +79,11 @@ id layoutGuide(NSView *owner)
     [owner addLayoutGuide:g];
 #pragma clang diagnostic pop
     return g;
-  } else {
-    NSView *g = [[NSView alloc] initWithFrame:NSZeroRect];
-    addToLayout(owner, g);
-    return g;
   }
+#endif
+  NSView *g = [[NSView alloc] initWithFrame:NSZeroRect];
+  addToLayout(owner, g);
+  return g;
 }
 
 static NSLayoutAttribute layoutAttribute(char ch)
@@ -138,40 +127,44 @@ static NSView *owner(id a)
 {
   if ([a respondsToSelector:@selector(superview)])
     return [a superview];
+#if MAC_OS_X_VERSION_MAX_ALLOWED >= 101100
   if ([a respondsToSelector:@selector(owningView)])
     return [a owningView];  // is a layout guide
+#endif
   return nil;
 }
 
 void activateConstraint(NSLayoutConstraint *c, BOOL active)
 {
+#if MAC_OS_X_VERSION_MAX_ALLOWED >= 101000
   if ([c respondsToSelector:@selector(setActive:)]) {
 #pragma clang diagnostic push
 #pragma clang diagnostic ignored "-Wdeprecated-declarations"
     c.active = active;
 #pragma clang diagnostic pop
-  } else {
-    // need to figure out lowest common ancestor
-    id a = c.firstItem;
-    id b = c.secondItem;
-    NSView *lca = nil;
-    NSView *ao = owner(a);
-    NSView *bo = owner(b);
-    if (b == nil)
-      lca = a;
-    else if (ao == b)
-      lca = b;
-    else if (bo == a)
-      lca = a;
-    else {
-      assert(ao == bo);
-      lca = ao;
-    }
-    if (active)
-      [lca addConstraint:c];
-    else
-      [lca removeConstraint:c];
+    return;
   }
+#endif
+  // need to figure out lowest common ancestor
+  id a = c.firstItem;
+  id b = c.secondItem;
+  NSView *lca = nil;
+  NSView *ao = owner(a);
+  NSView *bo = owner(b);
+  if (b == nil)
+    lca = a;
+  else if (ao == b)
+    lca = b;
+  else if (bo == a)
+    lca = a;
+  else {
+    assert(ao == bo);
+    lca = ao;
+  }
+  if (active)
+    [lca addConstraint:c];
+  else
+    [lca removeConstraint:c];
 }
 
 NSLayoutConstraint *layout(id a, id b, const char *rel, double gap,
@@ -201,7 +194,7 @@ NSString *ipeui_set_mnemonic(NSString *title, NSButton *button)
 {
   unichar mnemonic = 0;
   NSMutableString *result = [NSMutableString stringWithCapacity:[title length]];
-  int i = 0;
+  size_t i = 0;
   while (i < [title length]) {
     unichar ch = [title characterAtIndex:i];
     if (ch != '&') {
@@ -245,9 +238,9 @@ public:
   virtual void enableItem(int idx, bool value);
   virtual void acceptDialog(lua_State *L);
 private:
-  void fillComboBox(NSComboBox *cb, int idx);
-  void setTextView(NSTextView *tv, String s);
-  String getTextView(NSTextView *tv);
+  void fillComboBox(NSPopUpButton *cb, int idx);
+  void setTextView(NSTextView *tv, const std::string &s);
+  std::string getTextView(NSTextView *tv);
   void layoutControls();
 
 private:
@@ -271,11 +264,6 @@ private:
   return YES;
 }
 
-- (void) comboBoxSelectionDidChange:(NSNotification *) notification
-{
-  self.dialog->itemAction([[notification object] tag]);
-}
-
 - (void) tableViewSelectionDidChange:(NSNotification *) notification
 {
   self.dialog->itemAction([[notification object] tag]);
@@ -351,7 +339,7 @@ int PDialog::numberOfRows(int idx)
 
 NSString *PDialog::row(int idx, int row)
 {
-  return I2N(iElements[idx].items[row]);
+  return S2N(iElements[idx].items[row]);
 }
 
 void PDialog::setMapped(lua_State *L, int idx)
@@ -361,7 +349,7 @@ void PDialog::setMapped(lua_State *L, int idx)
   switch (m.type) {
   case ELabel:
   case EInput:
-    [((NSTextField *) ctrl) setStringValue:I2N(m.text)];
+    [((NSTextField *) ctrl) setStringValue:S2N(m.text)];
     break;
   case ETextEdit:
     setTextView((NSTextView *) ctrl, m.text);
@@ -378,7 +366,7 @@ void PDialog::setMapped(lua_State *L, int idx)
     break;
   case ECombo:
     {
-      NSComboBox *b = (NSComboBox *) ctrl;
+      NSPopUpButton *b = (NSPopUpButton *) ctrl;
       if (lua_istable(L, 3))
 	fillComboBox(b, idx);
       [b selectItemAtIndex:m.value];
@@ -396,7 +384,7 @@ void PDialog::retrieveValues()
     NSView *ctrl = iViews[i];
     switch (m.type) {
     case EInput:
-      m.text = N2I([((NSTextField *) ctrl) stringValue]);
+      m.text = N2C([((NSTextField *) ctrl) stringValue]);
       break;
     case ETextEdit:
       m.text = getTextView((NSTextView *) ctrl);
@@ -405,7 +393,7 @@ void PDialog::retrieveValues()
       m.value = [((NSTableView *) ctrl) selectedRow];
       break;
     case ECombo:
-      m.value = [((NSComboBox *) ctrl) indexOfSelectedItem];
+      m.value = [((NSPopUpButton *) ctrl) indexOfSelectedItem];
       break;
     case ECheckBox:
       m.value = [((NSButton *) ctrl) intValue];
@@ -422,23 +410,23 @@ void PDialog::enableItem(int idx, bool value)
     [((NSControl *) iViews[idx]) setEnabled:value];
 }
 
-void PDialog::fillComboBox(NSComboBox *cb, int idx)
+void PDialog::fillComboBox(NSPopUpButton *cb, int idx)
 {
   SElement &m = iElements[idx];
   [cb removeAllItems];
   for (int k = 0; k < int(m.items.size()); ++k)
-    [cb addItemWithObjectValue:I2N(m.items[k])];
+    [cb addItemWithTitle:S2N(m.items[k])];
 }
 
-void PDialog::setTextView(NSTextView *tv, String s)
+void PDialog::setTextView(NSTextView *tv, const std::string &s)
 {
-  NSAttributedString *n = [[NSAttributedString alloc] initWithString:I2N(s)];
+  NSAttributedString *n = [[NSAttributedString alloc] initWithString:S2N(s)];
   [[tv textStorage] setAttributedString:n];
 }
 
-String PDialog::getTextView(NSTextView *tv)
+std::string PDialog::getTextView(NSTextView *tv)
 {
-  return N2I([[tv textStorage] string]);
+  return std::string(N2C([[tv textStorage] string]));
 }
 
 // --------------------------------------------------------------------
@@ -480,7 +468,7 @@ void PDialog::layoutControls()
   // layout buttons
   NSView *lastButton = nil;
   int bcount = 0;
-  for (int i = 0; i < iElements.size(); ++i) {
+  for (size_t i = 0; i < iElements.size(); ++i) {
     SElement &m = iElements[i];
     if (m.row >= 0)
       continue;
@@ -502,9 +490,17 @@ void PDialog::layoutControls()
     layout(lastButton, content, "l=l", buttonGap);
   else if (bcount)
     layout(lastButton, content, "l>l", buttonGap);
+  else
+    // no button added using "addButton".  Probably an old ipelet.
+    layout(content, rows[iNoRows-1], "b=b", gap);
+
+  while (int(iColStretch.size()) < iNoCols)
+    iColStretch.push_back(0);
+  while (int(iRowStretch.size()) < iNoRows)
+    iRowStretch.push_back(0);
 
   // layout rows and columns
-  for (int i = 0; i < iElements.size(); ++i) {
+  for (size_t i = 0; i < iElements.size(); ++i) {
     SElement &m = iElements[i];
     if (m.row < 0)
       continue;
@@ -518,11 +514,9 @@ void PDialog::layoutControls()
       layout(rows[m.row + m.rowspan - 1], w, "b=b");
     layout(w, cols[m.col], "l=l");
     layout(w, cols[m.col + m.colspan - 1], "r=r");
+    if (m.type == EInput || m.type == ETextEdit)
+      layout(w, nil, "w>0", 100);
     // does it have stretch?
-    while (int(iColStretch.size()) < iNoCols)
-      iColStretch.push_back(0);
-    while (int(iRowStretch.size()) < iNoRows)
-      iRowStretch.push_back(0);
     BOOL rowStretch = NO;
     for (int r = m.row; r < m.row + m.rowspan; ++r)
       if (iRowStretch[r] > 0)
@@ -538,6 +532,16 @@ void PDialog::layoutControls()
     [w setContentHuggingPriority:colpri
 		  forOrientation:NSLayoutConstraintOrientationHorizontal];
   }
+  // make columns with stretch 1 equally wide
+  NSView *g1 = nil;
+  for (int i = 0; i < iNoCols; ++i) {
+    if (iColStretch[i] == 1) {
+      if (g1 == nil)
+	g1 = cols[i];
+      else
+	layout(g1, cols[i], "w=w");
+    }
+  }
 }
 
 // --------------------------------------------------------------------
@@ -553,7 +557,7 @@ bool PDialog::buildAndRun(int w, int h)
 		       styleMask:style
 			 backing:NSBackingStoreBuffered
 			   defer:YES];
-  [iPanel setTitle:I2N(iCaption)];
+  [iPanel setTitle:S2N(iCaption)];
   hDialog = iPanel;
   iDelegate = [[IpeDialogDelegate alloc] init];
   [iDelegate setDialog:this];
@@ -562,6 +566,7 @@ bool PDialog::buildAndRun(int w, int h)
   iViews = [NSMutableArray arrayWithCapacity:iElements.size()];
 
   NSView *content = [iPanel contentView];
+  NSView *focusCtrl = nil;
 
   for (int i = 0; i < int(iElements.size()); ++i) {
     SElement &m = iElements[i];
@@ -571,18 +576,20 @@ bool PDialog::buildAndRun(int w, int h)
     if (m.row < 0) {
       NSButton *b = [[NSButton alloc] initWithFrame:NSZeroRect];
       [b setButtonType:NSMomentaryPushInButton];
-      ipeui_set_mnemonic(I2N(m.text), b);
+      ipeui_set_mnemonic(S2N(m.text), b);
       [b setImagePosition:NSNoImage];
       [b setBezelStyle:NSRoundedBezelStyle];
       ctrl = b;
-      if (m.flags & EAccept)
-	[iPanel setDefaultButtonCell:[b cell]]; // changed rendering
+      if (m.flags & EAccept) {
+	[b setKeyEquivalent:@"\r"];
+	[b setKeyEquivalentModifierMask:NSCommandKeyMask];
+      }
     } else {
       switch (m.type) {
       case ELabel:
 	{
 	  NSTextField *t = [[NSTextField alloc] initWithFrame:NSZeroRect];
-	  t.stringValue= I2N(m.text);
+	  t.stringValue= S2N(m.text);
 	  t.bordered= NO;
 	  t.drawsBackground = NO;
 	  t.editable = NO;
@@ -593,7 +600,7 @@ bool PDialog::buildAndRun(int w, int h)
 	{
 	  NSButton *b = [[NSButton alloc] initWithFrame:NSZeroRect];
 	  [b setButtonType:NSMomentaryPushInButton];
-	  ipeui_set_mnemonic(I2N(m.text), b);
+	  ipeui_set_mnemonic(S2N(m.text), b);
 	  [b setImagePosition:NSNoImage];
 	  [b setBezelStyle:NSRoundedBezelStyle];
 	  ctrl = b;
@@ -603,7 +610,7 @@ bool PDialog::buildAndRun(int w, int h)
 	{
 	  NSButton *b = [[NSButton alloc] initWithFrame:NSZeroRect];
 	  [b setButtonType:NSSwitchButton];
-	  ipeui_set_mnemonic(I2N(m.text), b);
+	  ipeui_set_mnemonic(S2N(m.text), b);
 	  [b setState:m.value ? NSOnState : NSOffState];
 	  ctrl = b;
 	}
@@ -611,7 +618,7 @@ bool PDialog::buildAndRun(int w, int h)
       case EInput:
 	{
 	  NSTextField *t = [[NSTextField alloc] initWithFrame:NSZeroRect];
-	  [t setStringValue:I2N(m.text)];
+	  [t setStringValue:S2N(m.text)];
 	  if (m.flags & ESelectAll)
 	    [t selectText:content];
 	  ctrl = t;
@@ -626,16 +633,20 @@ bool PDialog::buildAndRun(int w, int h)
 	  tv.allowsUndo = YES;
 	  tv.continuousSpellCheckingEnabled = !!(m.flags & ESpellCheck);
 	  tv.automaticSpellingCorrectionEnabled = NO;
+	  tv.automaticQuoteSubstitutionEnabled = NO;
+	  tv.automaticTextReplacementEnabled = NO;
+	  tv.automaticDataDetectionEnabled = NO;
+	  tv.automaticDashSubstitutionEnabled = NO;
 	  setTextView(tv, m.text);
 	  view = tv;
+	  tv.textColor = [NSColor textColor];
 	}
 	break;
       case ECombo:
 	{
-	  NSComboBox *b = [[NSComboBox alloc] initWithFrame:NSZeroRect];
-	  b.editable = NO;
+	  NSPopUpButton *b = [[NSPopUpButton alloc] initWithFrame:NSZeroRect
+							pullsDown:NO];
 	  fillComboBox(b, i);
-	  [b setDelegate:iDelegate];
 	  [b selectItemAtIndex:m.value];
 	  ctrl = b;
 	}
@@ -661,18 +672,20 @@ bool PDialog::buildAndRun(int w, int h)
       }
     }
     if (ctrl) {
-      [ctrl setEnabled:!(m.flags & EDisabled)];
-      [ctrl setTag:i];
+      ctrl.enabled = !(m.flags & EDisabled);
+      ctrl.tag = i;
       if (m.type != EList) {
-	[ctrl setAction:@selector(ipeControl:)];
-	[ctrl setTarget:iDelegate];
+	ctrl.action = @selector(ipeControl:);
+	ctrl.target = iDelegate;
       }
       view = ctrl;
     }
+    if (m.flags & EFocused)
+      focusCtrl = view;
     if (scroll) {
       [view setAutoresizingMask:NSViewWidthSizable|NSViewHeightSizable];
       [scroll setDocumentView:view];
-      [scroll setHasVerticalScroller:YES];
+      scroll.hasVerticalScroller = YES;
       layout(scroll, nil, "h>0", 100.0);
       layout(scroll, nil, "w>0", 160.0);
     }
@@ -683,8 +696,19 @@ bool PDialog::buildAndRun(int w, int h)
   }
 
   layoutControls();
+  // set keyboard focus
+  if (focusCtrl)
+    [iPanel makeFirstResponder:focusCtrl];
+  // this is such a hack, but it seems to work for Ipe
+  NSMenu *editMenu = [[[NSApp mainMenu] itemAtIndex:2] submenu];
+  NSMenuItem *undoItem = [editMenu itemAtIndex:0];
+  NSMenuItem *redoItem = [editMenu itemAtIndex:1];
+  undoItem.action = @selector(undo:);
+  redoItem.action = @selector(redo:);
   int result = [NSApp runModalForWindow:iPanel];
   retrieveValues(); // for future reference
+  undoItem.action = @selector(ipeMenuAction:);
+  redoItem.action = @selector(ipeMenuAction:);
   iPanel = nil;
   return result;
 }
@@ -697,7 +721,7 @@ static int dialog_constructor(lua_State *L)
   const char *s = luaL_checkstring(L, 2);
 
   Dialog **dlg = (Dialog **) lua_newuserdata(L, sizeof(Dialog *));
-  *dlg = 0;
+  *dlg = nullptr;
   luaL_getmetatable(L, "Ipe.dialog");
   lua_setmetatable(L, -2);
   *dlg = new PDialog(L, parent, s);
@@ -805,16 +829,14 @@ int PMenu::execute(lua_State *L)
     return 0;
 }
 
-static NSImage *colorIcon(double red, double green, double blue)
+NSImage *colorIcon(double red, double green, double blue, int pixels)
 {
-  NSSize size = { COLORICONSIZE, COLORICONSIZE };
-  NSImage *icon = [NSImage imageWithSize:size
+  NSImage *icon = [NSImage imageWithSize:NSMakeSize(pixels, pixels)
 				 flipped:NO
 			  drawingHandler:^(NSRect rect) {
-      CGContextRef myContext = ipeGetCGContext();
+      CGContextRef myContext = [[NSGraphicsContext currentContext] CGContext];
       CGContextSetRGBFillColor(myContext, red, green, blue, 1.0);
-      CGContextFillRect(myContext,
-			CGRectMake(0, 0, COLORICONSIZE, COLORICONSIZE));
+      CGContextFillRect(myContext, CGRectMake(0, 0, pixels, pixels));
       return YES;
     }];
   return icon;
@@ -843,7 +865,7 @@ int PMenu::add(lua_State *L)
     if (hastable)
       luaL_argcheck(L, lua_istable(L, 5), 5,
 		    "argument is not a function or table");
-    const char *current = 0;
+    const char *current = nullptr;
     if (hascheck) {
       luaL_argcheck(L, lua_isstring(L, 6), 6,
 		    "argument is not a function or string");
@@ -895,14 +917,14 @@ int PMenu::add(lua_State *L)
 	double green = luaL_checknumber(L, -2);
 	double blue = luaL_checknumber(L, -1);
 	lua_pop(L, 3);         // pop result
-	NSImage *im = colorIcon(red, green, blue);
+	NSImage *im = colorIcon(red, green, blue, COLORICONSIZE);
 	[item setImage:im];
       }
       lua_pop(L, 2); // item, text
     }
     NSMenuItem *mitem = [[NSMenuItem alloc]
 			  initWithTitle:ipeui_set_mnemonic(C2N(title),nil)
-				 action:NULL
+				 action:nullptr
 			  keyEquivalent:@""];
     [mitem setSubmenu:sm];
     [iMenu addItem:mitem];
@@ -916,7 +938,7 @@ static int menu_constructor(lua_State *L)
 {
   // check_winid(L, 1);
   Menu **m = (Menu **) lua_newuserdata(L, sizeof(Menu *));
-  *m = 0;
+  *m = nullptr;
   luaL_getmetatable(L, "Ipe.menu");
   lua_setmetatable(L, -2);
   *m = new PMenu();
@@ -1010,7 +1032,7 @@ static int timer_constructor(lua_State *L)
   const char *method = luaL_checkstring(L, 2);
 
   Timer **t = (Timer **) lua_newuserdata(L, sizeof(Timer *));
-  *t = 0;
+  *t = nullptr;
   luaL_getmetatable(L, "Ipe.timer");
   lua_setmetatable(L, -2);
 
@@ -1029,40 +1051,130 @@ static int timer_constructor(lua_State *L)
 
 // --------------------------------------------------------------------
 
+static NSArray<NSString *> *make_filters(const char *s)
+{
+  NSMutableArray <NSString *> *exts = [NSMutableArray arrayWithCapacity:10];
+  const char *p = s;
+  while (*p) {
+    p += 2;  // skip *.
+    const char *q = p;
+    while (*q && *q != ';')
+      ++q;
+    NSString *ext = [[NSString alloc] initWithBytes:p
+					     length:q-p
+					   encoding:NSUTF8StringEncoding];
+    [exts addObject:ext];
+    if (*q == ';')
+      ++q;
+    p = q;
+  }
+  return exts;
+}
+
+@interface IpeFileDialogHelper : NSObject
+
+@property (weak) NSSavePanel *panel;
+@property NSPopUpButton *fileType;
+@property NSMutableArray *filters;
+
+- (instancetype) initFor:(NSSavePanel *) panel;
+- (void) setupWithLua:(lua_State *) L;
+- (void) changeFileType:(id) sender;
+@end
+
+@implementation IpeFileDialogHelper
+
+- (instancetype) initFor:(NSSavePanel *) panel
+{
+  self = [super init];
+  if (self) {
+    _panel = panel;
+  }
+  return self;
+}
+
+- (void) changeFileType:(id) sender
+{
+  [self setFilter:[self.fileType indexOfSelectedItem]];
+}
+
+- (void) setupWithLua:(lua_State *) L
+{
+  self.panel.message = C2N(luaL_checkstring(L, 3));
+  if (!lua_isnoneornil(L, 5)) {
+    const char *dir = luaL_checkstring(L, 5);
+    self.panel.directoryURL = [NSURL fileURLWithPath:C2N(dir) isDirectory:YES];
+  }
+  if (!lua_isnoneornil(L, 6)) {
+    auto url = [NSURL fileURLWithPath:C2N(luaL_checkstring(L, 6))];
+    self.panel.nameFieldStringValue = [url lastPathComponent];
+  }
+
+  self.panel.canCreateDirectories = YES;
+  self.panel.showsTagField = NO;
+  self.panel.canSelectHiddenExtension = YES;
+
+  self.fileType = [[NSPopUpButton alloc]
+		    initWithFrame:NSMakeRect(0., 0., 200.0, 40.0)
+			pullsDown:NO];
+
+  self.fileType.target = self;
+  self.fileType.action = @selector(changeFileType:);
+  self.panel.accessoryView = self.fileType;
+
+  if (!lua_istable(L, 4))
+    luaL_argerror(L, 4, "table expected for filters");
+  int nFilters = lua_rawlen(L, 4);
+  self.filters = [NSMutableArray arrayWithCapacity:nFilters/2];
+  for (int i = 1; i <= nFilters; ++i) {
+    lua_rawgeti(L, 4, i);
+    luaL_argcheck(L, lua_isstring(L, -1), 4, "filter entry is not a string");
+    const char *s = lua_tostring(L, -1);
+    if (i % 2 == 1)
+      [self.fileType addItemWithTitle:C2N(s)];
+    else
+      [self.filters addObject:make_filters(s)];
+    lua_pop(L, 1); // element i
+  }
+
+  int selected = 0;
+  if (!lua_isnoneornil(L, 7))
+    selected = luaL_checkinteger(L, 7) - 1;
+  [self.fileType selectItemAtIndex:selected];
+  [self setFilter:selected];
+}
+
+- (void) setFilter:(int) filterIndex
+{
+  if ([self.filters[filterIndex][0] isEqualToString:@"*"])
+    self.panel.allowedFileTypes = nil;
+  else
+    self.panel.allowedFileTypes = self.filters[filterIndex];
+}
+
+@end
+
 static int ipeui_fileDialog(lua_State *L)
 {
-  static const char * const typenames[] = { "open", "save", 0 };
+  static const char * const typenames[] = { "open", "save", nullptr };
   // not needed: NSWindow *win = check_winid(L, 1);
-  int type = luaL_checkoption(L, 2, 0, typenames);
-  const char *caption = luaL_checkstring(L, 3);
-  // Arg 4 is the filter, which is not used on Mac OS X
-  const char *dir = 0;
-  if (!lua_isnoneornil(L, 5))
-    dir = luaL_checkstring(L, 5);
-  // Arg 6 is the originally displayed filename, not used on Mac OS X
-  // Arg 7 is the selected filter, again not used on Mac OS X
-
+  int type = luaL_checkoption(L, 2, nullptr, typenames);
   if (type == 0) { // open file
     NSOpenPanel *panel = [NSOpenPanel openPanel];
-    [panel setMessage:C2N(caption)];
-    if (dir)
-      [panel setDirectoryURL:[NSURL fileURLWithPath:C2N(dir)
-			      isDirectory:YES]];
+    IpeFileDialogHelper *helper = [[IpeFileDialogHelper alloc] initFor:panel];
+    [helper setupWithLua:L];
     if ([panel runModal] == NSFileHandlingPanelOKButton) {
       lua_pushstring(L, N2C([[panel URLs][0] path]));
-      lua_pushinteger(L, 1);  // should be filter index
+      lua_pushinteger(L, [helper.fileType indexOfSelectedItem] + 1);
       return 2;
     }
   } else { // save file
     NSSavePanel *panel = [NSSavePanel savePanel];
-    [panel setMessage:C2N(caption)];
-    if (dir)
-      [panel setDirectoryURL:[NSURL fileURLWithPath:C2N(dir)
-			      isDirectory:YES]];
-    [panel setCanCreateDirectories:YES];
+    IpeFileDialogHelper *helper = [[IpeFileDialogHelper alloc] initFor:panel];
+    [helper setupWithLua:L];
     if ([panel runModal] == NSFileHandlingPanelOKButton) {
       lua_pushstring(L, N2C([[panel URL] path]));
-      lua_pushinteger(L, 1);  // should be filter index
+      lua_pushinteger(L, [helper.fileType indexOfSelectedItem] + 1);
       return 2;
     }
   }
@@ -1100,22 +1212,22 @@ static int ipeui_getColor(lua_State *L)
 static int ipeui_messageBox(lua_State *L)
 {
   static const char * const options[] =  {
-    "none", "warning", "information", "question", "critical", 0 };
+    "none", "warning", "information", "question", "critical", nullptr };
   static const char * const buttontype[] = {
     "ok", "okcancel", "yesnocancel", "discardcancel",
-    "savediscardcancel", 0 };
+    "savediscardcancel", nullptr };
 
   NSWindow *win = check_winid(L, 1);
   int type = luaL_checkoption(L, 2, "none", options);
   const char *text = luaL_checkstring(L, 3);
-  const char *details = 0;
+  const char *details = nullptr;
   if (!lua_isnoneornil(L, 4))
     details = luaL_checkstring(L, 4);
   int buttons = 0;
   if (lua_isnumber(L, 5))
     buttons = (int)luaL_checkinteger(L, 5);
   else if (!lua_isnoneornil(L, 5))
-    buttons = luaL_checkoption(L, 5, 0, buttontype);
+    buttons = luaL_checkoption(L, 5, nullptr, buttontype);
 
   (void) win;
   NSAlert *alert = [[NSAlert alloc] init];
@@ -1190,10 +1302,10 @@ static int ipeui_currentDateTime(lua_State *L)
 {
   NSDate *now = [NSDate date];
   NSCalendar *gregorian = [[NSCalendar alloc]
-			   initWithCalendarIdentifier:NSGregorianCalendar];
+			    initWithCalendarIdentifier:NSCalendarIdentifierGregorian];
   unsigned unitFlags =
-    NSYearCalendarUnit | NSMonthCalendarUnit | NSDayCalendarUnit |
-    NSHourCalendarUnit | NSMinuteCalendarUnit | NSSecondCalendarUnit;
+    NSCalendarUnitYear | NSCalendarUnitMonth | NSCalendarUnitDay |
+    NSCalendarUnitHour | NSCalendarUnitMinute | NSCalendarUnitSecond;
   NSDateComponents *st = [gregorian components:unitFlags fromDate:now];
   char buf[16];
   sprintf(buf, "%04ld%02ld%02ld%02ld%02ld%02ld",
@@ -1224,7 +1336,7 @@ static const struct luaL_Reg ipeui_functions[] = {
   { "messageBox", ipeui_messageBox },
   { "currentDateTime", ipeui_currentDateTime },
   { "startBrowser", ipeui_startBrowser },
-  { 0, 0},
+  { nullptr, nullptr }
 };
 
 // --------------------------------------------------------------------
