@@ -4,7 +4,7 @@
 /*
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (C) 1993-2016  Otfried Cheong
+    Copyright (c) 1993-2019 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -36,7 +36,7 @@
 WINID check_winid(lua_State *L, int i)
 {
   if (lua_isnil(L, i))
-    return 0;
+    return nullptr;
   WINID *w = (WINID *) luaL_checkudata(L, i, "Ipe.winid");
   return *w;
 }
@@ -58,70 +58,18 @@ static int winid_tostring(lua_State *L)
 
 static const struct luaL_Reg winid_methods[] = {
   { "__tostring", winid_tostring },
-  { 0, 0 }
+  { nullptr, nullptr }
 };
 
 // --------------------------------------------------------------------
 
-String::String()
-{
-  iImp = 0;
-}
-
-String::String(const String &rhs)
-{
-  iImp = rhs.iImp;
-  if (iImp)
-    iImp->iRefCount += 1;
-}
-
-String::String(const char *str)
-{
-  if (str) {
-    iImp = new Imp;
-    iImp->iRefCount = 1;
-    iImp->iData = strdup(str);
-  } else
-    iImp = 0;
-}
-
-String &String::operator=(const String &rhs)
-{
-  if (this != &rhs) {
-    if (iImp) {
-      iImp->iRefCount -= 1;
-      if (iImp->iRefCount == 0) {
-	free(iImp->iData);
-	delete iImp;
-      }
-    }
-    iImp = rhs.iImp;
-    if (iImp)
-      iImp->iRefCount += 1;
-  }
-  return *this;
-}
-
-String::~String()
-{
-  if (iImp) {
-    iImp->iRefCount -= 1;
-    if (iImp->iRefCount == 0) {
-      free(iImp->iData);
-      delete iImp;
-    }
-  }
-}
-
-// --------------------------------------------------------------------
-
 Dialog::Dialog(lua_State *L0, WINID parent, const char *caption)
   : iCaption(caption)
 {
   L = L0;
   iParent = parent;
   iLuaDialog = LUA_NOREF;
-  hDialog = NULL;
+  hDialog = nullptr;
   iIgnoreEscape = false;
   iNoRows = 1;
   iNoCols = 1;
@@ -143,14 +91,14 @@ void Dialog::callLua(int luaMethod)
 
   lua_rawgeti(L, LUA_REGISTRYINDEX, luaMethod);
   lua_rawgeti(L, LUA_REGISTRYINDEX, iLuaDialog);
-  lua_call(L, 1, 0);
+  luacall(L, 1, 0);
 }
 
 // name, label, action
 int Dialog::addButton(lua_State *L)
 {
   SElement m;
-  m.name = String(luaL_checkstring(L, 2));
+  m.name = std::string(luaL_checklstring(L, 2, nullptr));
   m.type = EButton;
   m.lua_method = LUA_NOREF;
   m.flags = 0;
@@ -158,9 +106,9 @@ int Dialog::addButton(lua_State *L)
   m.col = -1;
   m.rowspan = 1;
   m.colspan = 1;
-  m.text = String(luaL_checkstring(L, 3));
+  m.text = std::string(luaL_checklstring(L, 3, nullptr));
   if (lua_isstring(L, 4)) {
-    const char *s = lua_tostring(L, 4);
+    const char *s = lua_tolstring(L, 4, nullptr);
     if (!strcmp(s, "accept"))
       m.flags |= EAccept;
     else if (!strcmp(s, "reject"))
@@ -173,7 +121,7 @@ int Dialog::addButton(lua_State *L)
     m.lua_method = luaL_ref(L, LUA_REGISTRYINDEX);
   }
   m.minHeight = 16;
-  m.minWidth = 4 * m.text.len() + 8;
+  m.minWidth = 4 * m.text.length() + 8;
   if (m.minWidth < 64)
     m.minWidth = 64;
   iElements.push_back(m);
@@ -183,22 +131,24 @@ int Dialog::addButton(lua_State *L)
 int Dialog::add(lua_State *L)
 {
   static const char * const typenames[] =
-    { "button", "text", "list", "label", "combo", "checkbox", "input", 0 };
+    { "button", "text", "list", "label", "combo", "checkbox", "input", nullptr };
 
   SElement m;
-  m.name = String(luaL_checkstring(L, 2));
-  m.type = TType(luaL_checkoption(L, 3, 0, typenames));
+  m.name = checkstring(L, 2);
+  m.type = TType(luaL_checkoption(L, 3, nullptr, typenames));
   luaL_checktype(L, 4, LUA_TTABLE);
   m.lua_method = LUA_NOREF;
   m.flags = 0;
-  m.row = (int)luaL_checkinteger(L, 5) - 1;
-  m.col = (int)luaL_checkinteger(L, 6) - 1;
+  m.row = luaL_checkinteger(L, 5) - 1;
+  if (m.row < 0)
+    m.row = iNoRows + 1 + m.row;
+  m.col = luaL_checkinteger(L, 6) - 1;
   m.rowspan = 1;
   m.colspan = 1;
   if (!lua_isnoneornil(L, 7))
-    m.rowspan = (int)luaL_checkinteger(L, 7);
+    m.rowspan = luaL_checkinteger(L, 7);
   if (!lua_isnoneornil(L, 8))
-    m.colspan = (int)luaL_checkinteger(L, 8);
+    m.colspan = luaL_checkinteger(L, 8);
   if (m.row + m.rowspan > iNoRows)
     iNoRows = m.row + m.rowspan;
   if (m.col + m.colspan > iNoCols)
@@ -237,10 +187,10 @@ void Dialog::addLabel(lua_State *L, SElement &m)
 {
   lua_getfield(L, 4, "label");
   luaL_argcheck(L, lua_isstring(L, -1), 4, "no label");
-  m.text = String(lua_tostring(L, -1));
+  m.text = std::string(lua_tolstring(L, -1, nullptr));
   lua_pop(L, 1); // label
   m.minHeight = 16;
-  const char *p = m.text.z();
+  const char *p = m.text.c_str();
   int maxlen = 0;
   int curlen = 0;
   while (*p) {
@@ -261,14 +211,13 @@ void Dialog::addButtonItem(lua_State *L, SElement &m)
 {
   lua_getfield(L, 4, "label");
   luaL_argcheck(L, lua_isstring(L, -1), 4, "no button label");
-  const char *label = lua_tostring(L, -1);
-  m.text = String(label);
+  m.text = tostring(L, -1);
   lua_getfield(L, 4, "action");
   if (lua_isstring(L, -1)) {
-    const char *s = lua_tostring(L, -1);
-    if (!strcmp(s, "accept"))
+    auto s = tostring(L, -1);
+    if (s == "accept")
       m.flags |= EAccept;
-    else if (!strcmp(s, "reject"))
+    else if (s == "reject")
       m.flags |= EReject;
     else
       luaL_argerror(L, 4, "unknown action");
@@ -279,7 +228,7 @@ void Dialog::addButtonItem(lua_State *L, SElement &m)
     luaL_argerror(L, 4, "unknown action type");
   lua_pop(L, 2); // action, label
   m.minHeight = 16;
-  m.minWidth = 4 * m.text.len() + 8;
+  m.minWidth = 4 * m.text.length() + 8;
   if (m.minWidth < 64)
     m.minWidth = 64;
 }
@@ -288,7 +237,7 @@ void Dialog::addCheckbox(lua_State *L, SElement &m)
 {
   lua_getfield(L, 4, "label");
   luaL_argcheck(L, lua_isstring(L, -1), 4, "no label");
-  m.text = String(lua_tostring(L, -1));
+  m.text = tostring(L, -1);
   lua_getfield(L, 4, "action");
   if (!lua_isnil(L, -1)) {
     luaL_argcheck(L, lua_isfunction(L, -1), 4, "unknown action type");
@@ -298,7 +247,7 @@ void Dialog::addCheckbox(lua_State *L, SElement &m)
   lua_pop(L, 2); // action, label
   m.value = 0;
   m.minHeight = 16;
-  m.minWidth = 4 * m.text.len() + 32;
+  m.minWidth = 4 * m.text.length() + 32;
 }
 
 void Dialog::addInput(lua_State *L, SElement &m)
@@ -327,12 +276,12 @@ void Dialog::addTextEdit(lua_State *L, SElement &m)
     m.flags |= EFocused;
   lua_getfield(L, 4, "syntax");
   if (!lua_isnil(L, -1)) {
-    const char *s = lua_tostring(L, -1);
-    if (!strcmp(s, "logfile")) {
+    auto s = tostring(L, -1);
+    if (s == "logfile") {
       m.flags |= ELogFile;
-    } else if (!strcmp(s, "xml")) {
+    } else if (s == "xml") {
       m.flags |= EXml;
-    } else if (!strcmp(s, "latex")) {
+    } else if (s == "latex") {
       m.flags |= ELatex;
     } else
       luaL_argerror(L, 4, "unknown syntax");
@@ -352,11 +301,11 @@ void Dialog::setListItems(lua_State *L, int index, SElement &m)
   for (int i = 1; i <= no; ++i) {
     lua_rawgeti(L, index, i);
     luaL_argcheck(L, lua_isstring(L, -1), index, "items must be strings");
-    const char *item = lua_tostring(L, -1);
-    m.items.push_back(String(item));
-    int w = 4 * strlen(item) + 16;
+    auto item = tostring(L, -1);
+    int w = 4 * item.size() + 16;
     if (w > m.minWidth)
       m.minWidth = w;
+    m.items.emplace_back(item);
     lua_pop(L, 1); // item
   }
 }
@@ -377,9 +326,9 @@ void Dialog::addCombo(lua_State *L, SElement &m)
 
 int Dialog::findElement(lua_State *L, int index)
 {
-  const char *name = luaL_checkstring(L, index);
+  auto name = checkstring(L, index);
   for (int i = 0; i < int(iElements.size()); ++i) {
-    if (!strcmp(iElements[i].name.z(), name))
+    if (name == iElements[i].name)
       return i;
   }
   return luaL_argerror(L, index, "no such element in dialog");
@@ -387,8 +336,8 @@ int Dialog::findElement(lua_State *L, int index)
 
 int Dialog::set(lua_State *L)
 {
-  const char *s = luaL_checkstring(L, 2);
-  if (!strcmp(s, "ignore-escape")) {
+  auto s = checkstring(L, 2);
+  if (s == "ignore-escape") {
     iIgnoreEscape = lua_toboolean(L, 3);
     return 0;
   }
@@ -411,7 +360,7 @@ void Dialog::setUnmapped(lua_State *L, int idx)
   case ELabel:
   case ETextEdit:
   case EInput:
-    m.text = String(luaL_checkstring(L, 3));
+    m.text = checkstring(L, 3);
     break;
   case EList:
   case ECombo:
@@ -421,10 +370,9 @@ void Dialog::setUnmapped(lua_State *L, int idx)
 		    "list index out of bounds");
       m.value = n - 1;
     } else if (lua_isstring(L, 3)) {
-      const char *s = lua_tostring(L, 3);
+      auto s = tostring(L, 3);
       auto it = std::find_if(m.items.cbegin(), m.items.cend(),
-			     [s](const String &el)
-			     { return !strcmp(el.z(), s); });
+			     [&s](const std::string &el) { return s == el; });
       luaL_argcheck(L, it != m.items.end(), 3, "item not in list");
       m.value = (it - m.items.begin());
     } else {
@@ -453,7 +401,7 @@ int Dialog::get(lua_State *L)
   switch (m.type) {
   case ETextEdit:
   case EInput:
-    lua_pushstring(L, m.text.z());
+    lua_pushstring(L, m.text.c_str());
     return 1;
   case EList:
   case ECombo:
@@ -495,14 +443,14 @@ int Dialog::setEnabled(lua_State *L)
 
 int Dialog::setStretch(lua_State *L)
 {
-  static const char * const typenames[] = { "row", "column", 0 };
+  static const char * const typenames[] = { "row", "column", nullptr };
 
   while (int(iColStretch.size()) < iNoCols)
     iColStretch.push_back(0);
   while (int(iRowStretch.size()) < iNoRows)
     iRowStretch.push_back(0);
 
-  int type = luaL_checkoption(L, 2, 0, typenames);
+  int type = luaL_checkoption(L, 2, nullptr, typenames);
   int rowcol = (int)luaL_checkinteger(L, 3) - 1;
   int stretch = (int)luaL_checkinteger(L, 4);
 
@@ -532,7 +480,7 @@ static int dialog_destructor(lua_State *L)
   //fprintf(stderr, "Dialog::~Dialog()\n");
   Dialog **dlg = check_dialog(L, 1);
   delete (*dlg);
-  *dlg = 0;
+  *dlg = nullptr;
   return 0;
 }
 
@@ -547,8 +495,8 @@ static int dialog_execute(lua_State *L)
     luaL_argcheck(L, lua_isnumber(L, -1), 2, "width is not a number");
     lua_rawgeti(L, 2, 2);
     luaL_argcheck(L, lua_isnumber(L, -1), 2, "height is not a number");
-    w = lua_tointeger(L, -2);
-    h = lua_tointeger(L, -1);
+    w = lua_tointegerx(L, -2, nullptr);
+    h = lua_tointegerx(L, -1, nullptr);
     lua_pop(L, 2); // w & h
   }
   lua_pushboolean(L, (*dlg)->execute(L, w, h));
@@ -611,7 +559,7 @@ static const struct luaL_Reg dialog_methods[] = {
   { "get", dialog_get },
   { "setEnabled", dialog_setEnabled },
   { "accept", dialog_accept },
-  { 0, 0 }
+  { nullptr, nullptr }
 };
 
 // --------------------------------------------------------------------
@@ -635,7 +583,7 @@ static int menu_destructor(lua_State *L)
   //fprintf(stderr, "Menu::~Menu()\n");
   Menu **m = check_menu(L, 1);
   delete *m;
-  *m = 0;
+  *m = nullptr;
   return 0;
 }
 
@@ -658,7 +606,7 @@ static const struct luaL_Reg menu_methods[] = {
   { "__gc", menu_destructor },
   { "execute", menu_execute },
   { "add", menu_add },
-  { 0, 0 }
+  { nullptr, nullptr }
 };
 
 // --------------------------------------------------------------------
@@ -684,14 +632,14 @@ void Timer::callLua()
     lua_pop(L, 2); // pop weak table, nil
     return;
   }
-  lua_getfield(L, -1, iMethod.z());
+  lua_getfield(L, -1, iMethod.c_str());
   if (lua_isnil(L, -1)) {
     lua_pop(L, 3); // pop weak table, table, nil
     return;
   }
   lua_remove(L, -3); // remove weak table
   lua_insert(L, -2); // stack is now: method, table
-  lua_call(L, 1, 0); // call method
+  luacall(L, 1, 0); // call method
 }
 
 int Timer::setSingleShot(lua_State *L)
@@ -714,7 +662,7 @@ static int timer_destructor(lua_State *L)
   Timer **t = check_timer(L, 1);
   //fprintf(stderr, "Timer::~Timer()\n");
   delete *t;
-  *t = 0;
+  *t = nullptr;
   return 0;
 }
 
@@ -760,7 +708,7 @@ static const struct luaL_Reg timer_methods[] = {
   { "active", timer_active },
   { "setInterval", timer_setinterval },
   { "setSingleShot", timer_setsingleshot },
-  { 0, 0 }
+  { nullptr, nullptr }
 };
 
 // --------------------------------------------------------------------
