@@ -4,7 +4,7 @@
 /*
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (C) 1993-2016  Otfried Cheong
+    Copyright (c) 1993-2019 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -59,12 +59,12 @@ inline void push_string(lua_State *L, const QString &str)
   lua_pushstring(L, str.toUtf8());
 }
 
-inline QString tostring(lua_State *L, int i)
+inline QString toqstring(lua_State *L, int i)
 {
   return QString::fromUtf8(lua_tostring(L, i));
 }
 
-inline QString checkstring(lua_State *L, int i)
+inline QString checkqstring(lua_State *L, int i)
 {
   return QString::fromUtf8(luaL_checkstring(L, i));
 }
@@ -236,16 +236,16 @@ void PDialog::setMapped(lua_State *L, int idx)
   QWidget *w = iWidgets[idx];
   switch (m.type) {
   case ELabel:
-    (qobject_cast<QLabel *>(w))->setText(QString::fromUtf8(m.text.z()));
+    (qobject_cast<QLabel *>(w))->setText(QString::fromUtf8(m.text.c_str()));
     break;
   case ECheckBox:
     (qobject_cast<QCheckBox *>(w))->setChecked(m.value);
     break;
   case ETextEdit:
-    (qobject_cast<QTextEdit *>(w))->setText(QString::fromUtf8(m.text.z()));
+    (qobject_cast<QTextEdit *>(w))->setText(QString::fromUtf8(m.text.c_str()));
     break;
   case EInput:
-    (qobject_cast<QLineEdit *>(w))->setText(QString::fromUtf8(m.text.z()));
+    (qobject_cast<QLineEdit *>(w))->setText(QString::fromUtf8(m.text.c_str()));
     break;
   case EList:
     {
@@ -253,7 +253,7 @@ void PDialog::setMapped(lua_State *L, int idx)
       if (!lua_isnumber(L, 3)) {
 	l->clear();
 	for (int k = 0; k < int(m.items.size()); ++k)
-	  l->addItem(QString::fromUtf8(m.items[k].z()));
+	  l->addItem(QString::fromUtf8(m.items[k].c_str()));
       }
       l->setCurrentRow(m.value);
     }
@@ -264,7 +264,7 @@ void PDialog::setMapped(lua_State *L, int idx)
       if (!lua_isnumber(L, 3)) {
 	b->clear();
 	for (int k = 0; k < int(m.items.size()); ++k)
-	  b->addItem(QString::fromUtf8(m.items[k].z()));
+	  b->addItem(QString::fromUtf8(m.items[k].c_str()));
       }
       b->setCurrentIndex(m.value);
     }
@@ -279,7 +279,7 @@ bool PDialog::buildAndRun(int w, int h)
   for (int i = 0; i < int(iElements.size()); ++i) {
     SElement &m = iElements[i];
     if (m.row < 0) {
-      QPushButton *b = new QPushButton(QString::fromUtf8(m.text.z()), this);
+      QPushButton *b = new QPushButton(QString::fromUtf8(m.text.c_str()), this);
       iWidgets.push_back(b);
       if (m.flags & EAccept) {
 	b->setDefault(true);
@@ -290,14 +290,14 @@ bool PDialog::buildAndRun(int w, int h)
 	connect(b, SIGNAL(clicked()), SLOT(notifyLua()));
       iButtonArea->addWidget(b);
     } else {
-      QWidget *w = 0;
+      QWidget *w = nullptr;
       switch (m.type) {
       case ELabel:
-	w = new QLabel(QString::fromUtf8(m.text.z()), this);
+	w = new QLabel(QString::fromUtf8(m.text.c_str()), this);
 	break;
       case EButton:
 	{
-	  QPushButton *b = new QPushButton(QString::fromUtf8(m.text.z()), this);
+	  QPushButton *b = new QPushButton(QString::fromUtf8(m.text.c_str()), this);
 	  if (m.flags & EAccept)
 	    connect(b, SIGNAL(clicked()), SLOT(accept()));
 	  else if (m.flags & EReject)
@@ -309,7 +309,7 @@ bool PDialog::buildAndRun(int w, int h)
 	break;
       case ECheckBox:
 	{
-	  QCheckBox *ch = new QCheckBox(QString::fromUtf8(m.text.z()), this);
+	  QCheckBox *ch = new QCheckBox(QString::fromUtf8(m.text.c_str()), this);
 	  if (m.lua_method != LUA_NOREF)
 	    connect(ch, SIGNAL(stateChanged(int)), SLOT(notifyLua()));
 	  ch->setChecked(m.value);
@@ -319,7 +319,7 @@ bool PDialog::buildAndRun(int w, int h)
       case EInput:
 	{
 	  QLineEdit *e = new QLineEdit(this);
-	  e->setText(QString::fromUtf8(m.text.z()));
+	  e->setText(QString::fromUtf8(m.text.c_str()));
 	  if (m.flags & ESelectAll)
 	    e->selectAll();
 	  w = e;
@@ -335,7 +335,7 @@ bool PDialog::buildAndRun(int w, int h)
 	    (void) new XmlHighlighter(t);
 	  else if (m.flags & ELatex)
 	    (void) new LatexHighlighter(t);
-	  QString text = QString::fromUtf8(m.text.z());
+	  QString text = QString::fromUtf8(m.text.c_str());
 	  if (m.flags & ELogFile) {
 	    markupLog(t, text);
 	  } else
@@ -349,7 +349,7 @@ bool PDialog::buildAndRun(int w, int h)
 	{
 	  QComboBox *b = new QComboBox(this);
 	  for (int k = 0; k < int(m.items.size()); ++k)
-	    b->addItem(QString::fromUtf8(m.items[k].z()));
+	    b->addItem(QString::fromUtf8(m.items[k].c_str()));
 	  b->setCurrentIndex(m.value);
 	  w = b;
 	}
@@ -358,7 +358,7 @@ bool PDialog::buildAndRun(int w, int h)
 	{
 	  QListWidget *l = new QListWidget(this);
 	  for (int k = 0; k < int(m.items.size()); ++k)
-	    l->addItem(QString::fromUtf8(m.items[k].z()));
+	    l->addItem(QString::fromUtf8(m.items[k].c_str()));
 	  w = l;
 	}
 	break;
@@ -369,6 +369,8 @@ bool PDialog::buildAndRun(int w, int h)
       gridlayout()->addWidget(w, m.row, m.col, m.rowspan, m.colspan);
       if (m.flags & EFocused)
 	w->setFocus(Qt::OtherFocusReason);
+      if (m.flags & EDisabled)
+	w->setEnabled(false);
     }
   }
   setMinimumSize(w, h);
@@ -384,11 +386,11 @@ void PDialog::retrieveValues()
     QWidget *w = iWidgets[i];
     switch (m.type) {
     case EInput:
-      m.text =
-	String((qobject_cast<QLineEdit *>(w))->text().toUtf8());
+      m.text = std::string((qobject_cast<QLineEdit *>(w))->text().toUtf8());
       break;
     case ETextEdit:
-      m.text = String((qobject_cast<QTextEdit *>(w))->toPlainText().toUtf8());
+      m.text = std::string((qobject_cast<QTextEdit *>(w))->
+			   toPlainText().toUtf8());
       break;
     case EList:
       {
@@ -427,7 +429,7 @@ static int dialog_constructor(lua_State *L)
   const char *s = luaL_checkstring(L, 2);
 
   Dialog **dlg = (Dialog **) lua_newuserdata(L, sizeof(Dialog *));
-  *dlg = 0;
+  *dlg = nullptr;
   luaL_getmetatable(L, "Ipe.dialog");
   lua_setmetatable(L, -2);
   *dlg = new PDialog(L, parent, s);
@@ -461,8 +463,8 @@ PMenu::~PMenu()
 
 int PMenu::execute(lua_State *L)
 {
-  int vx = (int)luaL_checkinteger(L, 2);
-  int vy = (int)luaL_checkinteger(L, 3);
+  int vx = (int)luaL_checknumber(L, 2);
+  int vy = (int)luaL_checknumber(L, 3);
   QAction *a = iMenu->exec(QPoint(vx, vy));
   MenuAction *ma = qobject_cast<MenuAction *>(a);
   if (ma) {
@@ -487,8 +489,8 @@ static QIcon colorIcon(double red, double green, double blue)
 
 int PMenu::add(lua_State *L)
 {
-  QString name = checkstring(L, 2);
-  QString title = checkstring(L, 3);
+  QString name = checkqstring(L, 2);
+  QString title = checkqstring(L, 3);
   if (lua_gettop(L) == 3) {
     iMenu->addAction(new MenuAction(name, 0, QString(), title, iMenu));
   } else {
@@ -504,19 +506,19 @@ int PMenu::add(lua_State *L)
     if (hascheck) {
       luaL_argcheck(L, lua_isstring(L, 6), 6,
 		    "argument is not a function or string");
-      current = checkstring(L, 6);
+      current = checkqstring(L, 6);
     }
     int no = lua_rawlen(L, 4);
     QMenu *sm = new QMenu(title, iMenu);
     for (int i = 1; i <= no; ++i) {
       lua_rawgeti(L, 4, i);
       luaL_argcheck(L, lua_isstring(L, -1), 4, "items must be strings");
-      QString item = tostring(L, -1);
+      QString item = toqstring(L, -1);
       QString text = item;
       if (hastable) {
 	lua_rawgeti(L, 5, i);
 	luaL_argcheck(L, lua_isstring(L, -1), 5, "labels must be strings");
-	text = tostring(L, -1);
+	text = toqstring(L, -1);
 	lua_pop(L, 1);
       }
       if (hasmap) {
@@ -526,7 +528,7 @@ int PMenu::add(lua_State *L)
 	lua_call(L, 2, 1);     // function returns label
 	luaL_argcheck(L, lua_isstring(L, -1), 5,
 		      "function does not return string");
-	text = tostring(L, -1);
+	text = toqstring(L, -1);
 	lua_pop(L, 1);         // pop result
       }
       MenuAction *ma = new MenuAction(name, i, item, text, sm);
@@ -559,7 +561,7 @@ static int menu_constructor(lua_State *L)
 {
   QWidget *parent = check_winid(L, 1);
   Menu **m = (Menu **) lua_newuserdata(L, sizeof(PMenu *));
-  *m = 0;
+  *m = nullptr;
   luaL_getmetatable(L, "Ipe.menu");
   lua_setmetatable(L, -2);
 
@@ -594,7 +596,7 @@ void EditorThread::run()
 class EditorDialog : public QDialog
 {
 public:
-  EditorDialog(QWidget *parent = 0);
+  EditorDialog(QWidget *parent = nullptr);
 protected:
   void closeEvent(QCloseEvent *ev);
 };
@@ -616,7 +618,7 @@ void EditorDialog::closeEvent(QCloseEvent *ev)
 
 static int ipeui_wait(lua_State *L)
 {
-  QString cmd = checkstring(L, 2);
+  QString cmd = checkqstring(L, 2);
 
   EditorDialog *dialog = new EditorDialog();
 
@@ -684,7 +686,7 @@ static int timer_constructor(lua_State *L)
   const char *method = luaL_checkstring(L, 2);
 
   Timer **t = (Timer **) lua_newuserdata(L, sizeof(Timer *));
-  *t = 0;
+  *t = nullptr;
   luaL_getmetatable(L, "Ipe.timer");
   lua_setmetatable(L, -2);
 
@@ -706,7 +708,7 @@ static int timer_constructor(lua_State *L)
 static int ipeui_getColor(lua_State *L)
 {
   QWidget *parent = check_winid(L, 1);
-  QString title = checkstring(L, 2);
+  QString title = checkqstring(L, 2);
   QColor initial = QColor::fromRgbF(luaL_checknumber(L, 3),
 				    luaL_checknumber(L, 4),
 				    luaL_checknumber(L, 5));
@@ -728,11 +730,11 @@ static int ipeui_getColor(lua_State *L)
 
 static int ipeui_fileDialog(lua_State *L)
 {
-  static const char * const typenames[] = { "open", "save", 0 };
+  static const char * const typenames[] = { "open", "save", nullptr };
 
   QWidget *parent = check_winid(L, 1);
-  int type = luaL_checkoption(L, 2, 0, typenames);
-  QString caption = checkstring(L, 3);
+  int type = luaL_checkoption(L, 2, nullptr, typenames);
+  QString caption = checkqstring(L, 3);
   if (!lua_istable(L, 4))
     luaL_argerror(L, 4, "table expected for filters");
   QStringList filters;
@@ -740,16 +742,16 @@ static int ipeui_fileDialog(lua_State *L)
   for (int i = 1; i <= nFilters; i += 2) { // skip Windows versions
     lua_rawgeti(L, 4, i);
     luaL_argcheck(L, lua_isstring(L, -1), 4, "filter entry is not a string");
-    filters += checkstring(L, -1);
+    filters += checkqstring(L, -1);
     lua_pop(L, 1); // element i
   }
 
   QString dir;
   if (!lua_isnoneornil(L, 5))
-    dir = checkstring(L, 5);
+    dir = checkqstring(L, 5);
   QString name;
   if (!lua_isnoneornil(L, 6))
-    name = checkstring(L, 6);
+    name = checkqstring(L, 6);
   int selected = 0;
   if (!lua_isnoneornil(L, 7))
     selected = luaL_checkinteger(L, 7);
@@ -780,10 +782,9 @@ static int ipeui_fileDialog(lua_State *L)
       push_string(L, fns[0]);
       QString f = dialog.selectedNameFilter();
       int sf = 0;
-      while (filters[sf] != f && sf < nFilters/2)
+      while (sf < filters.size() && filters[sf] != f)
 	++sf;
-      if (sf < nFilters/2)
-	lua_pushinteger(L, sf + 1);
+      lua_pushinteger(L, (sf < filters.size()) ? sf + 1 : 0);
       return 2;
     }
   }
@@ -795,27 +796,26 @@ static int ipeui_fileDialog(lua_State *L)
 static int ipeui_messageBox(lua_State *L)
 {
   static const char * const options[] =  {
-    "none", "warning", "information", "question", "critical", 0 };
+    "none", "warning", "information", "question", "critical", nullptr };
   static const char * const buttontype[] = {
     "ok", "okcancel", "yesnocancel", "discardcancel",
-    "savediscardcancel", 0 };
+    "savediscardcancel", nullptr };
 
   QWidget *parent = check_winid(L, 1);
   int type = luaL_checkoption(L, 2, "none", options);
-  QString text = checkstring(L, 3);
-  const char *details = 0;
+  QString text = checkqstring(L, 3);
+  QString details;
   if (!lua_isnoneornil(L, 4))
-    details = luaL_checkstring(L, 4);
+    details = checkqstring(L, 4);
   int buttons = 0;
   if (lua_isnumber(L, 5))
-    buttons = (int)luaL_checkinteger(L, 5);
+    buttons = luaL_checkinteger(L, 5);
   else if (!lua_isnoneornil(L, 5))
-    buttons = luaL_checkoption(L, 5, 0, buttontype);
+    buttons = luaL_checkoption(L, 5, nullptr, buttontype);
 
   QMessageBox msgBox(parent);
   msgBox.setText(text);
-  if (details)
-    msgBox.setInformativeText(QString::fromUtf8(details));
+  msgBox.setInformativeText(details);
 
   switch (type) {
   case 0:
@@ -901,7 +901,7 @@ static const struct luaL_Reg ipeui_functions[] = {
   { "messageBox", ipeui_messageBox },
   { "waitDialog", ipeui_wait },
   { "currentDateTime", ipeui_currentDateTime },
-  { 0, 0},
+  { nullptr, nullptr }
 };
 
 // --------------------------------------------------------------------
